//  -*-Mode: C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
//
// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/FDInterest.cc,v $
// $Revision: 0.7 $
// $Date: 1998/08/04 18:03:12 $
// $Log: FDInterest.cc,v $
// Revision 0.7  1998/08/04 18:03:12  sellden
// Added workProc and CleanupProc
//
// Revision 0.6  1997/12/23 22:30:35  sellden
// Added support for X to perform polling for FDInterest instances
//
// Revision 0.5  1997/02/21 18:35:17  barnaby
// Modified port and network address setting
//
// Revision 0.4  1997/02/13 20:29:57  sellden
// Moved poll/select services to UltraPoll.cc/UltraPoll.h
//
// Revision 0.3  1997/01/21 03:58:13  sellden
// Wrapped reference to fakepoll.h w/ifdef __linux__.
// This needs to be consolidated into a single file for all platforms.
//
// Revision 0.2  1996/12/11 18:36:31  barnaby
// removed sys/filio.h with linux define
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
// Revision 6.5  1995/12/03  21:47:11  keith
// Fixed file descriptor leak in OpenTCPSocket().
//
// Revision 6.4  1995/12/03  20:13:49  keith
// Fixed file descriptor leaks...
//
// Revision 6.3  1995/07/28  20:33:14  spine
// Added copyright notice.
//
// Revision 6.2  1995/07/25  19:00:14  beth
// Fixed stupid bug...
//
// Revision 6.1  1995/07/25  18:53:49  keith
// in MainLoop, made another check to NeedsCleanupP, before anything
// else, in an attempt to better handle unexpected error conditions.
//
// Revision 6.0  1995/05/20  22:20:21  beth
// Mercator Version 6
//
// Revision 1.5  1995/04/01  19:40:38  keith
// End of work by Keith
//
// Revision 1.4  1995/03/06  17:52:05  keith
// *** empty log message ***
//
// Revision 1.3  1995/01/19  22:22:30  keith
// Automatic global checkin
//
// Revision 1.3  1995/01/19  22:22:30  keith
// Automatic global checkin
//
// Revision 1.2  1994/11/16  20:56:18  beth
// *** empty log message ***
//
// Revision 1.1  1994/11/16  20:46:23  spine
// Initial revision
//
// Revision 1.2  1994/08/24  18:23:53  kathryn
// delete Buffer fix
//
// Revision 1.1  1994/05/24  19:49:25  stacy
// Initial revision
//
// Revision 1.1  1994/04/06  19:53:36  kathryn
// Initial revision
//
//
 
static const char rcsid[]="@(#) $Header: /usr/mercator/newsource/7.0/RCS/FDInterest.cc,v 0.7 1998/08/04 18:03:12 sellden Exp $";

#ifdef USONIX
#include <stdio.h>
#endif
#include "FDInterest.h"
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#ifdef linux	// TB: Linux does not have streams
#include <termios.h>
#else
#include <stropts.h>
#endif
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#ifndef linux	// TB: Linux does not have filio
#include <sys/filio.h>
#endif
#include <sys/param.h>
#include <sys/uio.h>
#include <sys/socket.h>
#ifdef USONIX
#include <sys/ioctl.h>
#endif
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <strings.h>
#include <sys/stat.h>
#ifndef USE_X_EVENTS
#include <poll.h>
#endif

#include "Error.h"
#include "Util.h"

#ifdef USONIX
#define String XTSTRING
#include<X11/Intrinsic.h>
#undef String
extern XtAppContext app_context;
#endif

#ifndef USONIX	// TB: Removed this unecessary
declare(array,FDInterestPtr);
#endif

int		FDInterest::inited = 0;
#ifndef USE_X_EVENTS
struct pollfd	*FDInterest::fds = NULL;
#endif
FDInterest	**FDInterest::callbacks = NULL;
int		FDInterest::maxFds = 0;
int		FDInterest::quit = 0;

#ifdef USE_X_EVENTS
const short	FDInterest::ARead = (short)1;
const short	FDInterest::AWrite = (short)2;
const short	FDInterest::AExcept = (short)4;
const short	FDInterest::AEverything = ARead | AWrite | AExcept;
#else
#ifdef __linux__
const short	FDInterest::ARead = POLLIN | POLLPRI;
const short	FDInterest::AWrite = POLLOUT | POLLWRNORM;
const short	FDInterest::AExcept = (short)0; // No exceptionals in Linux poll?
const short	FDInterest::AEverything = ARead | AWrite | AExcept;
#else
const short	FDInterest::ARead = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI;
const short	FDInterest::AWrite = POLLOUT | POLLWRNORM | POLLWRBAND;
const short	FDInterest::AExcept = POLLERR | POLLHUP | POLLNVAL;
const short	FDInterest::AEverything = ARead | AWrite | AExcept;
#endif // ! __linux__
#endif

FDInterest::FDInterest(const char *n) :
    fdiName(Strdup(n)),
    cleanup(0)
{
    Error(EL_DEBUG, "FDInterest::FDInterest",
	  "new interest ", fdiName, " registered.", NULL);
    
    // if this is the 1st time this is called, initialize the arrays
    if (!inited) {
	struct rlimit rl;
	int i;
    
	inited = 1;
	
	// Figure the maximum size of the process descriptor table.
	if (getrlimit(RLIMIT_NOFILE, &rl) == -1) {
	    Error(EL_ERROR, "FDInterest::FDInterest",
		  "Cannot determine maximum descriptor table size: ",
		  Perror(), "  (Setting to 256...)", NULL);
	    maxFds = 256;
	} else {
	    maxFds = rl.rlim_cur;
	}

	callbacks = new FDInterestPtr[maxFds];
	
#ifndef USE_X_EVENTS
	fds = new struct pollfd[maxFds];
	
	for (i = 0; i < maxFds; i++) {
	    callbacks[i] = (FDInterest *) NULL;
	    
	    fds[i].fd = -1;
	    fds[i].events = 0;
	    fds[i].revents = 0;
	}
#endif
    }
}


FDInterest::~FDInterest(void)
{
    register int i;

    for (i=0 ; i<maxFds ; i++) {
	if (callbacks[i] == this) {
	    // We found a descriptor owned by this instance.
#ifndef USE_X_EVENTS // getting rid of poll-related stuff
	    fds[i].fd = -1;
	    fds[i].events = 0;
	    fds[i].revents = 0;
#endif
	    
	    callbacks[i] = NULL;
	}
    }

    delete fdiName;
}

int
FDInterest::AddFDInterest(int fd)
{
    return FDInterest::AddFDInterest(fd, this);
}

int
FDInterest::AddFDInterest(int fd, FDInterest * fdi)
{
    if ((fd < 0) || (fd > maxFds)) {
	char tmp1[16];
	char tmp2[16];
	sprintf(tmp1, "%d", fd);
	sprintf(tmp2, "%d", maxFds);
	Error(EL_ERROR, "FDInterest::AddFDInterest",
	      "descriptor ", tmp1, " out of range 0-", tmp2, NULL);
	return 0;
    }

#ifndef USE_X_EVENTS // getting rid of poll-related stuff
    if (fds[fd].fd != -1 || callbacks[fd] != NULL) {
	Error(EL_ERROR, "FDInterest::AddFDInterest",
	      "cannot add an interest to a descriptor already in use.", NULL);
	return 0;
    }

    fds[fd].fd = fd;
    fds[fd].events = 0;
    callbacks[fd] = fdi;
#endif
    
    return 1;
}

int
FDInterest::RemoveFDInterest(int fd)
{
    return FDInterest::RemoveFDInterest(fd, this);
}

int
FDInterest::RemoveFDInterest(int fd, FDInterest * fdi)
{
    if ((fd < 0) || (fd > maxFds)) {
	char tmp1[16];
	char tmp2[16];
	sprintf(tmp1, "%d", fd);
	sprintf(tmp2, "%d", maxFds);
	Error(EL_ERROR, "FDInterest::RemoveFDInterest",
	      "descriptor ", tmp1, " out of range 0-", tmp2, NULL);
	return 0;
    }
    
#ifndef USE_X_EVENTS // getting rid of poll-related stuff
    if (fds[fd].fd == -1) {
	Error(EL_ERROR, "FDInterest::RemoveFDInterest",
	      "trying to remove nonexistent interest.", NULL);
	return 0;
    }
#endif

    if (callbacks[fd] != fdi) {
	Error(EL_ERROR, "FDInterest::RemoveFDInterest",
	      "Cannot remove someone else's interest.", NULL);
	return 0;
    }
#ifndef USE_X_EVENTS // getting rid of poll-related stuff
    if (fds[fd].fd != fd) {
	Error(EL_ERROR, "FDInterest::RemoveFDInterest",
	      "Consistency failure!?", NULL);
	return 0;
    }

    fds[fd].fd = -1;
    fds[fd].events = 0;
    fds[fd].revents = 0;
#endif

    callbacks[fd] = NULL;
    
    return 1;
}

short
FDInterest::GetIOMask(int fd) const
{
    if (callbacks[fd] != this) {
	Error(EL_ERROR, "FDInterest::GetIOMask",
	      "Can't get IO Mask for a descriptor we don't own.", NULL);
	return 0;
    }
    
#ifdef USE_X_EVENTS // getting rid of poll-related stuff
    return (short) 0;
#else
    return fds[fd].events;
#endif
}

void
FDInterest::SetIOMask(int fd, short iomask)
{
    if (callbacks[fd] != this) {
	Error(EL_ERROR, "FDInterest::SetIOMask",
	      "Can't set IO Mask for a descriptor we don't own.", NULL);
	return;
    }

#ifndef USE_X_EVENTS // getting rid of poll-related stuff
    fds[fd].events = iomask;
#endif
}

void
FDInterest::Shutdown(int fd)
{
    Error(EL_DEBUG, "FDInterest::Shutdown", "SHUTTING DOWN...", NULL);
}

static void
PrintPollCode(int fd, int events)
{
#ifndef USE_X_EVENTS // getting rid of poll-related stuff
    if (events & POLLIN)
	cerr << "FD " << fd << " POLLIN\n";
#ifndef __linux__
    if (events & POLLRDNORM)
	cerr << "FD " << fd << " POLLRDNORM\n";
    if (events & POLLRDBAND)
	cerr << "FD " << fd << " POLLRDBAND\n";
    if (events & POLLWRBAND)
	cerr << "FD " << fd << " POLLWRBAND\n";
#endif
    if (events & POLLPRI)
	cerr << "FD " << fd << " POLLPRI\n";
    if (events & POLLOUT)
	cerr << "FD " << fd << " POLLOUT\n";
    if (events & POLLWRNORM)
	cerr << "FD " << fd << " POLLWRNORM\n";
    if (events & POLLERR)
	cerr << "FD " << fd << " POLLERR\n";
    if (events & POLLHUP)
	cerr << "FD " << fd << " POLLHUP\n";
    if (events & POLLNVAL)
	cerr << "FD " << fd << " POLLNVAL\n";
#endif
}

//
// This is the nerve center of Mercator.  All I/O processing goes through this
// one routine.
//
void
FDInterest::MainLoop(void)
{
    register int i;
    XEvent *xevent=(XEvent*)new XEvent;
    
    while (!quit) {
	register int result;
	int cleaned;

#ifdef USE_X_EVENTS
	//	XtAppMainLoop(::app_context);

		      	    if((XtIMAll) & XtAppPending(::app_context)){
	      XtAppNextEvent(::app_context, xevent);
	      if(quit) break;
		XtDispatchEvent(xevent);
				}
		    result = 0;
#else

		      	    if((XtIMAll) & XtAppPending(::app_context)){
	      XtAppNextEvent(::app_context, xevent);
		XtDispatchEvent(xevent);
				}

	result = poll(fds, maxFds, 100);
	if (result == -1) {
	    // If the syscall was interrupted, just continue.
	    if (errno != EINTR) {
		Error(EL_ERROR, "FDInterest::MainLoop",
		      "select: ", Perror(), NULL);
	    }
	    continue;
	}

	for (i = 0; i < maxFds; i++) {
	    // if we've processed all the fds, then get out.
	    if (result == 0)
		break;
	    
	    //
	    // At every pass, and before we call HandleActivity, we
	    // check to see if any instances need cleanup.
	    //
	    // The instance can request that we delete him at
	    // this point.  Regardless, we clear our his entry in
	    // our callbacks array.  (Reasons why the instance
	    // might not want to be completely deleted include
	    // that it is interested in several descriptors).
	    //
	    if (callbacks[i] && callbacks[i]->NeedsCleanupP()) {
		Error(EL_DEBUG, "FDInterest::MainLoop",
		      "Cleaning up after ", callbacks[i]->fdiName, NULL);
		delete callbacks[i];
		callbacks[i] = NULL;
		cleaned = 1;
	    }
	    
	    // Skip if nothing happened on this descriptor.
	    if (fds[i].revents == 0)
		continue;
	    
	    result--;

	    // If this descriptor's peer has shutdown, we *may* need to close up
	    // (we detect possible shutdown with 0 bytes to read).
	    if (fds[i].revents & ARead) {
		int bytes = 0;
		
#ifdef linux	// TB: No streams in linux
		if (ioctl(fds[i].fd, FIONREAD, &bytes) == -1) {
#else
		if (ioctl(fds[i].fd, I_NREAD, &bytes) == -1) {
#endif
		    Error(EL_ERROR, "FDInterest::MainLoop",
			  "ioctl NREAD failed: ", Perror(), NULL);
		} else if (bytes == 0) {
		    if (callbacks[i]) {
			Error(EL_DEBUG, "FDInterest::MainLoop",
			      "Peer of ", callbacks[i]->fdiName,
			      ": zero length read.", NULL);
		    }

		    //
		    // This is the code for cleaning up on a socket which appears
		    // to have gone away.  We do the following:
		    //   - Tell the instance that this descriptor appears to have
		    //     shutdown, via the Shutdown() method.
		    //   - See if the instance has been flagged for deletion via
		    //     the NeedsCleanupP() method. 
		    //
		    if (callbacks[i]) {
			FDInterest *instance = callbacks[i];
			
			instance->Shutdown(i);
			//
			// The instance can request that we delete him at
			// this point.  Regardless, we clear our his entry in
			// our callbacks array.  (Reasons why the instance
			// might not want to be completely deleted include
			// that it is interested in several descriptors).
			//
			if (instance->NeedsCleanupP()) {
			    Error(EL_DEBUG, "FDInterest::MainLoop",
				  "Cleaning up after ",
				  instance->fdiName, NULL);
			    delete instance;
			    callbacks[i] = NULL;
			    cleaned = 1;
			}
		    }
		}
	    }


	    if (callbacks[i]) {
		callbacks[i]->HandleActivity(i, fds[i].revents);
	    } else {
		//
		// clean out the pollset.  If the callback was NULL as a result of
		// a cleanup, then don't complain.  Otherwise something's going
		// on that's weird...
		//
		fds[i].fd = -1;
		fds[i].events = 0;
		if (!cleaned) {
		    char tmp[16];
		    sprintf(tmp, "%d", i);
		    Error(EL_ERROR, "FDInterest::MainLoop",
			  "Invalid callback on descriptor ", tmp, NULL);
		    cleaned = 0;
		}
	    }
	}
    }

    Error(EL_DEBUG, "FDInterest::MainLoop", "Shutting down...\n", NULL);

    // Tell each descriptor to shut down.
    for (i=0 ; i<maxFds ; i++) {
	if (callbacks[i]) {
	    callbacks[i]->Shutdown(i);
	    
	    fds[i].fd = -1;
	}
    }

    // Clean up the ones that are ready to be cleaned up.
    for (i=0 ; i<maxFds ; i++) {
	if (callbacks[i] && callbacks[i]->NeedsCleanupP()) {
	    delete callbacks[i];
	    callbacks[i] = NULL;
	}
#endif // USONIX: XtDispatchEvent
    }

    Error(EL_DEBUG, "FDInterest::MainLoop", "...shutdown completed.\n", NULL);
    
    return;
}

Boolean FDInterest::workProc(XtPointer myData)
  {
    register int i;
	register int result;
	int cleaned;

	if(!quit) {
	result = poll(fds, maxFds, 100);
	if (result == -1) {
	    // If the syscall was interrupted, just continue.
	    if (errno != EINTR) {
		Error(EL_ERROR, "FDInterest::MainLoop",
		      "select: ", Perror(), NULL);
	    }
	    return False;
	} // result == -1

	for (i = 0; i < maxFds; i++) {
	    // if we've processed all the fds, then get out.
	    if (result == 0)
		break;
	    
	    //
	    // At every pass, and before we call HandleActivity, we
	    // check to see if any instances need cleanup.
	    //
	    // The instance can request that we delete him at
	    // this point.  Regardless, we clear our his entry in
	    // our callbacks array.  (Reasons why the instance
	    // might not want to be completely deleted include
	    // that it is interested in several descriptors).
	    //
	    if (callbacks[i] && callbacks[i]->NeedsCleanupP()) {
		Error(EL_DEBUG, "FDInterest::MainLoop",
		      "Cleaning up after ", callbacks[i]->fdiName, NULL);
		delete callbacks[i];
		callbacks[i] = NULL;
		cleaned = 1;
	    }
	    
	    // Skip if nothing happened on this descriptor.
	    if (fds[i].revents == 0)
		continue;
	    
	    result--;

	    // If this descriptor's peer has shutdown, we *may* need to close up
	    // (we detect possible shutdown with 0 bytes to read).
	    if (fds[i].revents & ARead) {
		int bytes = 0;
		
#ifdef linux	// TB: No streams in linux
		if (ioctl(fds[i].fd, FIONREAD, &bytes) == -1) {
#else
		if (ioctl(fds[i].fd, I_NREAD, &bytes) == -1) {
#endif
		    Error(EL_ERROR, "FDInterest::MainLoop",
			  "ioctl NREAD failed: ", Perror(), NULL);
		} else if (bytes == 0) {
		    if (callbacks[i]) {
			Error(EL_DEBUG, "FDInterest::MainLoop",
			      "Peer of ", callbacks[i]->fdiName,
			      ": zero length read.", NULL);
		    }

		    //
		    // This is the code for cleaning up on a socket which appears
		    // to have gone away.  We do the following:
		    //   - Tell the instance that this descriptor appears to have
		    //     shutdown, via the Shutdown() method.
		    //   - See if the instance has been flagged for deletion via
		    //     the NeedsCleanupP() method. 
		    //
		    if (callbacks[i]) {
			FDInterest *instance = callbacks[i];
			
			instance->Shutdown(i);
			//
			// The instance can request that we delete him at
			// this point.  Regardless, we clear our his entry in
			// our callbacks array.  (Reasons why the instance
			// might not want to be completely deleted include
			// that it is interested in several descriptors).
			//
			if (instance->NeedsCleanupP()) {
			    Error(EL_DEBUG, "FDInterest::MainLoop",
				  "Cleaning up after ",
				  instance->fdiName, NULL);
			    delete instance;
			    callbacks[i] = NULL;
			    cleaned = 1;
			} // instance->needsCleanup
		    }
		}
	    }

	    if (callbacks[i]) {
		callbacks[i]->HandleActivity(i, fds[i].revents);
	    } else {
		//
		// clean out the pollset.  If the callback was NULL as a result of
		// a cleanup, then don't complain.  Otherwise something's going
		// on that's weird...
		//
		fds[i].fd = -1;
		fds[i].events = 0;
		if (!cleaned) {
		    char tmp[16];
		    sprintf(tmp, "%d", i);
		    Error(EL_ERROR, "FDInterest::MainLoop",
			  "Invalid callback on descriptor ", tmp, NULL);
		    cleaned = 0;
		}
	    }
	}

	    return False; // Do not remove from WorkProc list
	} // if !quit
	else {
    Error(EL_DEBUG, "FDInterest::MainLoop", "Shutting down...\n", NULL);

    // Tell each descriptor to shut down.
    for (i=0 ; i<maxFds ; i++) {
	if (callbacks[i]) {
	    callbacks[i]->Shutdown(i);
	    
	    fds[i].fd = -1;
	}
    }

    // Clean up the ones that are ready to be cleaned up.
    for (i=0 ; i<maxFds ; i++) {
	if (callbacks[i] && callbacks[i]->NeedsCleanupP()) {
	    delete callbacks[i];
	    callbacks[i] = NULL;
	}
    }

    Error(EL_DEBUG, "FDInterest::WorkProc",
	  "About to return True", NULL);

    (*FDInterest::CleanupProc)();
    return True;  // Remove from WorkProc list
	}
    }

int
OpenTCPSocket(const char *name, int port)
{
    int fd;
    struct sockaddr_in addr;

    /* set up address to bind to */
    memset((void *) &addr, '\0', sizeof(struct sockaddr_in));
#ifdef USONIX	/* TB */
    addr.sin_port = htons(port);
#else
    addr.sin_port = port;
#endif
    addr.sin_family = AF_INET;

    if (name == NULL) {
	// We are a server.  Use any IP address on this host.
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
    } else {
	struct hostent * hp;

	// Get the address of the host we want to talk to 
	hp = gethostbyname(name);
	if (hp == 0) {
	    Error(EL_ERROR, "OpenTCPSocket", "unknown host: ", name, NULL);
	    return -1;
	}
    
	// fill in the IP address and port
#ifdef USONIX	/* TB */
	addr.sin_addr.s_addr = * (u_long * ) hp->h_addr_list[0];
#else
	addr.sin_addr.s_addr = htonl(* (u_long * ) hp->h_addr_list[0]);
#endif
    }

    // create our stream socket
    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
	Error(EL_ERROR, "OpenTCPSocket", "opening socket: ", Perror(), NULL);
	close(fd);
	return -1;
    }

    // set some options to make it all nice
    if (name == NULL) {
	int one = 1;
#ifdef SVR4
	struct linger linger;

	linger.l_onoff = 0;
	linger.l_linger = 0;
#else
	int linger = 0;
#endif /* SVR4 */

	if ((setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
			(char *) &one, sizeof(int))) || 
	    (setsockopt(fd, SOL_SOCKET, SO_LINGER,
			(char *) &linger, sizeof(linger))) || 
	    (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
			(char *) &one, sizeof(int)))) {
	    perror("setsockopt");
	    close(fd);
	    return -1;
	}
    }

    if (name == NULL) {
	Error(EL_DEBUG, "OpenTCPSocket", "Binding to ",
	      inet_ntoa(addr.sin_addr), NULL);
	if (bind(fd, (struct sockaddr *) &addr,
		 sizeof(struct sockaddr_in)) < 0) {
	    Error(EL_ERROR, "OpenTCPSocket", "bind: ", Perror(), NULL);
	    close(fd);
	    return -1;
	}

	/* being accepting connections */
	listen(fd, 1);
    
    } else {
	Error(EL_DEBUG, "OpenTCPSocket", "Connecting to ",
	      inet_ntoa(addr.sin_addr), NULL);
	if (connect(fd, (struct sockaddr *) &addr,	
		    sizeof(struct sockaddr_in)) < 0 ) { 
	    Error(EL_ERROR, "OpenTCPSocket", "connect: ", Perror(), NULL);
	    close(fd);
	    return -1;   
	}
    }

    return fd;
}



/* **************************** Buffer Class ****************************** */


Buffer::Buffer() :
    data(NULL),
    index(0),
    size(0)
{
}

Buffer::Buffer(int sz) :
    data(new char[sz]),
    index(0),
    size(sz)
{
    memset(data, 0, size);
}

Buffer::~Buffer()
{
    if (data)
	delete [] data;
    index = size = 0;
}

int
Buffer::Full() const
{
    return (size && (index == size));
}

int
Buffer::Empty() const
{
    return (index == 0);
}

void
Buffer::Clear()
{
    if (data)
	delete [] data;
    
    data = NULL;
    size = index = 0;
}

int
Buffer::GetInt(void)
{
    if (this->size >= sizeof(int)) {
	int i;

	i = *((int *) data);  // yank out an int
	index -= sizeof(int); // decrement the index of the available space
	if (index)            // copy the leftover data to the beginning
	    memcpy(data, data + sizeof(int), index + 1);
	return i;
    } else {
	Error(EL_ERROR, "Buffer::GetInt", "not enough data for an int.", NULL);
	return 0;
    }
}

//
// This kind of assumes that input is non-blocking.
// Otherwise it would be pretty pointless.  Not that it isn't.
//
void
Buffer::Write(int fd)
{
    int result;

    if (index <= 0)
	return;

    result = write(fd, data, index);
    if (result == -1) {
	Error(EL_ERROR, "Buffer::Write", "write: ", Perror(), NULL);
	return;
    }

    // KEITH HACK
    fflush(fdopen(fd, "w"));

    if (size - result)
	memcpy(data, data + result, size - result);  // move up the data.
    
    index -= result;  // remove the number of bytes we used.
}

void
Buffer::Join(const char *indata, int insize)
{
    if (insize > size - index)
	this->Resize(insize + size);  // make it bigger, plus a little.

    memcpy(data + index, indata, insize);
    index += insize;
}

void
Buffer::Resize(int new_size)
{
    char *new_data = new char[new_size];

    if (new_data == NULL) {
	char tmp[16];
	sprintf(tmp, "%d", new_size);
	Error(EL_FATAL, "Buffer::Resize", "Couldn't allocate ", tmp,
	      " bytes.", NULL);
	exit(1);
    }

    if (data != NULL) {
	memcpy(new_data, data, size);
	delete [] data;
    }

    data = new_data;
    size = new_size;
}

int
Buffer::Read(int fd)
{
    int result;
    int toread;  // total bytes available to be read.
    int sz;      // bytes we actually want to read.
    
    if (Full())
	return 0;

    result = ioctl(fd, FIONREAD, &toread);
    if (result == -1) {
	perror("ioctl");
	return 0;
    }

    if (toread == 0)
	return 0;

    //
    // If this packet is empty then we need to read in
    // the size of it and resize.
    //
    if (size == 0) {
	int result, sz;
    
	if (toread < sizeof(unsigned long))
	    return 0;
    
	result = read(fd, (char *) &sz, sizeof(unsigned long));

	// complain if we didn't read enough or if we got an empty
	// packet from someone on last read
	if (result != sizeof(unsigned long)) {
	    Error(EL_ERROR, "Buffer::Read", "read: ", Perror(), NULL);
	    return 0;
	} else if (sz == 0) {
	    Error(EL_WARNING, "Buffer::Read", "Got an empty packet.", NULL);
	    return 0;
	}
    
	toread -= sizeof(unsigned long);
	Resize((int) sz);
    }

    // don't read more than we need.
    sz = toread;
    if (toread > (size - index))
	sz = size - index;

    result = read(fd, data + index, sz);
    if (result == -1) {
	Error(EL_ERROR, "Buffer::Read", "read: ", Perror(), NULL);
	return 0;
    } else if (result != sz) {
	Error(EL_ERROR, "Buffer::Read", "read less data then we wanted.", NULL);
    }

    index += result;
    return 1;
}


