// -*-Mode:C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: Mercator.h,v $
// Revision 0.3  1997/12/23 22:30:37  sellden
// Added SetRAPListener() method
//
// Revision 0.2  1996/12/11 18:38:49  barnaby
// Included stdio.h
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//

#ifndef __MERCATOR_H__
#define __MERCATOR_H__
 
//
// comments
//

//
// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/Mercator.h,v $
// $Revision: 0.3 $
// $Date: 1997/12/23 22:30:37 $
//

#ident "$Id: Mercator.h,v 0.3 1997/12/23 22:30:37 sellden Exp $"

#ifdef USONIX	// TB: Needed stio.h for FILE
#include <stdio.h>
#endif
#include <sys/types.h>
#include <ProcessManager.h>
#include "Error.h"

class Mercator;
class XServer;
class AppModelMgr;
class Audio;
class Speech;
class Interp;
class RAPListener;
class Braille;


/////
/////	There is one global Mercator instance that all classes share.
/////
extern Mercator *mercator;

class Mercator {
private:
    XServer      	*xServer;
    RAPListener	 	*rapListener;
    AppModelMgr  	*modelMgr;
    Audio 	    	*audio;
    Speech       	*speech;
    Braille		*braille;
    Interp	 	*interp;
    ProcessManager      *procMgr;
    
    static int		initialized;

    char		*configFilename;
    char		*attribFilename;

    int			tmpErrorLevel;
    char		*tmpErrorOutput;
    ErrorOutputs        tmpErrorOutputVal;
    char                *tmpFileName;
    fstream             tmpout;  // to save initial error/debug messages
    
    int                 infd;  // These will hold the input and output
    int                 outfd; // fd's, which may be named pipes or stdio
    static FILE         *outFile;
    
#if	defined(SUNOS_5_5) || defined(linux)	// TB: Linux signals have arg
    static void         PipeHandler(int);
    static void		InterruptHandler(int);
#else
    static void         PipeHandler();
    static void		InterruptHandler();
#endif /* SUNOS_5_5 */

public:
    Mercator();
    virtual ~Mercator();
 
    XServer     	*GetXServer() const;
    RAPListener		*GetRAPListener() const;
#ifdef USONIX
void SetRAPListener(RAPListener *newListener);
#endif
    AppModelMgr		*GetAppModelMgr() const;
    Audio	    	*GetAudio() const;
    Speech      	*GetSpeech() const;
    Braille		*GetBraille() const;
    Interp		*GetInterp() const;
    ProcessManager      *GetProcessManager() const;

    const int           GetOutfd() const;
    
    int		 	ProcessArgs(int argc, char *argv[]);
    void		ProcessEnvVariables();
    int			ParseConfigFile();
    int			ParseAttribFile();
    int			LoadTemplates();
    void                CreateNamedPipes();
    void		RemoveNamedPipes();
    void		ProcessDefaults();
    void		InstallSignalHandlers();

    void                LoadFilterDictionary();

    const char		*ConfigFilename() const;
    const char		*AttribFilename() const;

    int                 ExecList(const char * listName);

    void                StartConsole();
    int                 ConnectConsole();
    int                 ConnectConsole(long);

    void                SaveMsg();
    void                DumpMsg();

    void		FreeFilters();
    void		FreeTemplates();
    void		FreeConfig();
    void		FreeAttrib();
    
    //
    // This is the single almighty exit point for Mercator.
    //
    void		Quit();
};

////////////////////////////////////////////////////////////////////////

inline XServer *
Mercator::GetXServer() const
{
    return xServer;
}

inline RAPListener *
Mercator::GetRAPListener() const
{
    return rapListener;
}

#ifdef USONIX
inline void
Mercator::SetRAPListener(RAPListener *newListener)
{
  rapListener = newListener;
}
#endif

inline AppModelMgr *
Mercator::GetAppModelMgr() const
{
    return modelMgr;
}

inline Audio	*
Mercator::GetAudio() const
{
    return audio;
}

inline Speech *
Mercator::GetSpeech() const
{
    return speech;
}

inline Braille *
Mercator::GetBraille() const
{
    return braille;
}

inline Interp *
Mercator::GetInterp() const
{
    return interp;
}

inline ProcessManager *
Mercator::GetProcessManager() const
{
    return procMgr;
}

inline const char *
Mercator::AttribFilename() const
{
    return attribFilename;
}

inline const char *
Mercator::ConfigFilename() const
{
    return configFilename;
}

inline const int
Mercator::GetOutfd() const
{
    return outfd;
}
#endif /* __MERCATOR_H__ */
