// -*-Mode: C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: RAP.h,v $
// Revision 0.4  1998/06/06 21:04:31  sellden
// *** empty log message ***
//
// Revision 0.3  1997/12/23 22:30:40  sellden
// Large-scale changes... Moved from GaTech's Rap implementation to Will Walker's implementation
//
// Revision 0.2  1996/12/11 18:30:09  barnaby
// defined String as XtString for Intrinsics.h
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//

#ifndef _RAP_H_
#define _RAP_H_

#include <X11/Xlib.h>		// Window
#ifdef USONIX	// TB: rename of String because of libg++ conflict
#define String	XtString
#include <X11/Intrinsic.h>
#undef String
#else
#include <X11/Intrinsic.h>	// WidgetList, Widget
#endif
#include <X11/ICE/ICElib.h>	// IceConn
#include "GCValue.h"
#include "FDInterest.h"
#include "mDictValPtr.h"

extern "C" {
#include <X11/Rap/RapAgnt.h>
}

class RAP;
class RAPListener;
class Client;
class XtObject;
typedef Client *ClientPtr;

typedef RAP *RAPPtr;

typedef mDictValPtr<IceConn,RAP>	RAPInstanceDict;
typedef mDictValPtr<Window,Client>	WinClientDict;

//
// This was in XtHook; we need it to get Interp.cc to compile.  Eventually
// it'll be part of the args to SetValues.
//
typedef struct _ResourceKeyValue {
    char *name;
    char *value;
} ResourceKeyValue;


////////////////////////////////////////////////////////////////////////


class RAP : public FDInterest {

    friend class RAPListener;
    
private:
    //
    // INSTANCE DATA
    //
#ifdef USONIX
  RapAgntClientPtr rap_client;
#endif
    IceConn		iceConn;
    Window		clientWindow;
    int			majorOpcode;
    Client		*client;
    IceConnectStatus	connectStatus;

    // RAP maintains a per-connection GC dict.
    GCValueDict		*gcDict;

    // RAP maintains a mapping from IceConns to instances.
    static RAPInstanceDict	instances;

    // Map from window IDs to clients to tell if we've ever connected to a given
    // client before.
    static WinClientDict	clients;

    RAP();					// default ctor - never called
    RAP(IceConn iceConn, int majorOpcode);	// only called from RAPListener

#ifdef  USONIX
    RAP(RapAgntClientPtr client, int majorOpcode);	// only called from RAPListener
#endif

    static void		ProcessProtoMsgProc(IceConn iceConn,
					    IcePointer clientData,
					    int opcode,
					    unsigned long length,
					    Bool swap);
    //
    // These are the methods that actually DO something with the
    // structures returned from the various Handle methods.  Some of
    // them have a one-to-one mapping with protocol, and some are
    // "helper" routines.
    //
    // Reply handlers.
#ifdef USONIX
    void		ProcessFullQueryTreeReply(RapAgntShellList *data);
    void		ProcessQueryTreeReply(RapAgntShellList *data);
    void ProcessRapAgntObjects(RapAgntObjectPtr rapObject);
    void		ProcessGetValuesReply(RapValuePtr data);
    void		ProcessGetResourcesReply(RapAgntResourceList *data);
    void		ProcessGetGCValuesReply(RapGCValuesList *data);
#else
    void		ProcessQueryTreeReply(QueryTreeReply *data);
    void		ProcessFullQueryTreeReply(FullQueryTreeReply *data);
    void		ProcessHelloReply(HelloReply *data);
    void		ProcessGetValuesReply(GetValuesReply *data);
    void		ProcessGetResourcesReply(GetResourcesReply *data);
    void		ProcessGetGCValuesReply(GetGCValuesReply *data);
    void		ProcessLocateObjectReply(LocateObjectReply *data);
#endif

    // Notify helpers
#ifdef USONIX
  //    void		ProcessDestroyNotify(DestroyNotif *data);
  //    void		ProcessConfigureNotify(ConfigureNotif *data);
  //    void		ProcessCreateNotify(CreateNotif *data);
  //    void		ProcessChangeNotify(ChangeNotif *data);
  //    void		ProcessGeometryNotify(GeometryNotif *data);
    void		ProcessRequestNotify(xReq *data);
    void		ProcessEventNotify(xEvent *xevent);
#else
    void		ProcessDestroyNotify(DestroyNotif *data);
    void		ProcessConfigureNotify(ConfigureNotif *data);
    void		ProcessCreateNotify(CreateNotif *data);
    void		ProcessChangeNotify(ChangeNotif *data);
    void		ProcessGeometryNotify(GeometryNotif *data);
    void		ProcessRequestNotify(RequestNotif *data);
    void		ProcessEventNotify(EventNotif *data);
#endif

    // Helper functions - state change
    void		ProcessToplevelMapped(XtObject *obj, int initial);

    void		ProcessWidgetChildrenManaged(XtObject *obj,
						     CARD32 *widgets,
						     CARD32 numWidgets);
    void		ProcessWidgetChildrenUnmanaged(XtObject *obj,
						       CARD32 *widgets,
						       CARD32 numWidgets);
    
    void		ProcessWidgetManaged(XtObject *obj);
    void		ProcessWidgetUnmanaged(XtObject *obj);
    void		ProcessWidgetMapped(XtObject *obj);
    void		ProcessWidgetUnmapped(XtObject *obj);
    void		ProcessWidgetRealized(XtObject *obj, Window w);
    void		ProcessWidgetUnrealized(XtObject *obj);

    void		MapAndGrabChildren(XtObject *obj, int mapState);
    void		ManageChildren(XtObject *obj, int manageState);
    // Helper functions - events
    void		ProcessKeyPressEvent(xEvent *event);
    void		ProcessEnterNotifyEvent(xEvent *event);
    void		ProcessUnmapNotifyEvent(xEvent *event);
    void		ProcessMapNotifyEvent(xEvent *event);
    void		ProcessButtonPressEvent(xEvent *event);
    // Helper functions - requests
    void		ProcessCreateGCRequest(xReq *request);
    void		ProcessChangeGCRequest(xReq *request);
    void		ProcessCopyGCRequest(xReq *request);
    void		ProcessFreeGCRequest(xReq *request);
    void		ProcessReparentWindowRequest(xReq *request);
    void		ProcessTextRequests(xReq *request);

    GCValue		*FindGC(GContext gid);
    
public:
    virtual ~RAP();

    virtual int		HandleActivity(int fd, short type);
    virtual void	Shutdown(int fd);

#ifdef USONIX
    void		SetRapClient(RapAgntClientPtr newClient);
    RapAgntClientPtr		GetRapClient() const;
Window GetClientWindow() const;
#endif

    void		SetClient(Client *c);
    Client		*GetClient() const;

    GCValueDict		*GetGCDict() const;

    // Look stuff up in the WinClientDict
    static Client	*FindClientByWindow(Window);

    //
    // This is called from XServer when we get a MapNotify from the root window
    // for an object that is in our model.  We treat this as notification of
    // a popup window appearing (or a shell, in any event), and call thisi
    // method.
    //
    void		ProcessXServerMapNotify(Window);

    //
    // Requests from agent
    //
    int			HelloRequest(int block = FALSE);
    int			QueryTreeRequest(int block = FALSE);
    int			FullQueryTreeRequest(int block = FALSE);
    int			GetResourcesRequest(WidgetList widgets,
					    long numWidgets,
					    int block = FALSE);
    int			GetValuesRequest(Object widget,
					 const char **resourceNames,
					 long numResources,
					 int block = FALSE);
    // Note that we use Object here... should probably use everywhere
    int			SetValuesRequest(Object w,
					 const ResourceKeyValue *values,
					 long numKeyValues,
					 int block = FALSE);
    
    int			GetGCValuesRequest(unsigned long *gcs, long numGCs,
					   int block = FALSE);
#ifdef NOT_IMPLEMENTED
    int			WindowToObjectRequest(int block = FALSE);
    int			ObjectToWindowRequest(int block = FALSE);
#endif /* NOT_IMPLEMENTED */
    int			LocateObjectRequest(Object w, int block = FALSE);
    int			SelectEventRequest(short *events, short numEvents,
					   int block = FALSE);
    int			SelectRequestRequest(short *reqs, short numReqs,
					     int block = FALSE);
    int			AddNotifyRequest(short *notifies, short numNotifies,
					 int block = FALSE);
    int			RemoveNotifyRequest(short *notifies, short numNotifies,
					    int block = FALSE);
    int			CloseConnectionRequest(int block = FALSE);
};


////////////////////////////////////////////////////////////////////////

inline void
RAP::SetClient(Client *c)
{
    client = c;
#ifdef USONIX
// moved from ProcessHelloReply, which is no longer called
    clients.insert(clientWindow, c);
#endif
}

inline Client *
RAP::GetClient() const
{
    return client;
}

#ifdef USONIX
inline void
RAP::SetRapClient(RapAgntClientPtr newClient)
{
    rap_client = newClient;
}

inline RapAgntClientPtr
RAP::GetRapClient() const
{
    return rap_client;
}

inline Window
RAP::GetClientWindow() const
{
    return clientWindow;
}
#endif

inline GCValueDict *
RAP::GetGCDict() const
{
    return gcDict;
}

#endif /* _RAP_H_ */
