//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: RAPListener.cc,v $
// Revision 0.4  1998/08/03 20:49:50  sellden
// Added initial support for Rap modules, added MarshallDict code
//
// Revision 0.3  1997/12/23 22:30:40  sellden
// Large-scale changes... Moved from GaTech's Rap implementation to Will Walker's implementation
//
// Revision 0.2  1997/02/13 20:29:57  sellden
// Moved poll/select services to UltraPoll.cc/UltraPoll.h
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//

#ifndef USE_X_EVENTS
#include <poll.h>
#endif

#include <sys/ioctl.h>
#include<X11/Xproto.h>
#include <X11/Xatom.h>		// For XA_STRING
#ifndef USONIX
#define String	XtString
#include <X11/Intrinsic.h>
#include<X11/Shell.h>

extern "C" {
#include <X11/Rap/Rap.h>
#include <X11/Rap/RapProto.h>		// Protocol definition
#include <X11/Rap/RapAgnt.h>		// C language agent API
  // Get rid of old Rap files
  //#include "agent.h"
#include<X11/Xtea/Xtea.h>
}
#undef String
#endif

#include "RAPListener.h"

#include "XServer.h"
#include "Error.h"
#include "Client.h"
#include "Mercator.h"
#include "AppModelMgr.h"
#ifdef RAP_MODULES
#include "XtRap.h"
#else
#include "RAP.h"
#endif

//
// There is one almighty global instance of "Mercator"
//
extern Mercator *mercator;

#ifdef USONIX
extern Widget rap_agnt;
extern Widget top_level;
extern XtAppContext app_context;
#endif

////////////////////////////////////////////////////////////////////////
//
// Xtea-related stuff
//
////////////////////////////////////////////////////////////////////////

#ifdef USONIX
const long ICE_PROTOCOLS_FORMAT= 32;
#else
const char *EXTAGENT_NAME	= "ExternalAgent";
const char *ICE_NETWORK_IDS	= "IceNetworkIds";

const long EXTAGENT_FORMAT	= 32;
#endif

static IcePaVersionRec protoVersions[1];

////////////////////////////////////////////////////////////////////////
//
// Class special functions.
//
////////////////////////////////////////////////////////////////////////

RAPListener::RAPListener(XServer& xserver) :
    FDInterest("RAPListener"),
    protocolVersions(new IcePaVersionRec),
    listeners(NULL),
    listenerCount(0),
    majorOpcode(-1),
    networkIds(NULL),
    agentWindow((long unsigned int)NULL),
    protoNameAtom((long unsigned int)NULL),
    marshallDict(new mDictValPtr<CARD32, RapMarshallStruct>),
    marshallSequenceNumber(0)
{
#ifdef USONIX
#ifndef USE_X_EVENTS
  // Create pipe for use by rap agent.
  // When toolkit support is modularized, this will have to be
  // done differently.

  pipe(pipeFDs[rap_xt]);
  AddFDInterest(pipeFDs[rap_xt][0]);
  SetIOMask(pipeFDs[rap_xt][0], ARead);

#endif
#else
    IceSetIOErrorHandler(RAPListener::IOErrorHandler);
    
    protoVersions[0].major_version = RapMajorVersion;
    protoVersions[0].minor_version = RapMinorVersion;

    // Note protocol msg handler lives in RAP, not RAPListener.
#ifdef RAP_MODULES
    protoVersions[0].process_msg_proc = RapBase::ProcessProtoMsgProc; 
#else
    protoVersions[0].process_msg_proc = RAP::ProcessProtoMsgProc; 
#endif
#endif
}

RAPListener::~RAPListener()
{
    register int i;
    
    delete marshallDict;
    delete protocolVersions;

#ifndef USONIX
    for (i=0 ; i<listenerCount ; i++) {
	RemoveFDInterest(IceGetListenConnectionNumber(listeners[i]));
    }
    
    // Call ICE to free listeners, networkIds
    IceFreeListenObjs(listenerCount, listeners);
#endif
}


////////////////////////////////////////////////////////////////////////
//
// Static member functions - mostly ICE-required boilerplate.
//
////////////////////////////////////////////////////////////////////////

Bool
RAPListener::AuthorizeProc(char *hostname)
{
    Error(EL_DEBUG, "RAPListener::AuthorizeProc", "auth for host ",
	  hostname, NULL);
    return TRUE;
}

void
RAPListener::IOErrorHandler(IceConn iceConn)
{
#ifdef RAP_MODULES
  RapBase *instance = NULL;
    if ((instance = RapBase::instances.lookup(iceConn)) == 0) {
	Error(EL_ERROR, "RAPListener::IOErrorHandler",
	      "No client is associated with this ice connection.  ",
	      "What do we do?", NULL);
	return;
    }
#else
    RAP *instance = NULL;
    
    if ((instance = RAP::instances.lookup(iceConn)) == 0) {
	Error(EL_ERROR, "RAPListener::IOErrorHandler",
	      "No client is associated with this ice connection.  ",
	      "What do we do?", NULL);
	return;
    }
#endif

    if (instance->GetClient())
        Error(EL_DEBUG, "RAPListener::IOErrorHandler", "Client ",
	      instance->GetClient()->ClientName(),
	      " seems to have exited.", NULL);
    else 
        Error(EL_DEBUG, "RAPListener::IOErrorHandler",
	      "A client seems to have exited.", NULL);

#ifdef USONIX
    // The fd is not used
    instance->Shutdown(0);
#else
    instance->Shutdown(IceConnectionNumber(iceConn));
#endif
}

void
RAPListener::IOErrorProc(IceConn iceConn)
{
#ifdef RAP_MODULES
  RapBase *instance = NULL;
#else
    RAP *instance = NULL;
#endif
    
    // Shutdown stuff?  This is what Will's sample code does.  I need to
    // understand how these two error procedures will get called better.

#ifdef RAP_MODULES
    if ((instance = RapBase::instances.lookup(iceConn)) == 0) {
#else
    if ((instance = RAP::instances.lookup(iceConn)) == 0) {
#endif
	Error(EL_ERROR, "RAPListener::IOErrorProc",
	      "No client is associated with this ice connection.  ",
	      "What do we do?", NULL);
	return;
    }

    Error(EL_DEBUG, "RAPListener::IOErrorProc", "Client ",
	  instance->GetClient()->ClientName(),
	  " seems to have exited.", NULL);

#ifdef USONIX
    // The fd is not used
    instance->Shutdown(0);
#else
    instance->Shutdown(IceConnectionNumber(iceConn));
#endif
}

void
RAPListener::ProtoActivateProc(IceConn iceConn, IcePointer clientData)
{
    Error(EL_DEBUG, "RAPListener::ProtoActivateProc", "called.", NULL);
}

Status
RAPListener::ProtoSetupProc(IceConn iceConn, int majorVersion, int minorVersion,
			    char *vendorString, char *releaseString,
			    IcePointer *clientDataRet, char **failureReasonRet)
{
    char tmp[16];
    
    sprintf(tmp, "%d.%d", majorVersion, minorVersion);

    Error(EL_DEBUG, "RAPListener::ProtoSetupProc", "\n",
	  "\tVersion: ", tmp, "\n",
	  "\tVendor: ", vendorString, "\n",
	  "\tRelease: ", releaseString, NULL);

    return 1;
}

#ifdef USONIX
Widget RAPListener::GetProc()
{
  return ::rap_agnt;
}

void RAPListener::WriteProc(RapMarshallStruct *rms) 
{
  if(rms != NULL) {
    char buf[32];

    marshallDict->insert(marshallSequenceNumber, rms);
    printf("RAPListener::WriteProc - marshallSequenceNumber is %u\n", marshallSequenceNumber);

    // write the seq number followed by a non-numeric for parsing later
    sprintf(buf, "%u*", marshallSequenceNumber++);
    write(pipeFDs[rap_xt][1], buf, strlen(buf));
  }
}

void RAPListener::ClientProcMarshaller(Widget 			w, 
				       XtPointer			client_data, 
				       RapAgntClientCallbackStruct 	*callback_data)
{
  if(!mercator->GetRAPListener()->NeedsCleanupP()) {
    RapMarshallStruct *rms = new RapMarshallStruct;
    RapAgntClientCallbackStruct *callbackStruct = new RapAgntClientCallbackStruct;
    RapAgntClient *rapClnt = new RapAgntClient;

    *rapClnt = *(callback_data->client);
    *callbackStruct = *callback_data;
    callbackStruct->client = rapClnt;

    rms->raw = w;
    rms->client_data = (XtPointer)client_data;
    rms->callbackStruct = (XtPointer)callbackStruct;
    rms->reason = rap_ClientProc;
    rms->size = sizeof(RapMarshallStruct);

    mercator->GetRAPListener()->WriteProc(rms);
  }
}

void RAPListener::ClientProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntClientCallbackStruct 	*callback_data)
{
  mString reason;
    RapAgntObjectPtr 	object;
    object = (RapAgntObjectPtr)callback_data->client->client_data;

    /* Now set the sensitivity or destroy the client depending upon
     * why this routine was called.  An insensitive widget means we
     * tried connecting with the client using the ICE X Rendezvous
     * protocol, but it hasn't responded yet.
     */
    switch(callback_data->reason) {
	case RapAgntNewClient: {
#ifdef USONIX
	  reason = "RapAgntNewClient";
	  if(object == NULL)
	    Error( EL_WARNING, "RAPListener::ClientProc",
		   "object was NULL", NULL);
	  else {
	    XtSetSensitive((Widget)object->client_data,False);
	  }
#else
XtSetSensitive((Widget)object->client_data,False);
#endif
	    break;
	}
	case RapAgntClientConnect: {
#ifdef USONIX
	  reason = "RapAgntClientConnect";
	  // If we've seen this client before, ignore it
#ifdef RAP_MODULES
	     if (RapBase::instances.lookup(callback_data->client->ice_conn) == 0){
#else
	     if (RAP::instances.lookup(callback_data->client->ice_conn) == 0){
#endif

	  if(object == NULL)
	    Error( EL_WARNING, "RAPListener::ClientProc",
		   "object was NULL", NULL);
	  else {
	    	    XtSetSensitive((Widget)object->client_data,True);
	  }

	  char ibuf[16];
	  sprintf(ibuf, "%lu", callback_data->client->ice_conn);

	  Error(EL_DEBUG, "RAPListener::ClientProc",
		"About to construct RAP instance with iceConn = ", ibuf, NULL);

	  // hack alert: sending '1' as opcode until I remove it
	  // from the arg list for constructor.
#ifdef RAP_MODULES
	    	  RapBase *rap = new XtRap(callback_data->client, 1);
#else
	    	  RAP *rap = new RAP(callback_data->client, 1);
#endif
		  // Bogus fd
		  rap->HandleActivity(0, ARead);
	  }
#else
XtSetSensitive((Widget)object->client_data,True);
#endif
	    break;
	}
	
	/* Remove the client completely.
	 */
	case RapAgntInitiateFailed: 
	case RapAgntClientDisconnect: {
#ifdef USONIX
	  reason = "RapAgntClientDisconnect or RapAgntInitiateFailed";
	  IceConn iceConn = callback_data->client->ice_conn;
#ifdef RAP_MODULES
	  RapBase *instance = RapBase::instances.lookup(iceConn);
#else
	  RAP *instance = RAP::instances.lookup(iceConn);
#endif

	  if(instance != NULL) {
	  // We can do this because the arg - FD - is not used
	  	      instance->Shutdown(0);
	  } else {
	    Error(EL_WARNING, "RAPListener::ClientProc",
		  "Trying to shut down Rap instance - ",
		  "RAP::instances.lookup failed", NULL);
	  }

#else
FreeClientInfo(callback_data->client);
	    XtDestroyWidget((Widget)object->client_data);
	    if (current_client == callback_data->client) {
		if (current_object && (current_object->client == current_client)) {
		    XtSetSensitive(object_button,False);
		    current_object = NULL;
		}
		XtSetSensitive(client_button,False);
		current_client = NULL;
	    }
#endif
	    break;
	}
    }

	// Allocated from ClientProcMarshaller
	delete callback_data;

	Error(EL_DEBUG, "RAPListener::ClientProc",
	      "Just Checking in...", (char*)reason, NULL);
	cout<<"Window ID = "<<((RapAgntClientPtr)callback_data->client->window)<<endl;
}

    void RAPListener::HookProcMarshaller(Widget 			w, 
					 XtPointer			client_data, 
					 RapAgntHookCallbackStruct 	*callback_data)
      {
	if(!mercator->GetRAPListener()->NeedsCleanupP()) {
	  RapMarshallStruct *rms = new RapMarshallStruct;
	  RapAgntHookCallbackStruct *callbackStruct = new RapAgntHookCallbackStruct;
	  RapAgntClientPtr client = new RapAgntClient;
	  RapAgntObjectPtr object = new RapAgntObject;

	  *callbackStruct = *callback_data;
	  *client = *(callbackStruct->client);
	  callbackStruct->client = client;
	  *object = *(callbackStruct->object);
	  callbackStruct->object = object;

	  rms->raw = w;
	  rms->client_data = (XtPointer)client_data;
	  rms->callbackStruct = (XtPointer)callback_data;
	  rms->reason = rap_HookProc;
	  rms->size = sizeof(RapMarshallStruct);

	  mercator->GetRAPListener()->WriteProc(rms);
	}
      }

void RAPListener::HookProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntHookCallbackStruct 	*callback_data)
{
  // Allocated in HookProcMarshaller
  delete callback_data->client;
  delete callback_data->object;
  delete callback_data;

	Error(EL_DEBUG, "RAPListener::HookProc",
	      "Just Checking in...", NULL);
}

 void RAPListener::ObjectInfoProcMarshaller(Widget 			w, 
					    XtPointer			client_data, 
					    RapAgntObjectInfoCallbackStruct 	*callback_data)
   {
     if(!mercator->GetRAPListener()->NeedsCleanupP()) {
       RapMarshallStruct *rms = new RapMarshallStruct;
       RapAgntObjectInfoCallbackStruct *callbackStruct = new RapAgntObjectInfoCallbackStruct;

       *callbackStruct = *callback_data;

       rms->raw = w;
       rms->client_data = (XtPointer)client_data;
       rms->callbackStruct = (XtPointer)callbackStruct;
       rms->reason = rap_ObjectInfoProc;
       rms->size = sizeof(RapMarshallStruct);

       mercator->GetRAPListener()->WriteProc(rms);
     }
   }

void RAPListener::ObjectInfoProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntObjectInfoCallbackStruct 	*callback_data)
{
  mString reason;

#ifdef RAP_MODULES
  RapBase *instance = NULL;
    
  if ((instance = RapBase::instances.lookup(callback_data->client->ice_conn)) == 0) {
    reason = "Error - no RAP instance";
    Error(EL_DEBUG, "RAPListener::ObjectInfoProc",
	  "No client is associated with this ice connection.  ",
	  "What do we do?", NULL);
  }
#else
  RAP *instance = NULL;
    
  if ((instance = RAP::instances.lookup(callback_data->client->ice_conn)) == 0) {
    reason = "Error - no RAP instance";
    Error(EL_DEBUG, "RAPListener::IOErrorHandler",
	  "No client is associated with this ice connection.  ",
	  "What do we do?", NULL);
  }
#endif

    switch(callback_data->reason) {
    case RapQueryTreeReply: {
      reason = "RapQueryTreeReply";

      // Using ProcessFullQueryTreeReply() for both cases
      instance->ProcessFullQueryTreeReply((RapAgntShellList*)callback_data->data);
      break;
    }
    case RapFullQueryTreeReply: {
      reason = "RapFullQueryTreeReply";
      instance->ProcessFullQueryTreeReply((RapAgntShellList*)callback_data->data);
      break;
    }
    case RapGetGCValuesReply: {
      reason = "RapGetGCValuesReply";
      instance->ProcessGetGCValuesReply((RapGCValuesList *)callback_data->data);

      break;
    }
    case RapGetResourcesReply: {
      reason = "RapGetResourcesReply";
      instance->ProcessGetResourcesReply((RapAgntResourceList *)callback_data->data);

      break;
    }
    case RapGetValuesReply: {
      reason = "RapGetValuesReply";
      instance->ProcessGetValuesReply((RapValuePtr)callback_data->data);

      break;
    }
    case RapLocateObjectReply: {
      reason = "RapLocateObjectReply";
#ifndef USONIX
      instance->ProcessLocateObjectReply((LocateObjectReply*)callback_data->data);
#endif

      break;
    }
    default: {
      reason = "unknown";

    }
    }

    // Allocated in ObjectInfoProcMarshaller
    delete callback_data;

	Error(EL_DEBUG, "RAPListener::ObjectInfoProc",
	      "Reason = ", (char *)reason, NULL);
}

void RAPListener::EventProcMarshaller(Widget 			w, 
				      XtPointer			client_data, 
				      RapAgntEventCallbackStruct 	*callback_data)
{
  if(!mercator->GetRAPListener()->NeedsCleanupP()) {
    RapMarshallStruct *rms = new RapMarshallStruct;
    RapAgntEventCallbackStruct *callbackStruct = new RapAgntEventCallbackStruct;
    xEvent *xev = NULL;

    *callbackStruct = *callback_data;

    switch(callbackStruct->reason) {
    case RapAgntEventNotify:
      xev = new xEvent;

      *xev = *(xEvent*)(callbackStruct->data);
      callbackStruct->data = xev;
      break;
    default:
      break;
    }

    rms->raw = w;
    rms->client_data = (XtPointer)client_data;
    rms->callbackStruct = (XtPointer)callbackStruct;
    rms->reason = rap_EventProc;
    rms->size = sizeof(RapMarshallStruct);

    mercator->GetRAPListener()->WriteProc(rms);
  }
}

void RAPListener::EventProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntEventCallbackStruct 	*callback_data)
{
  mString reason;


#ifdef RAP_MODULES
  RapBase *instance = NULL;
    
  if ((instance = RapBase::instances.lookup(callback_data->client->ice_conn)) == 0) {
    Error(EL_ERROR, "RAPListener::EventProc",
	  "No client is associated with this ice connection.  ",
	  "What do we do?", NULL);
  }
#else
  RAP *instance = NULL;
    
  if ((instance = RAP::instances.lookup(callback_data->client->ice_conn)) == 0) {
    Error(EL_DEBUG, "RAPListener::EventProc",
	  "instanceDict lookup failed", NULL);

    Client *client = RAP::clients.lookup(callback_data->client->window);
    if(client != NULL) 
      instance = client->GetRAP();
    else
      Error(EL_DEBUG, "RAPListener::EventProc",
	    "WinClientDict lookup failed, too", NULL);

  }

  if(instance == NULL) {
    Error(EL_ERROR, "RAPListener::EventProc",
	  "No client is associated with this ice connection.  ",
	  "What do we do?", NULL);
    return;
  }
#endif

  switch(callback_data->reason) {
  case RapAgntGetEventsReply: {
    reason = "RapAgntGetEventsReply";
    //    instance->ProcessGetEventsReply((GetEventsReply*)callback_data->data);
    break;
  }
  case RapAgntEventNotify: {
    xEvent *xevent = (xEvent*)(callback_data->data);
    reason = "RapAgntEventNotify";
            instance->ProcessEventNotify(xevent);

  // Allocated in EventProcMarshaller
	    	    	    delete (xEvent*)(callback_data->data);
    break;
  }
  default: {
    reason = "unknown";
  }
  }

  // Allocated in EventProcMarshaller
  delete callback_data;

  	Error(EL_DEBUG, "RAPListener::EventProc",
  	      "Reason = ", (char*)reason, NULL);
}

void RAPListener::RequestProcMarshaller(Widget 			w, 
					XtPointer			client_data, 
					RapAgntRequestCallbackStruct 	*callback_data)
{
  if(!mercator->GetRAPListener()->NeedsCleanupP()) {
    RapMarshallStruct *rms = new RapMarshallStruct;
    RapAgntRequestCallbackStruct *callbackStruct = new RapAgntRequestCallbackStruct;
    int len = ((xReq*)(callback_data->data))->length * 4;
    char buf[32], buf2[32];
    xReq *xreq = (xReq*)new char[len]; //malloc(len);
    xImageTextReq* xitr = NULL;
    xPolyTextReq *xptr = NULL;

    *callbackStruct = *callback_data;

    switch (((xReq*)(callbackStruct->data))->reqType) {
    case X_ImageText8:
    case X_ImageText16: 
      xitr = (xImageTextReq*)(callbackStruct->data);
      memcpy (xreq, xitr, len);

      /*
      sprintf(buf, "%lx", ((xImageTextReq*)(callbackStruct->data))->drawable);
      sprintf(buf2, "%lx", xitr->drawable);

      Error(EL_DEBUG, "RAPListener::RequestProcMarshaller",
	    "Image-text request - old window = ", buf, ", new window = ", buf2, NULL);
      */
      break;
    case X_PolyText8:
    case X_PolyText16: 
      xptr = (xPolyTextReq*)(callbackStruct->data);
      memcpy (xreq, xptr, len);

      /*
      sprintf(buf, "%lx", ((xPolyTextReq*)(callbackStruct->data))->drawable);
      sprintf(buf2, "%lx", xptr->drawable);

      Error(EL_DEBUG, "RAPListener::RequestProcMarshaller",
	    "Poly-text request - old window = ", buf, ", new window = ", buf2, NULL);
      */
      break;
    default: 
      *xreq = *(xReq*)(callbackStruct->data);
      break;
    }

    callbackStruct->data = xreq;

    rms->raw = w;
    rms->client_data = (XtPointer)client_data;
    rms->callbackStruct = (XtPointer)callbackStruct;
    rms->reason = rap_RequestProc;
    rms->size = sizeof(RapMarshallStruct);

    mercator->GetRAPListener()->WriteProc(rms);
  }
}

void RAPListener::RequestProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntRequestCallbackStruct 	*callback_data)
{
#ifdef RAP_MODULES
  RapBase *instance = NULL;
    
  if ((instance = RapBase::instances.lookup(callback_data->client->ice_conn)) == 0) {
    Error(EL_ERROR, "RAPListener::RequestProc",
	  "No client is associated with this ice connection.  ",
	  "What do we do?", NULL);
  }
#else
  RAP *instance = NULL;
    
  if ((instance = RAP::instances.lookup(callback_data->client->ice_conn)) == 0) {
    Error(EL_ERROR, "RAPListener::IOErrorHandler",
	  "No client is associated with this ice connection.  ",
	  "What do we do?", NULL);
  }
#endif

  switch(callback_data->reason) {
  case RapAgntGetRequestsReply: {
    //    instance->ProcessGetRequestsReply((GetRequestsReply*)callback_data->data);
    break;
  }
  case RapAgntRequestNotify: {
    xReq *xreq = (xReq *) (callback_data->data);
            instance->ProcessRequestNotify(xreq);

	    // Allocated in RequestProcMarshaller
	    	    delete[] (char*)(callback_data->data);
    break;
  }
  default: {
  }
  }

  // Allocated in RequestProcMarshaller
  delete callback_data;
} // RequestProc

#endif

void
RAPListener::SetupAgent(XServer& xserver)
{
#ifndef USONIX
    XTextProperty	textProp;

    textProp.value = (unsigned char *) networkIds;
    textProp.encoding = XA_STRING;
    textProp.format = 8;
    textProp.nitems = strlen(networkIds);

    xserver.SetTextProperty(agentWindow, &textProp,
			    xserver.InternAtom(ICE_NETWORK_IDS, FALSE));
#endif
}

int
RAPListener::Register(XServer& xserver)
{
#ifdef USONIX
    agentWindow = XtWindow(::top_level);

#else
    Status status;
    char errorStr[256];
    register int i;
    
    majorOpcode = IceRegisterForProtocolReply(RapName,
					      "X Consortium",
					      "1.0",
					      1,
					      protoVersions,
					      0,
					      NULL,
					      NULL,
					      RAPListener::AuthorizeProc,
					      RAPListener::ProtoSetupProc,
					      RAPListener::ProtoActivateProc,
					      RAPListener::IOErrorProc);

    if (majorOpcode == -1) {
	Error(EL_FATAL, "RAPListener::Register",
	      "IceRegisterForProtocolReply failed!  Cannot continue.", NULL);
	return 0;
    }

    status = IceListenForConnections(&listenerCount,
				     &listeners,
				     256,
				     errorStr);

    if (!status) {
	Error(EL_FATAL, "RAPListener::Register",
	      "IceListenForConnections failed: ", errorStr, NULL);
	return 0;
    }

    RAPListener::networkIds = IceComposeNetworkIdList(listenerCount, listeners);

    RAPListener::agentWindow = xserver.CreateWindow(xserver.GetRootWindow(),
					    0, 0, 1, 1, 1,
					    xserver.GetBlackPixel(),
					    xserver.GetBlackPixel());

    if (agentWindow == None) {
	Error(EL_FATAL, "RAPListener::Register",
	      "Could not create agent window.", NULL);
	return 0;
    }

    SetupAgent(xserver);
    protoNameAtom = xserver.InternAtom(RapName, FALSE);

    for (i=0 ; i<listenerCount ; i++) {
	IceSetHostBasedAuthProc(listeners[i], RAPListener::AuthorizeProc);
	
	AddFDInterest(IceGetListenConnectionNumber(listeners[i]));
	SetIOMask(IceGetListenConnectionNumber(listeners[i]), ARead);
    }
#endif
    
    return 1;
}

int
RAPListener::InitiateConnection(XServer& xserver, Window clientWindow)
{
    XClientMessageEvent cme;
    char tmp[16], tmpHex[16];
    
    sprintf(tmp, "%ld", clientWindow);
    sprintf(tmpHex, "%lx", clientWindow);

    //
    // We used to send an event to ANY window which is not the application
    // shell window for an already-seen client.  Instead, we now send
    // an event to any window which is mapped on the root window and
    // is not in our model.
    //
    if (mercator->GetAppModelMgr()->GetXtObject(clientWindow)) 
	return FALSE;
#if 0
    //
    // Return if this window ID has already been seen before.
    //
    if (RAP::clients.lookup(clientWindow) != nil) {
	Error(EL_DEBUG, "RAPListener::InitiateConnection",
	      "Already seen window ", tmp, "(", tmpHex,
	      "), ignoring...", NULL);
	return FALSE;
    }
#endif
    Error(EL_DEBUG, "RAPListener::InitiateConnection",
	  "Initiating connection to client window ", tmp, "(",
	  tmpHex, ")...", NULL);
    
#ifdef USONIX
#ifdef RAP_MODULES
    if (RapBase::clients.lookup(clientWindow) != NULL) {
#else
    if (RAP::clients.lookup(clientWindow) != NULL) {
#endif
	Error(EL_DEBUG, "RAPListener::InitiateConnection",
	      "Already seen window ", tmp, "(", tmpHex,
	      "), ignoring...", NULL);
	return FALSE;
    }
        RapAgntContactExistingClients(::rap_agnt);

    xserver.Flush();
	return TRUE;
#else
    cme.window = clientWindow;
    cme.type = ClientMessage;
    cme.message_type = xserver.InternAtom(EXTAGENT_NAME, FALSE);
    cme.format = EXTAGENT_FORMAT;
    cme.data.l[0] = protoNameAtom;
    cme.data.l[1] = xserver.InternAtom(ICE_NETWORK_IDS, FALSE);
    cme.data.l[2] = agentWindow;
    cme.data.l[3] = 0L;
    cme.data.l[4] = 0L;

    if (!xserver.SendEvent(clientWindow, TRUE, 0, (XEvent *) &cme))
	Error(EL_ERROR, "RAPListener::InitiateConnection",
	      "SendEvent failed!", NULL);

    xserver.Flush();

    return TRUE;
#endif
}



////////////////////////////////////////////////////////////////////////
//
// Member functions.
//
////////////////////////////////////////////////////////////////////////

int
RAPListener::HandleActivity(int _fd, short mask)
{
    // we probably don't have to do this... let ICE do it for us...
    if (mask & ARead) {
#ifdef USONIX
	if(_fd == pipeFDs[rap_xt][0]) {
	  RapMarshallStruct *rms;
	  char buf[256];
	  CARD32 seq = 0;
	  int numChars = 0, index=0;
	  CARD32 sequenceList[marshallDict->size()];
	  mDictValPtrIterator<CARD32, RapMarshallStruct> iter(*marshallDict);

	  // Clear out the read buffer
	if (ioctl(_fd, FIONREAD, &numChars) == -1) {
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "ioctl failed: ", Perror(), NULL);
	    return 0;
	}

	while(numChars > 0) {
		read(_fd, buf, (numChars <= 256) ? numChars : 256);
		numChars -= 256;
	    }

	  for(iter.reset(); iter.isValid(); iter.next()) {
	    sequenceList[index++] = iter.current();
	    printf("RAPListener::HandleActivity - servicing marshallSequenceNumber %u\n",
		   sequenceList[index-1]);
	  //	  if(numChars > 0) {
	  //	    seq = atol(buf);

	  //	    printf("RAPListener::HandleActivity - string read is %s\n", buf);
	  //	    printf("RAPListener::HandleActivity - marshallSequenceNumber is %u\n", seq);
	    rms = marshallDict->lookup(iter.current());

	    if(rms == NULL) {
	      Error(EL_ERROR, "RAPListener::HandleActivity",
		    "RapMarshallStruct lookup failed", NULL);
	      return 0;
	    }

	    switch(rms->reason) {
	    case rap_ClientProc:
	      ClientProc(rms->raw, rms->client_data,
				      (RapAgntClientCallbackStruct*)rms->callbackStruct);

	      	      delete rms;
	      break;
	    case rap_ObjectInfoProc:
	      ObjectInfoProc(rms->raw, rms->client_data,
					  (RapAgntObjectInfoCallbackStruct*)rms->callbackStruct);

	      	      delete rms;
	      break;
	    case rap_HookProc:
	      HookProc(rms->raw, rms->client_data,
				    (RapAgntHookCallbackStruct*)rms->callbackStruct);

	      	      delete rms;
	      break;
	    case rap_EventProc:
	      EventProc(rms->raw, rms->client_data,
				     (RapAgntEventCallbackStruct*)rms->callbackStruct);

	      	      delete rms;
	      break;
	    case rap_RequestProc:
	      RequestProc(rms->raw, rms->client_data,
				       (RapAgntRequestCallbackStruct*)rms->callbackStruct);

	      	      delete rms;
	      break;
	    default:
	      Error(EL_WARNING, "RAPListener::HandleActivity",
		    "reason is unsupported", NULL);
	    }

	  } // iterator loop

	  // We've responded to all in the dict, so empty it
	  for(int i=0; i<index; i++)
	    marshallDict->del(sequenceList[i]);

	  // Check
	  if(marshallDict->size() > 0) {
	    char buf[32];
	    sprintf(buf, "%d", marshallDict->size());
	    Error (EL_WARNING, "RAPListener::HandleActivity",
		   "marshallDict should be empty, but contains ", buf, " items.", NULL);
	  }

	  return 1;
	} else {
	  Error(EL_DEBUG, "RAPListener::HandleActivity",
		"Called with the wrong descriptor?", NULL);

	  return 0;
	}

#else
	IceConn newIceConn;
	IceListenObj l;
	IceAcceptStatus statusRet;
	register int i;
	RAP *rap;
	
	//
	// Find the listener object which corresponds to this descriptor.
	// This is cheezy, but (1) we probably won't have too many
	// listeners, and (2) connection setup doesn't happen too often.
	//
	for (i=0 ; i<listenerCount ; i++) {
	    if (_fd == IceGetListenConnectionNumber(listeners[i])) {
		l = listeners[i];
	    }
	}

	Error(EL_DEBUG, "RAPListener::HandleActivity",
	      "connection request on addr ",
	      IceGetListenConnectionString(l), NULL);

	newIceConn = IceAcceptConnection(l, &statusRet);

	if (statusRet != IceAcceptSuccess) {
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "Could not accept ICE connection.", NULL);
	    return 0;
	}

	//
	// At this point we've been contacted by a RAP-speaking client and
	// we're engaged in the startup negotiation process.  We create a
	// new RAP instance to process the connection; it'll finish the
	// connection process and continue on to hanlding actual RAP
	// messages.  We need to padd it the IceConn and major opcode for
	// the protocol.
	//
	rap = new RAP(newIceConn, majorOpcode);

	return 1;
#endif // USONIX
    }
    
    if (mask & AWrite) {
	// IGNORE?

	return 1;
    }
    if (mask & AExcept) {
	//
	// An exceptional condition has occurred on a listener descriptor.
	// In theory these should never happen; if we get an exception,
	// shut down Mercator.
	//
#ifndef USE_X_EVENTS
	if (mask & POLLERR) 
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "Exceptional condition detected: POLLERR\n",
		  "\tCannot continue.", NULL);
	if (mask & POLLHUP)
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "Exceptional condition detected: POLLHUP\n",
		  "\tCannot continue.", NULL);
	if (mask & POLLNVAL)
	    Error(EL_ERROR, "RAPListener::HandleActivity",
		  "Exceptional condition detected: POLLNVAL\n",
		  "\tCannot continue.", NULL);
#endif

	this->MarkForCleanup();
	FDInterest::Quit();

	return 0;
    }
}

 void RAPListener::Shutdown(int fd)
{
  mDictValPtrIterator<IceConn, RAP> iter(RAP::instances);

  char buf[32];
  sprintf(buf, "%d", RAP::instances.size());
  Error(EL_DEBUG, "REAPListener::Shutdown",
	"Size of RAP::instances is ", buf, NULL);

  for(iter.reset(); iter.isValid(); iter.next()){
    Error(EL_DEBUG, "RAPListener::Shutdown",
	  "Calling Shutdown for a RAP instance", NULL);
    RAP::instances.lookup(iter.current())->Shutdown(0);
  }

  sprintf(buf, "%d", RAP::instances.size());
  Error(EL_DEBUG, "REAPListener::Shutdown",
	"Size of RAP::instances is ", buf, NULL);

  Error(EL_DEBUG, "RAPListener::Shutdown",
	"Calling kill_all on RAP::instances...", NULL);
  RAP::instances.kill_all(); // calls delete for each

  sprintf(buf, "%d", RAP::instances.size());
  Error(EL_DEBUG, "REAPListener::Shutdown",
	"Size of RAP::instances is ", buf, NULL);

  // We'll let Mercator's destructor delete its RAPListener instance
  //    MarkForCleanup();
}
