// -*-Mode: C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: RAPListener.h,v $
// Revision 0.4  1998/08/03 20:49:50  sellden
// Added initial support for Rap modules, added MarshallDict code
//
// Revision 0.3  1998/06/06 21:04:31  sellden
// *** empty log message ***
//
// Revision 0.2  1997/12/23 22:30:40  sellden
// Large-scale changes... Moved from GaTech's Rap implementation to Will Walker's implementation
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//

#ifndef _RAP_LISTENER_H_
#define _RAP_LISTENER_H_

#include <X11/Xlib.h>
#ifdef USONIX
#define String	XtString
#include <X11/Intrinsic.h>
#include<X11/Shell.h>

extern "C" {
#include <X11/Rap/Rap.h>
#include <X11/Rap/RapProto.h>		// Protocol definition
#include <X11/Rap/RapAgnt.h>		// C language agent API
#include<X11/Xtea/Xtea.h>
}
#undef String
#endif
#include <X11/ICE/ICElib.h>

#include "FDInterest.h"
#include "mDictValPtr.h"

enum rap_toolkit {rap_xt, rap_gtk, rap_qt};
enum rap_callback {rap_ClientProc, rap_ObjectInfoProc,
		   rap_EventProc, rap_RequestProc, rap_HookProc};

typedef struct _RapMarshallStruct {
int reason;
long size;
Widget raw;
XtPointer client_data;
XtPointer callbackStruct;
} RapMarshallStruct;

typedef mDictValPtr<CARD32, RapMarshallStruct> MarshallDict;

class XServer;


class RAPListener : public FDInterest {
private:
  int pipeFDs[4][2];  // One pair of fds per toolkit
  CARD32 marshallSequenceNumber;
  MarshallDict *marshallDict;

    IcePaVersionRec		*protocolVersions;
    IceListenObj		*listeners;
    int				listenerCount;
    int				majorOpcode;
    char			*networkIds;
    Window			agentWindow;
    Atom			protoNameAtom;

    void			SetupAgent(XServer& xserver);

    static Bool			AuthorizeProc(char *hostname);
    static void			IOErrorHandler(IceConn iceConn);
    static void			IOErrorProc(IceConn iceConn);
    static ProcessProtoMsgProc;
    static Status		ProtoSetupProc(IceConn iceConn,
					       int majorVersion,
					       int minorVersion,
					       char *vendorString,
					       char *releaseString,
					       IcePointer *clientDataRet,
					       char **failureReasonRet);
    static void			ProtoActivateProc(IceConn iceConn,
						  IcePointer clientData);
    void			ClassInit();

public:
    RAPListener(XServer& xserver);
    virtual ~RAPListener();

    //
    // HandleActivity can spawn new RAP objects when a connection is
    // made.
    //
    virtual int			HandleActivity(int fd, short mask);
    //
    // Shutdown() is called when zero-length reads are detected.  For
    // listener descriptors, this probably means that someone is trying
    // to connect() to us.  Don't treat it as a connection loss.
    //
#ifdef USONIX
    virtual void		Shutdown(int fd);
#else
    virtual void		Shutdown(int fd)
				    { return; }
#endif

    int				Register(XServer& xserver);

    //
    // InitiateConnection can be called via the XServer whenever the
    // connection initiation conditions are satisfied.  The return
    // value is TRUE if the code actually tries to initiate a connection
    // (the window was previously unseen), and FALSE otherwise.
    //
    int				InitiateConnection(XServer& xserver,
					   Window clientWindow);
#ifdef USONIX
  // Callbacks handed to main window of Ultrasonix
  static Widget GetProc();
  void HookProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntHookCallbackStruct 	*callback_data);
  static void HookProcMarshaller(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntHookCallbackStruct 	*callback_data);
  void EventProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntEventCallbackStruct 	*callback_data);
  static void EventProcMarshaller(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntEventCallbackStruct 	*callback_data);
  void RequestProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntRequestCallbackStruct 	*callback_data);
  static void RequestProcMarshaller(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntRequestCallbackStruct 	*callback_data);
  void ClientProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntClientCallbackStruct 	*callback_data);
  static void ClientProcMarshaller(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntClientCallbackStruct 	*callback_data);
  void ObjectInfoProc(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntObjectInfoCallbackStruct 	*callback_data);
  static void ObjectInfoProcMarshaller(Widget 			w, 
	   XtPointer			client_data, 
	   RapAgntObjectInfoCallbackStruct 	*callback_data);
void WriteProc(RapMarshallStruct *rms);
#endif
};

#endif /* _RAP_LISTENER_H_ */
