// -*- Mode: C++ -*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: Resource.h,v $
// Revision 0.3  1998/05/21 23:02:59  sellden
// Replaced unchar with unsigned char
//
// Revision 0.2  1997/12/23 22:30:39  sellden
// Added PrintAll() method, added byte-swapping code, likely broken
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//
//
// Management of Xt resources.
//
// This code defines several classes for resource handling:
//    -ResourceCache    Provides storage to a set of resources.
//	                Typcaily one ResourceCache will be
//		        instantiated per XtObject.  This is the only
//		        externally-visible class provided by this file.
//    -Resource		Storage class for resource.  Provides name,
//		        type, class, etc.
//    -ResourceValue	Class for handling variable-format values.
//			Can probably be merged back into Resource.
//    -ResourceTypeManager    Provides a registry system for Xt types,
//			      and handling for multiple type formats.
//
// Possible improvements:
//    - This code probably spends 80% of its time in strcmp().  We might want
//	   to create a QuarkManager to build hash values from strings to save
//	   time on string compares.
//    - A Trigger class could give us a systematic way of effecting change when
//	   resources are updated.
//
// NOTE that everything but ResourceCache is hidden in the .cc file since it's
// an implementation detail (fast compiles, ahhhhh... :-)
//
//		keith edwards	14mar95
//

#ifndef _RESOURCE_H_
#define _RESOURCE_H_

#include <exception.h>
#include <string.h>
#include <sys/types.h>

#include "mString.h"
#include "mDictPtrPtr.h"

#include "Util.h"

class Resource;
class ResourceCache;

typedef Resource *ResourcePtr;
typedef const char *CharPtr;

typedef mDictPtrPtr<mString, Resource> ResourceDict;

extern void InitializeResourceTypeManager();

////////////////////////////////////////////////////////////////////////

class ResourceCache {

    friend ostream& operator<<(ostream&, const ResourceCache&);
    
private:
    ResourceDict		*resources;

public:
    ResourceCache();
    virtual ~ResourceCache();

    int				Exists(const char *name) const;
    int				Initialized(const char *name) const throw(xmsg);
    ostream&			Print(const char *name, ostream& s) const;
    ostream&			PrintValue(const char *name, ostream& s) const;

    int				Add(const char *name,
				    const char *className,
				    const char *type,
				    const char *val = NULL) throw(xmsg);
    int				Set(const char *name,
				    const char *type, 
				    const char *val) throw(xmsg);
    int				Remove(const char *name);

    int				SetValueRaw(const char *n,
					    const char *val,
					    const char *rep = NULL) throw(xmsg);
    
    int				SetValue(const char *name,
					 const char *value) throw(xmsg);
    int				SetValue(const char *name,
					 unsigned char value) throw(xmsg);
    int				SetValue(const char *name,
					 ushort value) throw(xmsg);
    int				SetValue(const char *name,
					 ulong value) throw(xmsg);
    
    int				GetValue(const char *name,
					 char **value,
					 const char **t=NULL) const throw(xmsg);
    int				GetValue(const char *name,
					 unsigned char *value,
					 const char **t=NULL) const throw(xmsg);
    int				GetValue(const char *name,
					 ushort *value,
					 const char **t=NULL) const throw(xmsg);
    int				GetValue(const char *name,
					 ulong *value,
					 const char **t=NULL) const throw(xmsg);
#ifdef USONIX
void printAll();
#endif
};


////////////////////////////////////////////////////////////////////////


#endif /* _RESOURCE_H_ */
