//  -*-Mode: C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: TclAction.cc,v $
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//
//
// $Author: rcsg $
// $Source: /usr/mercator/newsource/7.0/RCS/TclAction.cc,v $
// $Revision: 0.1 $
// $Date: 1996/12/02 20:06:39 $
//
 
static const char rcsid[]="@(#) $Header: /usr/mercator/newsource/7.0/RCS/TclAction.cc,v 0.1 1996/12/02 20:06:39 rcsg Exp $";


#include "TclAction.h"
#include "Error.h"
#include "Util.h"
#include <stdlib.h>
#include <strings.h>

// ************************************************************************
// **                     Class Special Functions                    ******
// ************************************************************************

TclAction::TclAction(Tcl_Interp *interp, const char *n) :
    Action(n),
    interpreter(interp)
{
}

// ************************************************************************

TclAction::~TclAction(void)
{
}

// ************************************************************************
// **                          Public Functions                      ******
// ************************************************************************

int
TclAction::Invoke(int argc, char *argv[], const char **resString)
{
    char *str;
    char *command;
    int result;
    int stringUsed = 0;

    if (argc == 0) {
	//NO args.
	// For some reason, Sentinel reports that Tcl_Eval is freeing name.
	// We'll dup a new one to pass in.
	char *tmpName = strdup(name);
	result = Tcl_Eval(interpreter, tmpName);
	free(tmpName);
    } else {
	str = Tcl_Concat(argc, argv);
	command = new char[strlen(str) + strlen(name) + 2];
	strcpy(command, name);
	strcat(command, " ");
	strcat(command, str);
	result = Tcl_Eval(interpreter, command);
	// memory allocated via malloc(), release via free().
	stringUsed = 1;
    }

    if (resString)
	*resString = interpreter->result;
    
    if (result != TCL_OK) {
	if (stringUsed)
	    Error(EL_ERROR, "TclAction::Invoke", "error in action ",
		  command, ": ", interpreter->result, NULL);
	else
	    Error(EL_ERROR, "TclAction::Invoke", "error in action ",
		  name, " ", ": ", interpreter->result, NULL);
	
	return 0;
    }

    if (stringUsed) {
	free(str);
	delete [] command;
    }
    
//    if (strcmp("False",interpreter->result) == 0)
//	return(!TCL_OK);
    
//    return(result);
    return 1;
}

// ************************************************************************
