//  -*-Mode:C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: XtObject.cc,v $
// Revision 0.7  1998/05/21 23:02:14  sellden
// Replaced unchar with unsigned char
//
// Revision 0.6  1997/12/23 22:30:38  sellden
// Added support in CreateResource() to handle string values
//
// Revision 0.5  1997/03/29 00:12:54  sellden
// Standardized dictionary semantics
//
// Revision 0.4  1997/02/16 09:06:26  sellden
// Removed dictionary constructors which take a hashing function
// pointer.  Hashing functionality is now determined by
// overriding the class HashActions.
//
// Revision 0.3  1997/02/08 11:15:19  sellden
// Updated use of mDictValPtrIterators based on
// OTC_Iterator semantics.
//
// Revision 0.2  1996/12/11 18:45:57  barnaby
// Included ScreenReader.h
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//
// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/XtObject.cc,v $
// $Revision: 0.7 $
// $Date: 1998/05/21 23:02:14 $
//

static const char rcsid[] = "$Header: /usr/mercator/newsource/7.0/RCS/XtObject.cc,v 0.7 1998/05/21 23:02:14 sellden Exp $";

#include "XtObject.h"
#include "TclAction.h"
#include "Template.h"
#include "TextRep.h"
#include "Client.h"
#include "AppModelMgr.h"
#include "Error.h"
#include "Util.h"
#include "Resource.h"
#include "RAP.h"
#include "mListVal.h"

#ifdef USONIX	// TB: Need definitions of ScreenReader class
#include "ScreenReader.h"
#endif

//counter used for generating unique XtObject ids.
static unsigned long unique_id_count = 0;


//
// String hasher for dictionaries.
//
#ifdef USONIX
/* static object holding the class dictionary */
mDictValPtr<mString,eventDict> *XtObject::classDictionary=new mDictValPtr<mString,eventDict>();
#else
static unsigned StringHasher(const mString& str) { return str.hash(); }

/* static object holding the class dictionary */
mDictValPtr<mString,eventDict> *XtObject::classDictionary=new mDictValPtr<mString,eventDict>(StringHasher);

// cheezy hash function for Objects
static unsigned objectHashFunc(const Object &o)
{
    return ((unsigned)o);
}
#endif

// ************************************************************************
// **                     Class Special Functions                    ******
// ************************************************************************
XtObject::XtObject(Client *_client, Object _id, const char *_name,
                   const char *_class, XtObject *_parent, 
		   Window _window) :
    id(_id),
    myName(Strdup(_name)),
    myClass(Strdup(_class)),
    longName(new char*),
    uniqueName(NULL),
    parent(NULL),
    window(_window),
    borderWidth(0),
    x(-1),
    y(-1),
    width(-1),
    height(-1),
    mapped(False),
    managed(False),
    resources(new ResourceCache),
#ifdef USONIX
    eventActions(new mDictPtrPtr<EventStruct,Action>()),
#else
    eventActions(new mDictPtrPtr<EventStruct,Action>(EventStructHash)),
#endif
    sreader(NULL),
    textrep(NULL),
    client(_client),
#ifdef USONIX
    children(new mDictValPtr<Object,XtObject>()),
#else
    children(new mDictValPtr<Object,XtObject>(objectHashFunc)),
#endif
    popReturn(NULL),
    grabsInstalled(0)
{ 
    char buffer[64];

    //
    // Set our creation timestamp. (This is no longer a struct timeval!)
    //
    xtObjectIdentifier=unique_id_count;

    // Note use of double indirection so we can keep constness in 
    // GetLongName()			-k
    *longName = NULL;

    sprintf(buffer, "XtObject%d", unique_id_count++);
    uniqueName = Strdup(buffer);
    
    SetParent(_parent);

    // Fire the creation template action
    (void) Template::GetInteger(*this, "creation");
}
    
// ************************************************************************
    
XtObject::~XtObject()
{
    // Fire the template deletion action
    (void) Template::GetInteger(*this, "destruction");
    
    delete [] myName;
    delete [] myClass;
    delete [] uniqueName;
    delete [] (*longName);
    delete longName;
    
    // Tell our parent we're not longer its child, being as we no longer
    // exist... -k
    if (parent && parent->children)
	parent->children->del(id);
    
    //
    // Delete our children now.  NOTE THE WEIRD WAY OF DOING THIS.
    // Everytime we call AMM::DeleteXtObject(), our dictionary of
    // children will be reset as the children are deleted.  So we
    // need to reinitialize the iterator each time.
    //
    // What I'd *like* to do is simply set parent = NULL on all the
    // children before we delete them, to prevent the children from
    // removing themselves from their parent's (that's us) dictionary.
    // IAN - WHY DOESN'T YOUR ITERATOR RETURN THE ACTUAL DATA IN THE DICT,
    // RATHER THAN A COPY!?
    //
    int done = 0;
    if (children) {
      while (!done) {
	mDictValPtrIterator<Object, XtObject> iter(*children);
      
#ifdef USONIX
	iter.reset();
	if (iter.isValid()) {
#else
	if (iter.next()) {
#endif
	  XtObject *obj = children->lookup(iter.current());
	  mercator->GetAppModelMgr()->DeleteXtObject(obj);
	} else {
	  done = 1;
	}
      }
    }
    
    delete resources;
    delete eventActions;
    delete children;

    delete sreader;
    delete textrep;
}

// ************************************************************************
// **                          Public Functions                      ******
// ************************************************************************

mDictValPtr<Object,XtObject> *XtObject::GetChildren() const
{
    return children;
}

// ************************************************************************
#if 0
int
XtObject::X() const
{
    if (IsGadget()) {
	// This isn't really gadget specific.  We get the resource value for
	// "x".  If it's in the list of unsafeResources, GetResourceValues will
	// call out to the app to get the value.
	return GetResourceValueShort("x");
    } else {
	return x;
    }
}

int
XtObject::Y() const
{
    if (IsGadget()) {
	return GetResourceValueShort("y");
    } else {
	return y;
    }
}
#endif
// ************************************************************************
//
// These are not inlined so that we don't have to include RapProto.h in the
// .h file.
//
int
XtObject::IsRealized() const
{
    return window != RapUnrealized;
}

int
XtObject::IsGadget() const
{
    return window == RapGadget;
}

// ************************************************************************

void
XtObject::SetName(const char *n)
{
    delete [] myName;
    myName = Strdup(n);
}

// ************************************************************************

void
XtObject::SetClass(const char *c)
{
    delete [] myClass;
    myClass = Strdup(c);
}

// ************************************************************************

const char *
XtObject::GetLongName() const
{
    if (longName == NULL) {
	Error(EL_ABORT, "XtObject::GetLongName",
	      "longname for ", GetName(), " should never be NULL!\n"
	      "\tThis is a nonrecoverable error because it represents an ",
	      "internal consistency\n\tfailure.", NULL);
	return NULL;
    }

    if (*longName == NULL) {
	// We'll have to build it this first time.
	if (parent) {
	    const char *parentLongName = parent->GetLongName();
	    *longName = new char[strlen(parentLongName) + strlen(myName) + 2];
	    strcpy(*longName, parentLongName);
	    strcat(*longName, ".");
	    strcat(*longName, myName);
	} else {
	    *longName = Strdup(myName);
	}
	    
    }

    return *longName;
}

// ************************************************************************

void
XtObject::SetParent(XtObject *parent_ptr)
{
    this->parent = parent_ptr;
    if (parent_ptr)
	parent_ptr->AddChild(this);
}

// ************************************************************************

Window
XtObject::GetValidWindowId() const
{
    if (IsGadget()) {
	//
	// If we're a gadget, we return our ancestor's window.
	//
	XtObject *tmp = GetParent();

	while (tmp && tmp->IsGadget())
	    tmp = tmp->GetParent();

	if (tmp == NULL) {
	    Error(EL_ERROR, "XtObject::GetWindowId",
		  "Gadget has no parent?  ", GetName(), NULL);
	    return window;
	} else {
	    return tmp->GetWindowId();
	}
    } else {
	return window;
    }
}

// ************************************************************************

void
XtObject::SetMapped(int m)
{
    int oldmapped = mapped;
    
    mapped = m;
    
    // If we're mapped, then we MUST be managed.
    if (m) {
	SetManaged(m);
	(void) Template::GetInteger(*this, "mapped");
    } else {
	(void) Template::GetInteger(*this, "unmapped");
    }

    // If we're  just becoming mapped, and we've got bogus X, Y, Width, Height
    // values, figure 'em out.
    if (!oldmapped && mapped) {
	if (x == -1 || y == -1 || width == -1 || height == -1) {
	    XWindowAttributes attr;
	    
	    if (mercator->GetXServer()->GetWindowAttributes(*this, &attr)) {
		Width(attr.width);
		Height(attr.height);
		X(attr.x);
		Y(attr.y);
		BorderWidth(attr.border_width);

		if (attr.map_state == IsViewable)
		    mercator->GetXServer()->SendExposeEvent(*this);
	    } else {
		Error(EL_ERROR, "XtObject::SetMapped",
		      "GetWindowAttributes failed for object ",
		      GetName(), NULL);
	    }
	}
    }
}

// ************************************************************************

void
XtObject::SetManaged(int m)
{
    managed = m;
    
    if (m)
	(void) Template::GetInteger(*this, "managed");
    else
	(void) Template::GetInteger(*this, "unmanaged");
}

// ************************************************************************

int
XtObject::GetMapped() const
{
    if (IsGadget()) {
	//
 	// If we're a gadget...
	//	- If we're not managed, we can't be mapped.
	//      - Otherwise, we're mapped if there's a mapped+managed
	//	  pathway to the top of the tree.
	if (!GetManaged()){
	    return False;
	} else {
	    XtObject *tmp = GetParent();

	    // If the gadget has no parent (which should never happen),
	    // it can't be mapped.
	    if (tmp == NULL)
		return False;

	    while (tmp) {
		// If we're on a gadget
		if (tmp->IsGadget()) {
		    if (tmp->GetManaged()) {
			// keep going
			tmp = tmp->GetParent();
			continue;
		    } else {
			// Unmanaged gadget, we can't be mapped.
			return False;
		    }
		} else {
		    // It's not a gadget.
		    if (tmp->GetManaged() && tmp->GetMapped()) {
			// keep going
			tmp = tmp->GetParent();
			continue;
		    } else {
			// Unmanaged or unmapped parent, bail out.
			return False;
		    }
		}
	    }
	    // If we made it to here without bailing, we've hit the top of
	    // the tree without finding an unmapped or unmanaged parent.
	    // Return true.
	    return True;
	}
    } else {
	return mapped;
    }
}

// ************************************************************************

void
XtObject::SetResourceValue(const char *name, const char *value)
{
    if (name == NULL) {
	Error(EL_ERROR, "XtObject::SetResourceValue",
	      "NULL name argument, object = ", GetName(), NULL);
	return;
    }

    resources->SetValue(name, value);
}

void
XtObject::SetResourceValue(const char *name, unsigned char value)
{
    if (name == NULL) {
	Error(EL_ERROR, "XtObject::SetResourceValue",
	      "NULL name argument, object = ", GetName(), NULL);
	return;
    }
    
    resources->SetValue(name, value);
}

void
XtObject::SetResourceValue(const char *name, ushort value)
{
    if (name == NULL) {
	Error(EL_ERROR, "XtObject::SetResourceValue",
	      "NULL name argument, object = ", GetName(), NULL);
	return;
    }
    
    resources->SetValue(name, value);
}

void
XtObject::SetResourceValue(const char *name, ulong value)
{
    if (name == NULL) {
	Error(EL_ERROR, "XtObject::SetResourceValue",
	      "NULL name argument, object = ", GetName(), NULL);
	return;
    }
    
    resources->SetValue(name, value);
}

// ************************************************************************

int
XtObject::ResourceIsUnsafe(const char *n) const
{
    mString name(n);
    int unsafe = 0;
    mString str,target=n;

    //
    // If the resource we're looking for is flagged as "unsafe" then
    // we know that we can't rely on our local cached value... generate
    // protocol to poll the new value.
    //
    const mListVal<mString>& unsafeResources =
	Template::GetStringList(*this, "unsafeResources");
    
    if (unsafeResources.search(target,str))
	return 1;

    return 0;
}

void
XtObject::DoBlockGetValues(const char *name) const
{
    int argc = 1;
    char *argv[1];

    argv[0] = Strdup(name);
    
    Error(EL_DEBUG, "XtObject::GetResourceValue",
	  "Fetching unsafe resource ", name, " from object ", GetName(), NULL);
    //
    // NOTE last arg - this is blocking.
    //
    client->GetRAP()->GetValuesRequest(this->GetObjID(), 
				       (const char **) argv, argc, TRUE);
}

const char *
XtObject::GetResourceValueString(const char *n, const char **rep) const
{
    char *data;
    
    if (n == NULL) {
	Error(EL_ERROR, "XtObject::GetResourceValue",
	      "NULL name argument, object = ", GetName(), NULL);
	return NULL;
    }
    
    if (ResourceIsUnsafe(n))
	DoBlockGetValues(n);
    
    resources->GetValue(n, &data, rep);
    return data;
}

unsigned char
XtObject::GetResourceValueByte(const char *n, const char **rep) const
{
    unsigned char data;
    
    if (n == NULL) {
	Error(EL_ERROR, "XtObject::GetResourceValue",
	      "NULL name argument, object = ", GetName(), NULL);
	return (unsigned char)NULL;
    }
    
    if (ResourceIsUnsafe(n))
	DoBlockGetValues(n);
    
    resources->GetValue(n, &data, rep);
    return data;
}

ushort
XtObject::GetResourceValueShort(const char *n, const char **rep) const
{
    ushort data;
    
    if (n == NULL) {
	Error(EL_ERROR, "XtObject::GetResourceValue",
	      "NULL name argument, object = ", GetName(), NULL);
	return (ushort)NULL;
    }
    
    if (ResourceIsUnsafe(n))
	DoBlockGetValues(n);
    
    resources->GetValue(n, &data, rep);
    return data;
}

ulong
XtObject::GetResourceValueLong(const char *n, const char **rep) const
{
    ulong data;
    
    if (n == NULL) {
	Error(EL_ERROR, "XtObject::GetResourceValue",
	      "NULL name argument, object = ", GetName(), NULL);
	return (ulong)NULL;
    }
    
    if (ResourceIsUnsafe(n))
	DoBlockGetValues(n);
    
    resources->GetValue(n, &data, rep);
    return data;
}

// ************************************************************************

void
XtObject::AddChild(const XtObject *child)
{
    
    if (child == NULL) {
	Error(EL_ERROR, "XtObject::AddChild", "NULL object passed.", NULL);
	return;
    }
    
    if (children->lookup(child->id) == 0) {
	children->insert(child->id, (XtObject *)child);
    } else {
	Error(EL_ERROR, "XtObject::AddChild", "child ", child->GetName(),
	      " already exists in ", GetName(), NULL);
    }
}

// ************************************************************************

void
XtObject::DeleteChild(const XtObject *child)
{
    if (child == NULL) {
	Error(EL_ERROR, "XtObject::DeleteChild", "NULL object passed.", NULL);
	return;
    }
    
    if ((children == NULL) || (children->lookup(child->id) == 0)){
	Error(EL_ERROR, "XtObject::DeleteChild",
	      child->GetName(), " is not a child of ", GetName(), NULL);
	return;
    }
  
    children->del(child->id);
}

// ************************************************************************

void
XtObject::PrintInfo() const
{
    Error(EL_DEBUG, "XtObject::PrintInfo",
	  uniqueName, " = ", myName, " Class = ", myClass, NULL);
}

// ************************************************************************

//
// CheckIsShell reports whether the object *appears* to be a shell.  The
// old mechanism was just to check for parent == NULL.  This is probably
// not good since apps may randomly create non-parented widgets for
// whatever bizarre reasons they may have.
//
// The test we use is:
//    - If parent is NULL then the object is a shell.
//    - If the isShell TCL action reports true for the object's class, then
//      it is a shell.
//
// We should probably ditch the parent == NULL test in favor of checking WM
// properties.
//
// CHANGE - 11jul95
// The Parent thing is really screwing us with RAP.  So we're going to
// the Shell template variable.  To make it work, set the defaultClassShell
// to TRUE and then explicitly set it to FALSE in every non-shell template.
// (hack)
//
int
XtObject::CheckIsShell() const
{
#if 0
    if (GetParent() == NULL) {
	Error(EL_STATUS, "XtObject::CheckIsShell",
	      "Object is a toplevel because it has no parent: ",
	      GetName(), NULL);
	return 1;
    } else {
    if (GetParent() == NULL) {
#endif
	if (Template::GetBoolean(*this, "shell")) {
	    Error(EL_STATUS, "XtObject::CheckIsShell",
		  "Object is a toplevel because isShell reports true: ",
		  GetName(), NULL);
	    return 1;
	} else {
	    return 0;
	}
#if 0
    }
    }
#endif

}

// ************************************************************************

int
XtObject::DispatchEvent(EventPtr event)
{
    // if an EventBind has been created for the event passed in then
    // invoke the action and return TRUE.  Else return False.
 
    ActionPtr action;
    eventDictPtr dict;
    // look to see if the client is interested in the event
    action = eventActions->lookup(event);

    // if no, see if its class widget is interested
    if (action == NULL) {
	dict = XtObject::classDictionary->lookup(myClass); 
	
	if (dict == NULL) {
            //event binding not found - no action invoked - return false
	    return False; 
	}
	action = dict->lookup(event); 
	
	if (action == NULL) {
	    //event binding not found - no action invoked - return false
	    return False; 
	}
    }
 
    // we've got the action now
    action->Invoke(0,NULL);
 
    return True; // event binding found and invoked - return true
}

// ************************************************************************

void
XtObject::CreateEventBind(EventPtr event, Action *function)
{
    eventActions->insert(event, function);
}

// ************************************************************************

void
XtObject::CreateEventBind(const char *widgetClass, EventPtr event,
			  Action *function)
{ 
    eventDictPtr dict;
 
    dict = XtObject::classDictionary->lookup(widgetClass);
    if (dict == NULL) {
#ifdef USONIX
	dict = new mDictPtrPtr<EventStruct,Action>();
#else
	dict = new mDictPtrPtr<EventStruct,Action>(EventStructHash);
#endif
	dict->insert(event,function);
	XtObject::classDictionary->insert(widgetClass,dict);
    } else { 
	dict->insert(event,function);
    }
}

// ************************************************************************

int
XtObject::DeleteEventBind(EventPtr event)
{
    if (eventActions->lookup(event) == 0) {
	Error(EL_ERROR, "XtObject::DeleteEventBind",
	      "event was never inserted.", NULL);
	return False;
    } else {
	eventActions->del(event);
	return True;
    }
}

// ************************************************************************

int
XtObject::DeleteEventBind(const char *widgetClass, EventPtr event)
{ 
    eventDictPtr dict;
  
    dict = XtObject::classDictionary->lookup(widgetClass);
    if (dict == NULL) {
	Error(EL_DEBUG, "XtObject::DeleteEventBind", "Class ", widgetClass,
	      " not found.", NULL);
	return False;
    }

    if (dict->lookup(event) == 0) {
	Error(EL_DEBUG, "XtObject::DeleteEventBind",
	      "event was never associated with class ", widgetClass, NULL);
	return False;
    }
    dict->del(event);
    return True;
}

// ************************************************************************

mDictPtrPtr<EventStruct,Action> * 
XtObject::GetActions(const char *widgetClass)
{ 
    eventDictPtr dict;

    dict = XtObject::classDictionary->lookup(widgetClass); 
    if (dict == NULL)
	return NULL;

    return (dict);
}

// ************************************************************************

void
XtObject::CreateResource(const char *n, const char *c, const char *t,
			 const char *v)
{
#ifdef USONIX
  // Account for values coming in as strings
  RapValue *rapVal = (RapValuePtr) v;
  long *dLong = (long*)NULL;
  unsigned long uTmp;
  bool *dBool = (bool*)NULL;
  char *dString = (char*)NULL;

      if (strncmp(t + strlen(t) - 6, rapVal->type, 6)) { 
    // RapValue type does not match RapAgntResource type
    if (!strcmp(t+ strlen(t) - 9, "Dimension")) {
      // Matches Dimension, HorizontalDimension, VerticalDimension
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    } else if (!strcmp(t + strlen(t) - 8, "Position")) {
      // Matches Position, HorizontalPosition, VerticalPosition
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strncmp(t, "Boolean", 4)) {
      // Matches Boolean and Bool
      dBool = new bool;
      *dBool = (bool)atol((char*)rapVal->value);
      v = (char*)dBool;
    }else if (!strcmp(t, "Screen")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t + strlen(t) -3, "Int")) {
      // Matches Int, HorizontalInt, VerticalInt
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "Colormap")) {
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "Pixel")) {
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "Pixmap")) {
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "TranslationTable")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "AcceleratorTable")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "WidgetList")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "Cardinal")) {
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "Short")) {
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "Function")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "Callback")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "Visual")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "Atom")) {
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "Gravity")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "InitialState")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "Bitmap")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "Window")) {
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "Widget")) {
      dLong = new long;
      *dLong = atol((char*)rapVal->value);
      v = (char*)dLong;
    }else if (!strcmp(t, "StringArray")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "SmcConn")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "CommandArgArray")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "EnvironmentArray")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
      //    }else if (!strcmp(t, "DirectoryString")) {

    }else if (!strcmp(t, "RestartStyle")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else if (!strcmp(t, "Pointer")) {
      // As yet unhandled.
      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
    }else{
      // Unknown type
      Error(EL_WARNING, "XtObject::CreateResource",
	    "Unhandled type: ", t, NULL);

      // Use the string representation as the value
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
      v = (char*)dString;
  }
  }else{
    // RapValue type matches RapAgntResource type
      dString = new char[strlen((char*)rapVal->value) +1];
      strcpy(dString, (char*)rapVal->value);
    v = (char*)dString;
  }
#endif
    try {
        resources->Add(n, c, t, v);
    }
    catch (xmsg& message) {
	if (strcmp(message.why(), "ResourceAlreadyExists") == 0) 
	    resources->Set(n, t, v);
	else
	    Error(EL_ERROR, "XtObject::CreateResource",
		  "Problem setting resource ", n, ": ",
	  	  message.why(), NULL);
    }
#ifdef USONIX
      delete dLong;
      delete dBool;
#endif
}

void
XtObject::SetResource(const char *n, const char *t, const char *v)
{
    resources->Set(n, t, v);
}

int
XtObject::ResourceExists(const char *name) const
{
    return resources->Exists(name);
}

void
XtObject::PrintResource(const char *name, ostream& s) const
{
    resources->Print(name, s);
}

void
XtObject::PrintResourceValue(const char *name, ostream& s) const
{
    resources->PrintValue(name, s);
}

void
XtObject::PrintResources(ostream& s) const
{
    s << *resources;
}

//
// operator == and operator< implemented so these can be 
// stored in data structures.  
//
// NOTE: We determine these based on their unique identifier. Not only
// the dictionary code needs this but the ::modelManager code in 
// Interp.cc needs the operator< so it can do sorting.
//
int XtObject::operator==(const XtObject& someObj) const
{
    return xtObjectIdentifier==someObj.xtObjectIdentifier?1:0;
}
int XtObject::operator<(const XtObject& someObj) const
{
    return xtObjectIdentifier<someObj.xtObjectIdentifier?1:0;
}

#ifdef USONIX
void
XtObject::UpdateCoords()
{
  Resource *res;
  resources->printAll();

  if(ResourceExists("x") == 0)
    X(GetResourceValueShort("x"));

  if(ResourceExists("y") == 0)
    Y(GetResourceValueShort("y"));

  if(ResourceExists("height") == 0)
    Height(GetResourceValueShort("height"));

  if(ResourceExists("width") == 0)
    Width(GetResourceValueShort("width"));

  if(ResourceExists("borderWidth") == 0)
    BorderWidth(GetResourceValueShort("borderWidth"));

}
#endif

// ************************************************************************

