// This may look like C code, but it is really -*- C++ -*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#ifndef MERCATOR_DICT_PTR_PTR_H
#define MERCATOR_DICT_PTR_PTR_H

// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/mDictPtrPtr.h,v $
// $Revision: 0.6 $
// $Date: 1998/05/21 23:02:08 $
//
// $Log: mDictPtrPtr.h,v $
// Revision 0.6  1998/05/21 23:02:08  sellden
// Added forward decl for iterator template
//
// Revision 0.5  1997/02/16 09:06:26  sellden
// Removed dictionary constructors which take a hashing function
// pointer.  Hashing functionality is now determined by
// overriding the class HashActions.
//
// Revision 0.4  1997/02/10 05:28:09  sellden
// Completed transfer to OTC_Map and OTC_Iterator.
//
// Revision 0.3  1996/12/12 15:59:22  zimba
// ose implementation of mDictPtrPtr
//
//Revision 1.7  1995/07/28  21:21:05  spine
//Added copyright notice.
//
//Revision 1.6  1995/07/25  22:05:34  keith
//Cleaned up some usage of strings, dicts, and lists.  Ran with -pta to
//fix some previously unseen bugs in the parameterized classes.
//
//Revision 1.5  1995/07/24  17:44:26  beth
//Ian's Version
//
//Revision 1.3  1995/06/05  23:22:58  suelong
//*** empty log message ***
//
//Revision 1.2  1995/06/05  23:11:24  suelong
//commented out syntax error on line 11
//
// * Revision 1.1  1995/06/05  22:24:46  iansmith
// * Initial revision
//  *
//
#include <OTC/collctn/map.hh>

#ifdef USONIX // BLS: EGCS compiler was complaining
template<class K, class V> class mDictPtrPtrIterator;
#endif

// 
// This is the closure object. You should subclass this to get walking
// behavior that you want. If your action function returns 0 the
// walk will terminate. If you need startup or closedown help,
// you can implement either of those two functions.
//
template <class K, class V> class mDictPtrPtrClosure {
public:
    // you can use this to setup your closure... 
    // it will get called with the client data passed to walk
    virtual void startup(void *);
    // you will still get a closedown if you short circuit
    // out of the action procedure
    virtual void closedown();
    // remember to return 1 if you want the walk to continue
    virtual int action(K *, V*);
};

template <class K, class V> inline void mDictPtrPtrClosure<K,V>::startup(void *clientData)
		{ }
template <class K, class V> inline void mDictPtrPtrClosure<K,V>::closedown()
		{ }
template <class K, class V> inline int mDictPtrPtrClosure<K,V>::action(K *, V *)
		{ return 0; }

///////////////////////////////////////////////////////////////////////////
//
// This is a dictionary of pointers to pointers. You cannot put
// values (real objects) in this dictionary.  These use pointer
// copy semantics (shallow copy). You must implement the
// K must have following methods to use this class:
// operator==
// operator<
// There must also be a hashing function for K with the following
// sigature
// unsigned hFun(const K& );
// 
// You are not allowd to have duplicates of keys in this dict.
// You can have duplicate values.
//
template <class K, class V> class mDictPtrPtr {
  friend class mDictPtrPtrIterator<K,V>;

private:
  // The implementation
  OTC_Map<K*,V*> *impl;

public:
  // Default constructur
  mDictPtrPtr(){
    impl=new OTC_Map<K*,V*>();
  };

  /* Hash functions now defined by overriding 
   class HashActions, part of the OTC library
  // constructor
  mDictPtrPtr(unsigned (*hashKey)(const K&)){
    impl=new OTC_Map<K*,V*>();
  };
  */

  // copy constructor
  mDictPtrPtr(const mDictPtrPtr& other){
    impl=new OTC_Map<K*,V*>(*(other.impl));
  };

  // destructor
  virtual ~mDictPtrPtr(){delete impl;};

  // operator=
  mDictPtrPtr<K, V>& operator=(const mDictPtrPtr<K, V>& other);

  //insert... returns 1 if the key was already present, 0 otherwise
  int insert(K* k, V * v){
    int result=0;//assume not in the dictonary
    if(impl->contains(k)){
      result=1;
      del(k);
    }
    impl->add(k,v);

    return result;
  };

  //lookup... returns 0 if the value is not present in the dictionary
  // equality of keys is with the operator== on K
  V *lookup(K *k) const{
    V* v=NULL;
    if(impl->contains(k))
      v=impl->item(k);

    return v ;
  };

  // del  ... delete's an object from the dictionary (but doesn't call
  // EITHER of the destructors... this is a NOOP if the item is 
  // not present
  void del(K *k);

  // del_item ... same as del but calls destructors on both
  void del_item(K *k);

  // walk ... pass a closure and client data
  void walk(mDictPtrPtrClosure<K, V>& closure,
	    void *clientData) const;

  // size
  unsigned 			size() const
{return impl->population();};
};

////////////////////////////////////////////////////////////////////////////

template <class K, class V> class mDictPtrPtrIterator {
private:
  // The implementation
  OTC_Iterator<K*> impl;

  // Disallow default constructor
  mDictPtrPtr(){};

public:
  // constructor
  mDictPtrPtrIterator(const mDictPtrPtr<K, V> &dict)
{impl=dict.impl->keys();};

  // Copy constructor
  mDictPtrPtrIterator(const mDictPtrPtrIterator<K, V>& iter): 
    impl(iter.impl){};

  // destructor
  virtual ~mDictPtrPtrIterator(){};

  // get next ptr... returns 0 when done
  K *next(){
    K* result=NULL;
    if(impl.isValid()){
      //advance to the next item
      impl.next();
      if(impl.isValid())
	result=impl.item();
    }
    return result ;
  };

K* current() const
  {return impl.item(); };

void reset() 
  {impl.reset(); };

int isValid() const
  {return impl.isValid(); };
};

template<class K, class V> void
mDictPtrPtr<K,V>::del(K *k)
{
  if(impl->contains(k))
    impl->remove(k);
}

template<class K, class V> void
mDictPtrPtr<K,V>::del_item(K *k)
{
  if(impl->contains(k)){
    V* v = impl->item(k);
    impl->remove(k);
    delete(k);
    delete(v);
  }
}

template<class K, class V> void
mDictPtrPtr<K,V>::walk(mDictPtrPtrClosure<K, V>& closure,
     void *clientData) const
{
  //issue the startup
  closure.startup(clientData);

    // Loop thru the items
  mDictPtrPtrIterator<K,V> iter(*impl);
  int index, dictSize = size();

  for(index=0; index < dictSize; index++, iter.next())
    {
      if(closure.action(iter.current(), impl->item(iter.current())) == 0)
	break;

    }
}

template<class K, class V> mDictPtrPtr<K,V>&
mDictPtrPtr<K,V>::operator=(const mDictPtrPtr<K, V>& other){
  *impl=*(other.impl);
  return *this;
}

#endif /* MERCATOR_DICT_PTR_PTR_H */
