// This may look like C code, but it is really -*- C++ -*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#ifndef MERCATOR_DICT_VAL_PTR_H
#define MERCATOR_DICT_VAL_PTR_H

// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/mDictValPtr.h,v $
// $Revision: 0.9 $
// $Date: 1998/05/21 23:02:03 $
//
// $Log: mDictValPtr.h,v $
// Revision 0.9  1998/05/21 23:02:03  sellden
// Added forward decl for iterator template
//
// Revision 0.8  1997/12/23 22:30:45  sellden
// Fixed broken Walk() method
//
// Revision 0.7  1997/02/16 09:06:26  sellden
// Removed dictionary constructors which take a hashing function
// pointer.  Hashing functionality is now determined by
// overriding the class HashActions.
//
// Revision 0.6  1997/02/10 05:28:54  sellden
// Moved static HashActions class to mString.h.
//
// Revision 0.5  1997/02/08 11:18:17  sellden
// Completed transfer to OTC_Map as base class.
//
// Revision 0.4  1996/12/12 14:33:11  zimba
// ose implementation of mDictValPtr
//
//Revision 1.11  1995/08/04  20:02:58  selbie
//*** empty log message ***
//
//Revision 1.10  1995/08/03  22:33:14  selbie
//added declaration for "kill_all()"
//
//Revision 1.9  1995/07/28  21:21:41  spine
//Added copyright notice.
//
//Revision 1.8  1995/07/25  22:05:34  keith
//Cleaned up some usage of strings, dicts, and lists.  Ran with -pta to
//fix some previously unseen bugs in the parameterized classes.
//
//Revision 1.7  1995/07/25  20:18:53  keith
//Fixed return types on inlined expression.
//
//Revision 1.6  1995/07/25  20:15:08  keith
//Cleaned up the code a little, fixed a memory leak in iter::current().
//
//Revision 1.5  1995/07/25  17:24:11  keith
//Added def of "nil", fixed but in del_item().
//
//Revision 1.4  1995/07/24  17:44:26  beth
//Ian's Version
//
//Revision 1.2  1995/06/05  23:11:24  suelong
//commented out syntax error on line 11
//
// * Revision 1.1  1995/06/05  22:25:11  iansmith
// * Initial revision
//  *
//
#include <OTC/collctn/map.hh>

#ifdef USONIX // BLS: EGCS compiler was complaining
template<class K, class V>class mDictValPtrIterator;
#endif

// This is the closure object. You should subclass this to get walking
// behavior that you want. If your action function returns 0 the
// walk will terminate. If you need startup or closedown help,
// you can implement either of those two functions.
//
template <class K, class V> class mDictValPtrClosure {
public:
    //
    // you can use this to setup your closure... 
    // it will get called with the client data passed to walk
    //
    virtual void 	startup(void *clientData);
    //
    // you will still get a closedown if you short circuit
    // out of the action procedure
    //
    virtual void 	closedown();
    //
    // remember to return 1 if you want the walk to continue
    //
    virtual int 	action(K&, V*);
};

template <class K, class V> inline void mDictValPtrClosure<K,V>::startup(void*)
		{ }
template <class K, class V> inline void mDictValPtrClosure<K,V>::closedown()
		{ }
template <class K, class V> inline int mDictValPtrClosure<K,V>::action(K&, V *)
		{ return 0; }

//
// This is a dictionary of values to pointers.
// K must have following methods to use this class:
//     operator==
//     operator<
//     copy constructor
//     operator=
//
// There must also be a hashing function for K with the following
// sigature:
//     unsigned hFun(const K& );
// 
// You are not allowd to have duplicates of keys in this dict.
// You can have duplicate values.
//
template <class K, class V> class mDictValPtr {
    friend class mDictValPtrIterator<K, V>;

private:
    // The implementation
    OTC_Map<K, V*> *impl;

public:
    // Default constructor.
    mDictValPtr();

  /* Outmoded constructor - hash funcs are now being implemented
   by overriding class HashActions, which is part of the OTC library.
    // Construct supplying a hash function
    mDictValPtr(unsigned (*hashKey)(const K&));
    */

    // copy constructor
    mDictValPtr(const mDictValPtr& x);

    // Destructor
    virtual ~mDictValPtr();

    // operator=
    mDictValPtr<K, V>& 	operator=(const mDictValPtr<K,V> &);

    // insert... returns 1 if the key was already present, 0 otherwise
    int	insert(const K& k, V * v);

    // lookup... returns 0 if the value is not present in the dictionary.
    // equality of keys is with the operator== on K
    V *lookup(const K& k) const;

    // del  ... deletes an object from the dictionary 
    // since we have copies of K we delete our copy but
    // we don't touch the V *.
    void del(const K& k);

    // del_item ... same as del but calls destructor on V *
    void del_item(const K& k);

    // walk ... pass a closure and client data
    void walk(mDictValPtrClosure<K, V>& closure,
	      void *clientData) const;

    // size
    unsigned size() const{return(impl->population());};

    // special method to call delete on all entries
      void kill_all();
};

///////////////////////////////////////////////////////////////////////////
//
// Ok, now we need an iterator.
//
// Unless constructed with an empty list, the OTC iterator is
// valid immediately after construction.
template <class K, class V> class mDictValPtrIterator {
private:
  OTC_Iterator<K> impl;

  // Disallow default constructor.
  mDictValPtrIterator();
    
public:
  // constructor
  mDictValPtrIterator(const mDictValPtr<K, V>& dict)
{impl=dict.impl->keys();};

  // Copy constructor
  mDictValPtrIterator(const mDictValPtrIterator<K, V>& iter): 
    impl(iter.impl){};

  // destructor
  virtual ~mDictValPtrIterator(){};

  // operator =
mDictValPtrIterator<K,V>   operator=(const mDictValPtrIterator<K, V>& other){
        impl=other.impl;
    return *this;
  };

  // get next... returns 0 if no more, 1 if there is more
  int next(){
    int done=0;//return this at end of list
    if(impl.isValid()){
      //move the iterator
      impl.next();
      //set the return to 0 if we are off the end
      if(!impl.isValid())
	done=0;
      else
      done=1;
    }
    return done;
  };

  // get current - returns a copy to ensure safety!
  K current() const
{return(impl.item());};

int isValid() const
{return impl.isValid();};

// Reset the iterator
void reset()
{impl.reset();};
};

template<class K, class V>
    mDictValPtr<K,V>::mDictValPtr(){
      impl=new OTC_Map<K,V*>();
    }

template<class K, class V>
    mDictValPtr<K,V>::mDictValPtr(const mDictValPtr& x){
            impl=new OTC_Map<K,V*>(*x.impl);
    }

template<class K, class V>
    mDictValPtr<K,V>::~mDictValPtr(){
      delete impl;
    }

template<class K, class V>
    mDictValPtr<K, V>& 	mDictValPtr<K,V>::operator=(const mDictValPtr<K,V> &x){
      *impl=*(x.impl);
            return *this;
    }

template<class K, class V> int
    mDictValPtr<K,V>::insert(const K& k, V * v){
      int found=0;//usually won't be there already
      if(impl->contains(k)){
	found=1;
	impl->remove(k);
      }
      impl->add(k,v);
      return(found);
    }

template<class K, class V> V*
    mDictValPtr<K,V>::lookup(const K& k) const{
      V* v=NULL;
      if(impl->contains(k)){
	v=impl->item(k);
      }
      return v;
    }

template<class K, class V> void
    mDictValPtr<K,V>::del(const K& k){
      if(impl->contains(k))
	    impl->remove(k);
    }

template<class K, class V> void
    mDictValPtr<K,V>::del_item(const K& k){
      if(impl->contains(k)){
	delete impl->item(k);
	impl->remove(k);
      }
}

template<class K, class V> void
mDictValPtr<K,V>::walk(mDictValPtrClosure<K, V>& closure,
void *clientData) const{
      //send the startup
      closure.startup(clientData);

      //Walk thru dictionary
mDictValPtrIterator<K,V> iter(*this);

for(iter.reset(); iter.isValid(); iter.next()) {
  // BLSX: extra step required as arg 1 is K&
  K cur = iter.current();
	if(closure.action(cur, impl->item(iter.current()))==0)
	  break;
      }
    }

template<class K, class V> void
mDictValPtr<K,V>::kill_all()
{
  mDictValPtrIterator<K,V> iter(*this);
  int dictSize = size();

  for(int index=0; index<dictSize; index++, iter.next())
    {
    del_item(iter.current());
    }

}


#endif /* MERCATOR_DICT_VAL_PTR_H */
