// This may look like C code, but it is really -*- C++ -*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#ifndef MERCATOR_DICT_VAL_VAL_H
#define MERCATOR_DICT_VAL_VAL_H

// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/mDictValVal.h,v $
// $Revision: 0.4 $
// $Date: 1998/05/21 23:01:52 $
//
// $Log: mDictValVal.h,v $
// Revision 0.4  1998/05/21 23:01:52  sellden
// Added forward decl for iterator template
//
// Revision 0.3  1997/02/10 07:41:45  sellden
// Completed transfer to OTC_Map and OTC_Iterators.
//
// Revision 0.2  1996/12/06 21:12:01  zimba
// single file template using otc libraries!! yeah
//
// Revision 1.2  1996/12/02 03:22:46  jzimba
// working toward gnu map
//
// Revision 1.1  1996/12/02 00:33:58  jzimba
// Initial revision
//
//Revision 1.6  1995/07/28  21:22:45  spine
//Added copyright notice.
//
//Revision 1.5  1995/07/25  22:05:34  keith
 //Cleaned up some usage of strings, dicts, and lists.  Ran with -pta to
//fix some previously unseen bugs in the parameterized classes.
//
//Revision 1.4  1995/07/24  17:44:26  beth
//Ian's Version
//
//Revision 1.3  1995/06/10  15:46:23  iansmith
//*** empty log message ***
//
//Revision 1.3  1995/06/09  13:25:19  iansmith
//Changed the action() function of the closure to use references for
//both K and V.
//
//Changed the insert() function to use const references for V.
//
//Changed the lookup() function to have the right signature.
//
//Change the current() function of the iterator to return a K not a K&
//as this implies a return of a local.
//
// Revision 1.2  1995/06/07  20:39:26  iansmith
// This is a checkin to insure that RCS has a version for ValVal.
//

// We implement this with a rogue wave dictionary right now
#include <OTC/collctn/map.hh>

#ifdef USONIX // BLS: EGCS compiler was complaining
template<class K, class V>class mDictValValIterator;
#endif

// This is the closure object. You should subclass this to get walking
// behavior that you want. If your action function returns 0 the
// walk will terminate. If you need startup or closedown help,
// you can implement either of those two functions.
//
template <class K, class V> class mDictValValClosure{
  //note that there is a base class OTC_Vissitor
public:
  // you can use this to setup your closure... 
  // it will get called with the client data passed to walk
  virtual void startup(void *clientData);
  // you will still get a closedown if you short circuit
  // out of the action procedure
  virtual void closedown();
  // remember to return 1 if you want the walk to continue
  virtual int action(K &, V &);
};

template <class K,class V> inline void mDictValValClosure<K, V>::startup(void*)
{ }
template <class K, class V> inline void mDictValValClosure<K, V>::closedown()
{ }
template <class K, class V> inline int mDictValValClosure<K, V>::action(K&, V&)
{ return 0; }

//////////////////////////////////////////////////////////////////////////
//
// This is a dictionary of values to values. You must  put
// values (real objects) in this dictionary.  These use deep
// copy semantics (copy constructor).  K must implement 
// the  following methods to use this class:
// operator==
// operator<
// copy constructor
// operator=
//
// There must also be a hashing function for K with the following
// sigature
// unsigned hFun(const K& );
// 
// You are not allowd to have duplicates of keys in this dict.
// You can have duplicate values.
//
template <class K, class V> class mDictValVal {
  friend class mDictValValIterator<K,V>;

  // The implementation
  OTC_Map<K,V> *impl;

  // Disallow default constructor
  mDictValVal();

public:
  // constructor
  mDictValVal(unsigned (*hashKey)(const K &));

  // copy constructor
  mDictValVal(const mDictValVal& );

  // destructor
  virtual ~mDictValVal();

  // operator=
  mDictValVal<K,V>& 		operator=(const mDictValVal<K,V>&);

  //insert... returns 1 if the key was already present, 0 otherwise
  int 			insert(const K&, const V&);

  //lookup... returns 0 if the value is not present in the dictionary
  // equality of keys is with the operator== on K
  // doesn't modify V if not there
  int 			lookup(const K&,V&) const;

  // del  ... delete's an object from the dictionary 
  void 			del(const K&);

  // walk ... pass a closure and client data
  void 			walk(mDictValValClosure<K, V>&closure,
			     void *clientData) const;

  // size
  unsigned size() const {return(unsigned)impl->population();};
};

//////////////////////////////////////////////////////////////////////////
//
// Ok, now we need an iterator.
//
template <class K, class V> class mDictValValIterator {
private:
  OTC_Iterator<K> impl;
  // Disallow default constructor.
  mDictValValIterator();
    
public:
  // constructor
  mDictValValIterator(const mDictValVal<K,V>&);
  // Copy Constructor
mDictValValIterator(const mDictValValIterator<K> &iter);
  // destructor
  virtual ~mDictValValIterator();
  // operator =
  mDictValValIterator(const mDictValValIterator<K,V>&);
  // get next... returns 0 if no more, 1 if there is more
  int next();
  // get current
  K current() const;
};

//
// mDictValVal Implementation
//

// constructor... build the implementation and make it empty
template <class K, class V>
mDictValVal<K,V>::mDictValVal(unsigned (*hashKey)(const K&)) :
    impl(new OTC_Map<K,V>())
{
}

// copy constructor 
template <class K, class V>
mDictValVal<K,V>::mDictValVal(const mDictValVal<K,V> &x) :
    impl(new Map<K,V>(x.impl))
{
}

// destructor
template <class K, class V>
mDictValVal<K,V>::~mDictValVal()
{
    delete impl;
}

// operator =
template <class K, class V> mDictValVal<K,V>&
mDictValVal<K,V>::operator=(const mDictValVal<K,V> &x)
{
    *impl = *x.impl;
    return *this;
}

// insert
template <class K, class V> int
mDictValVal<K,V>::insert(const K &k, const V &v)
{
  int result=0;
  if(impl->contains(k)){
    result=1;
    impl->remove(k);
  }
  impl->add(k,v);
  return result;
}

// lookup
template <class K, class V> int
mDictValVal<K,V>::lookup(const K &k,V &v) const
{
  int flag=0;//assume it isn't in there
  if(impl->contains(k)){
    v=impl->item(k);
    flag=1;
  }
  return flag;
}

// del..... get rid of 'em
template <class K, class V> void
mDictValVal<K,V>::del(const K &k)
{
  if(impl->contains(k))
    //blsx: is there an OTC_Map::del() ?
    impl->del(k);
}

template <class K, class V> void 
mDictValVal<K,V>::walk(mDictValValClosure<K, V>& closure, void *clientData) const
{
  OTC_Iterator<K,V> iter(*this);
  // startup info
  closure.startup(clientData);

  for(iter.reset(); iter.isValid(); iter.next()){
    // if they return 0, bail out
    if(closure.action(iter.item(), impl->item(iter.item())) == 0) 
      break;
  }
  // Call closedown function
  closure.closedown();
}

//////////////////////////////////////////////////////////////////////////
// iterator
template <class K,class V> 
mDictValValIterator<K,V>::mDictValValIterator(const mDictValVal<K,V> &dict) 
{
  impl=dict.impl->keys();
}

// Copy Constructor
template <class K,class V>
mDictValValIterator<K,V>::mDictValValIterator(const mDictValValIterator<K> &iter)
{
  impl = iter.impl;
}

// destructor
template <class K,class V>
mDictValValIterator<K,V>::~mDictValValIterator()
{
}

// move up to the next value
template <class K, class V> int
mDictValValIterator<K,V>::next()
{
  int result = 0;
  if(impl.isValid()){
    impl.next();
    if(impl.isValid())
      result = 1;

  }
  return result;
}

// get the  current value
template <class K, class V> K
mDictValValIterator<K,V>::current() const
{
  return impl.item();
}

#endif /* MERCATOR_DICT_VAL_VAL_H */
