// This may look like C code, but it is really -*- C++ -*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
//
//
//
#ifndef MERCATOR_LIST_VAL_H
#define MERCATOR_LIST_VAL_H

// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/mListVal.h,v $
// $Revision: 0.4 $
// $Date: 1998/05/21 23:01:39 $

// $Log: mListVal.h,v $
// Revision 0.4  1998/05/21 23:01:39  sellden
// Added forward decl for iterator template
//
// Revision 0.3  1997/02/01 23:03:32  sellden
// Completed migration to OTC_List and OTC_Iterator.
//
// Revision 0.2  1996/12/12 14:32:21  zimba
// ose implementation of mListVal
//
//Revision 1.7  1995/08/12  22:00:28  keith
//fixed problem with current()
//
//Revision 1.6  1995/08/12  21:39:47  keith
//Added two versions of list.at().
//
//Revision 1.5  1995/08/12  21:32:42  keith
//Added iter.prev(), iter.insertAfterPoint(), list.prepend(), and made
//iter.current() a const function (since it returns a copy).
//
//Revision 1.4  1995/07/28  21:24:15  spine
//Added copyright notice.
//
//Revision 1.3  1995/07/24  17:43:39  beth
//Ian's Version
//
//Revision 1.1  1995/05/25  14:50:43  iansmith
//Initial revision
//

#include <OTC/collctn/list.hh>
#include <OTC/collctn/iterator.hh>


#ifdef USONIX // BLS: EGCS compiler was complaining
template<class T> class mListValIterator;
#endif

/////////////////////////////////////////////////////////////////////////
//
// mListValClosure
//
template <class T> class mListValClosure {
public:
    // use this to setup your closure
    virtual void startup(void *clientData) {};
    // you'll still get a closedown if you short circuit
    //  out of the action routine
    virtual void closedown() {};
    // if the action routine returns 0 the closure stops
    // processing
    // if you don't subclass, we end up only calling the action routine once
    virtual int action(T &) {return 0;};
};


/////////////////////////////////////////////////////////////////////////
//
// mListVal
//
// 
// Linked List for Mercator
// You can put values in this list. These lists use a copy-in
// copy-out strategy.
//
// You MUST define the following:
// operator==
// operator=
// operator<
// copy constructor
//
template <class T> class mListVal 
{
  friend class mListValIterator<T>;
    
private:
  // The implementation
  OTC_List<T> *impl;
    
public:
  mListVal();
  mListVal(const mListVal &);		// deep copy
  virtual ~mListVal();		// doesn't delete contents

  mListVal<T> & operator=(const mListVal<T> &);	// deep copy

  //
  // Get the item at position i, may be used as l-value
  //
  T& at(int i){
    //    if((i> -1) && (i<impl->population()))
      return(impl->item(i));};

  //
  // Get the item as position i (copy), may NOT be used as l-value.
  //
  T at(int i) const{
    //    if((i> -1) && (i<impl->population()))
      return(impl->item(i));};

  //
  // add to the end
  //
  void 	append(const T& item){impl->addLast(item);};
  //
  // add to the beginning
  //
  void 	prepend(const T& item){impl->addFirst(item);};
  //
  // search for a value... returns 0 if not found, 1 if found
  // if found, it puts the value found into the second 
  // reference parameter... if not found foundValue is not touched
  //
  int 	search(const T &target, T &foundValue) const;
  //
  // remove the first item equal to T
  // comparison is with the == operator on T
  //  if it finds an item equal it returns 1 otherwise 0
  //
  int 	del_item(const T &);
  //
  // if you want to iterate, subclass the mListClosure and
  // pass it in here... the client data is passed to the startup routine
  //
  void 	walk(mListValClosure<T> *closure,void *clientData) const;
  //
  // how big is the list?
  //
  unsigned 	entries() const{return impl->population();};
  //
  // is it empty?
  //
  int 	empty() { return (impl->isEmpty());};
};


/////////////////////////////////////////////////////////////////////////
//
// mListValIterator
//
// OTC iterators are only undefined
// a) If they are constructed as a null iterator - that is,
//    not associated with any list
//
//  b)  They iterate off the end of a list.
//
//  c) They are assigned to an iterator that is invalid.
//
//  You should not  remove elements
// from the list during iteration.
//
template <class T> class mListValIterator {
private:
  OTC_Iterator<T> impl;
  mListValIterator();			// disallowed
    
public:
  // Constructors
    mListValIterator(const mListVal<T> &myList)
{       impl = myList.impl->items();};

    mListValIterator(const mListValIterator& other)
  {impl = other.impl; };

  // Destructor
    virtual ~mListValIterator(){};

    mListValIterator<T> & operator=(const mListValIterator<T>& other)
  {impl = other.impl;
  return (*this);};
    
    //
  // Iterate - return 0 when you reach the end
    // It is in an undefined state at that point....
    // you have to query using current() to get the current value
    //
    int 	next();
    //
    // Move back one item.
    //
  //    int		prev();
    //
    // get the current value under the iterator.  NOTE THAT THIS IS A COPY.
    //
    T 		current() const;
    //
    // reset the iterator to the undefined state it had at creation time
    //
    void 	reset()
  {impl.reset();};

    //
    // Add a new T after the position of the iterator.
    //
  //    void	insertAfterPoint(const T& t);
};

//template copare function
template <class T> int
mListVal<T>::del_item(const T & val)
{
  int found=0;//assume not found
  for(int i=0; i<impl->population(); i++){
    if(val == at(i)){
      found=1;
      impl->removeItem(i);
      break;
    }
  }
  return found;
}

template <class T> void
mListVal<T>::walk(mListValClosure<T> *closure, void *clientData) const
{
  closure->startup(clientData);
  // walk the values
  for(int i=0;i<impl->population();i++){
    if(closure->action(at(i))==0) 
      break;
  }
  // give 'em a closedown
  closure->closedown();
}

//
// constructor (empty list)
//
template <class T>
mListVal<T>::mListVal()
{
  impl=new OTC_List<T>;
}

//
// copy constructor.. this will result in a copy of all members of the list
//
template <class T>
mListVal<T>::mListVal(const mListVal<T> &v)
{
  impl=new OTC_List<T>(*(v.impl));
}

//
// destructor... this deletes the data structure and all of its copies
//
template <class T>
mListVal<T>::~mListVal()
{
  delete impl;
}

//
// assignment is a deep copy of the lists
//
template <class T> mListVal<T>&
mListVal<T>::operator=(const mListVal<T> &t)
{
  *impl = (*(t.impl));
  return (*this);
}

template <class T> int 	
mListVal<T>::search(const T &target, T & foundValue) const{
  int found=0;//assume not found
  for(int i=0; i<impl->population(); i++){
    if(target == at(i)){
      found=1;
      foundValue=at(i);
      break;
    }
  }
  return(found);
}; 

////////////////////////////////////////////////////////////////////////////
//
// mListValIterator
//
//
// advance
//
template <class T> int
mListValIterator<T>::next() 
{
  int done=1;//return this at end of list
  if(impl.isValid()){
    //move the iterator
    impl.next();
    //set the return to 0 if we are off the end
    if(!impl.isValid())
      done=0;

  }
  /*
else
    impl.reset();
    */

  return done;
}

/*
//
// reverse
//
template <class T> int
mListValIterator<T>::prev()
{
  //Not implemented
}
*/

//
// get current value
//
template <class T> T
mListValIterator<T>::current() const
     // blsx: possible source of exception
{
  //  if (impl.isValid())
  return(impl.item());
}

//
// Add a new item after the position of the iterator.
//
/*
template <class T> void
mListValIterator<T>::insertAfterPoint(const T& val)
{
  //  impl.insertAfterPoint(t);
}
*/

#endif  /* MERCATOR_LIST_VAL_H */
