//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: mString.cc,v $
// Revision 0.5  1997/02/08 11:17:30  sellden
// Added int intHash() required for OTC_Maps.
//
// Revision 0.4  1997/01/29 22:03:31  sellden
// Added the much wated for hashing function
//
// Revision 0.3  1997/01/09 21:38:38  sellden
// Wrapped all changes thus far in 'ifdef USONIX'.
//
// Revision 0.2  1996/12/04 19:51:33  sellden
// Changed to GNU String.  Note loss of String::hash().
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//
#include "mRegexp.h"
#include "mString.h"

mString::mString() : impl() { }
mString::mString(const char *c) : impl(c) { }
mString::mString(const char *c, int n) : impl(c, n) {}
mString::mString(const mString& m) : impl(m.impl) { }

mString::operator char *() const { return Strdup(impl); }
mString::operator const char *() const { return impl; }

mString& mString::operator=(const mString& m) {
  impl = m.impl; return *this;
}

mString& mString::operator+=(const mString &m) {
  impl += m.impl; return *this;
}

mString& mString::operator+=(const char *c) {
  impl += mString(c); return *this;
}

unsigned mString::hash() const { 
#ifdef USONIX
  int len=length();
  unsigned prod=1;

  for(int index=0; index < len; index++)
    prod*=(unsigned)impl[index];

  return prod;
#else
  return impl.hash(); 
#endif
}

#ifdef USONIX
int mString::intHash() const { 
  int len=length();
  int prod=1;

  for(int index=0; index < len; index++)
    prod*=(int)impl[index];

  return ((prod <0)? -prod : prod);
}
#endif

unsigned mString::length() const { 
  return impl.length(); 
}

int mString::operator==(const mString& m) const {
  return impl == m.impl;
}

int mString::operator<(const mString& m) const {
  return impl < m.impl;
}

int mString::index(const char *c, int i) const {
  return impl.index(c, i);
}

int mString::index(const mString &m, int i) const {
  return impl.index(m.impl, i);
}

int mString::index(const mRegexp &r, int i) const {
  return impl.index(r.impl, i);
}

int mString::index(const mRegexp &r, unsigned int *len, 
			  unsigned int i) const {
  // In the C++ lib, this functionality is more neatly found in Regex.
#ifdef USONIX
  return r.impl.search(impl, strlen(impl), (int&)len, i);
#else
  return impl.index(r.impl, len, i);
#endif
}

#ifdef USONIX
int mString::insert(int i, const char *c) {
  if ((i < 0) || ( i > impl.length()))
    return 0;
  impl(i, 0) = c;
  return 1;
}

int mString::insert(int i, mString& m) {
  if ((i < 0) || ( i > impl.length()))
    return 0;
  impl(i, 0) = m;
  return 1;
}

#else

int mString::insert(int i, const char *c) {
  if (impl(i,0).isNull())
    return 0;
  impl(i, 0) = c;
  return 1;
}

int mString::insert(int i, mString& m) {
  if (impl(i,0).isNull())
    return 0;
  impl(i, 0) = m;
  return 1;
}
#endif

int mString::subst(mRegexp &r, int i, mString &m) {

#ifdef USONIX
  int j, matchLen=0;
  if ((j = r.impl.search(impl, impl.length(), matchLen, i)) >= 0) {
    impl(j, matchLen) = m;
#else
  int j;
  if ((j = impl.index(r.impl, i)) >= 0) {
    impl(r.impl, i) = m.impl;
#endif
    return j + strlen((const char *)m.impl);
  }
  else
    return -1;
}

int mString::subst(mRegexp &r, int i, const char *c) {
#ifdef USONIX
  int j, matchLen=0;
  if ((j = r.impl.search(impl, impl.length(), matchLen, i)) >= 0) {
    impl(j, matchLen) = c;
#else
  int j;
  if ((j = impl.index(r.impl, i)) >= 0) {
    impl(r.impl, i) = c;
#endif
    return j + strlen(c);
  }
  else
    return -1;
}

mString* mString::substString(mRegexp &r, int i) {
  mString *ret = new mString();
#ifdef USONIX
  int j, matchLen=0;
  if ((j = r.impl.search(impl, impl.length(), matchLen, i)) >= 0) 
    ret->impl = impl(j, matchLen);
  return ret;
#else
  ret->impl = impl(r.impl, i); return ret;
#endif
}

void mString::replace(int i, int len, mString& m) {
  impl(i, len) = m.impl; 
}

void mString::replace(int i, int len, const char *c) {
  impl(i, len) = c;
}
