// This may look like C code, but it is really -*- C++ -*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: mString.h,v $
// Revision 0.6  1997/02/10 05:28:54  sellden
// Added override functionality for hashing and ranking
// mString and mString* - used by OTC_Map.
//
// Revision 0.5  1997/02/08 11:17:30  sellden
// Added int intHash() required for OTC_Maps.
//
// Revision 0.4  1997/01/09 21:38:00  sellden
// Wrapped all changes thus far in 'ifdef USONIX'.
//
// Revision 0.3  1996/12/11 18:24:11  barnaby
// Added rwnil definition
//
// Revision 0.2  1996/12/04 19:50:19  sellden
// Changed to GNU String.  Note loss of String::hash().
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//
#ifndef M_STRING_H
#define M_STRING_H

#ifdef USONIX
#include <String.h>
#include<OTC/collctn/hashactn.hh>
#include<OTC/collctn/rankactn.hh>
#else
#include <rw/cstring.h>
#endif

#include "mRegexp.h"
#include "Util.h"

#ifdef USONIX	// TB: Removal of RW
#define nil NULL
#else
#ifndef nil
#define nil rwnil
#endif /* nil */
#endif


class mString {
private:
#ifdef USONIX
  String impl;
#else
RWCString 		impl;
#endif
  
public:
  friend class mRegexp;
  // construct with null string
  mString();
  // construct from const char *
  mString(const char *c);
  mString(const char *c, int n);
  // copy constructor
  mString(const mString& m);
  // conversion to char * (returns a copy, so the caller MUST delete it)
  operator char *() const;
  // conversion to const char *.  Caller MUST NOT modify or delete!
  operator const char *() const;
  // assignment
  mString& operator=(const mString &m);
  mString& operator+=(const mString &m);
  mString& operator+=(const char *c);
  
  unsigned hash() const;
#ifdef USONIX
  int intHash() const;
#endif
  unsigned length() const;
  
  int operator==(const mString& m) const;
  int operator<(const mString& m) const;

  int index(const char *c, int i=0) const;
  int index(const mString &m, int i=0) const;
  int index(const mRegexp &r, int i=0) const;
  int index(const mRegexp &r, unsigned int *len, unsigned int i=0) const;

  int insert(int i, const char *c);
  int insert(int i, mString& m);

  // Substitute mString or char * string at position marked by regular
  //   expression r, beginning at offset i within the mString.  The 
  //   offset of the final character substituted is returned, -1 if no
  //   substitution occurred.
  int subst(mRegexp &r, int i, mString& m);
  int subst(mRegexp &r, int i, const char *c);
  // Returns the actual string located by the regular expression matcher
  //   beginning at offset i within the mString.
  mString* substString(mRegexp &r, int i);

  // Replace the subsequence beginning at offset i with length len
  //   with mString m or char * c.
  void replace(int i, int len, mString& m);
  void replace(int i, int len, const char *c);
};

#ifdef USONIX
// Override default hashing function for OTC_Map
class OTC_HashActions<mString*>
{
public:
    static int		hash(mString* const& theItem)
				{ 				    return theItem->intHash();}
				// Returns a hash value for <theItem>.
};

// Override default ranking for OTC_Map
class OTC_RankActions<mString*>
{
public:
inline static int rank(mString* const& theItem1, mString* const& theItem2)
{
  if((*theItem1) < (*theItem2)) return -1;
  if((*theItem1) == (*theItem2)) return 0;
  return 1;
};
};

// Override default hashing function for OTC_Map
class OTC_HashActions<mString>
{
public:
    static int		hash(mString const& theItem)
				{ 				    return theItem.intHash();}
				// Returns a hash value for <theItem>.
};
#endif

#endif /* M_STRING_H */
