// -*-Mode:C++-*-
//
//
//
//	Note:  The use of this program shall be restricted to educational,
//	non-commercial use only.  All inquiries regarding licensing this
//	software for commercial use may be addressed to:
//	
//	Georgia Tech Research Corporation
//	Technology Licensing
//	Centennial Research Building - Rm. 275
//	400 Tenth Street, N.W.
//	Atlanta, Georgia  30332-0415
//	
//	*************************************************
//	
//	       Copyright 1996
//	Georgia Tech Research Corporation
//	    All Rights Reserved
//
// $Log: mercator.cc,v $
// Revision 0.4  1998/08/03 20:42:43  sellden
// Started separating out code for main window
//
// Revision 0.3  1998/05/28 19:42:10  sellden
// Added quit button to the Ultrasonix window.
//
// Revision 0.2  1997/12/23 22:30:46  sellden
// Added main window for Ultrasonix
//
// Revision 0.1  1996/12/02 20:06:39  rcsg
// Initial Port Revision
//
//
//
// The main for the Mercator project
// This basically starts up everything and lets them wait
// for input from the desired applications. 
// 

//
// $Author: sellden $
// $Source: /usr/mercator/newsource/7.0/RCS/mercator.cc,v $
// $Revision: 0.4 $
// $Date: 1998/08/03 20:42:43 $
//

static const char rcsid[]="@(#) $Header: /usr/mercator/newsource/7.0/RCS/mercator.cc,v 0.4 1998/08/03 20:42:43 sellden Exp $";


#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <fstream.h>
#include "Mercator.h"
#include "FDInterest.h"
#include "Error.h"
#include "Config.h"
#include "XServer.h"
#include "Interp.h"
#include "Template.h"
#include "Attribute.h"
#include "mListVal.h"
#include "Audio.h"
#include "Braille.h"
#include "Speech.h"
#include"RAPListener.h"
#include "MainWindow.h"
#include  <X11/StringDefs.h>
#include  <X11/Xaw/Box.h>
#include  <X11/Xaw/Command.h>

/////
/////	Here is the one, almighty Mercator instance.  Yes, it is
/////	global.
/////
Mercator *mercator = NULL;

/*
Globals
*/

Widget rap_agnt=NULL;
extern Widget top_level;
extern XtAppContext app_context;

void (* FDInterest::CleanupProc)();


void createAgent() {
  int arg_cnt=0;
  Arg		arg_list[10];

    XtSetArg(arg_list[arg_cnt],RapAgntNautoInitiate,
	     True); arg_cnt++;
    XtSetArg(arg_list[arg_cnt], XtNbaseWidth,
	     200); arg_cnt++;
    XtSetArg(arg_list[arg_cnt], XtNbaseHeight,
	     200); arg_cnt++;

    ::rap_agnt = RapAgntCreate(::top_level, "rap_agnt",
			       arg_list,arg_cnt, RAPListener::GetProc);

#ifdef LOTS_O_PRINTFS
printf("rap_agnt = %x\n", ::rap_agnt);
#endif

    XtAddCallback(::rap_agnt,RapAgntNclientCallback,
		  (XtCallbackProc)RAPListener::ClientProcMarshaller,0);
    XtAddCallback(::rap_agnt,RapAgntNobjectInfoCallback,
		  (XtCallbackProc)RAPListener::ObjectInfoProcMarshaller,0);
    XtAddCallback(::rap_agnt,RapAgntNhookCallback,
		  (XtCallbackProc)RAPListener::HookProcMarshaller,0);
    XtAddCallback(::rap_agnt,RapAgntNeventCallback,
		  (XtCallbackProc)RAPListener::EventProcMarshaller,0);
    XtAddCallback(::rap_agnt,RapAgntNrequestCallback,
		  (XtCallbackProc)RAPListener::RequestProcMarshaller,0);
}

static void Die() {
    // Free up our filters, template, config, and attribute information.
    mercator->FreeFilters();
    mercator->FreeTemplates();
    mercator->FreeConfig();
    mercator->FreeAttrib();

    // Clean up.
    delete mercator;

    // This gives the error: Couldn't get per display information
    //    XtDestroyApplicationContext(::app_context);

    Error(EL_DEBUG, "FDInterest::CleanupProc",
	  "Exiting Ultrasonix...", NULL);

  exit(0);

return;
}

int
main(int argc, char *argv[], char *envp[])
{
    
    cout << "\n\tCopyright 1992, 1993, 1994, 1995, 1996\n";
    cout << "\tGeorgia Tech Research Corporation\n";
    cout << "\tAll rights reserved.\n\n";

    cout << "\n\tNote:  The use of this program shall be restricted to educational,\n";
    cout << "non-commercial use only.  All inquiries regarding licensing this\n";
    cout << "software for commercial use may be addressed to:\n";
    cout << "\n\tGeorgia Tech Research Corporation\n";
    cout << "\tTechnology Licensing\n";
    cout << "\tCentennial Research Building - Rm. 275\n";
    cout << "\t400 Tenth Street, N.W.\n";
    cout << "\tAtlanta, Georgia  30332-0415\n";

/* Need to set the global top_level before calling the Mercator ctor */
// Sets global values of top_level and app_context 
 createTopLevel(argc, argv);

    mercator = new Mercator();

    	    createUI();

    // Open a RAP listener
    mercator->SetRAPListener(new RAPListener(*mercator->GetXServer()));

    // Create the agent
	    createAgent();


    // Have RAPListener register itself.
    if (!mercator->GetRAPListener()->Register(*mercator->GetXServer())) {
	Error(EL_FATAL, "Mercator::Mercator",
	      "Couldn't register RAP protocol.", NULL);
    }

    mercator->InstallSignalHandlers();
    mercator->ProcessEnvVariables();

    if (!mercator->ProcessArgs(argc, argv))
	Error(EL_FATAL, argv[0], "Cannot continue.", NULL);

    /* attribute file must load before config file */
    if (!mercator->ParseAttribFile())
        Error(EL_ERROR, argv[0], "Error parsing attribute file: ",
	      mercator->AttribFilename(),  NULL);

    if (!mercator->ParseConfigFile())
	Error(EL_ERROR, argv[0], "Error parsing configuration file: ",
	      mercator->ConfigFilename(), ".", NULL);
    
    mercator->ProcessDefaults();
    mercator->GetInterp()->SourceInitialTCLfile();
    
    // Load our templates.
    if (!mercator->LoadTemplates())
	Error(EL_ERROR, argv[0], "Error while loading templates.", NULL);
    // Load definitions for ScreenReader filters
    mercator->LoadFilterDictionary();

    mercator->StartConsole();

    mercator->DumpMsg();
    
    FDInterest::CleanupProc = Die;

    XtAppAddWorkProc(::app_context, FDInterest::workProc, (XtPointer)mercator);
    XtAppMainLoop(::app_context);

    // Free up our filters, template, config, and attribute information.
    mercator->FreeFilters();
    mercator->FreeTemplates();
    mercator->FreeConfig();
    mercator->FreeAttrib();

    // Clean up.
    delete mercator;
  
    return 0;
}
