/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
//
//      Copyright 1992, 1993, 1994, 1995
//      Georgia Tech Research Corporation
//      All rights reserved.
//
*/
/*
* $Log: RapAgnt-in.c,v $
* Revision 0.2  1997/11/20 21:57:38  sellden
* Added Marty's pointer-copy fix to AddObject()
*
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*                                                                      */
/* RapAgnt-in.c                                                         */
/*                                                                      */
/*      Handles all incoming RAP messages to the agent.			*/
/*                                                                      */
/************************************************************************/
#define NEED_EVENTS
#include <X11/Xmd.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#include "RapAgntP.h"

#include <X11/Xproto.h>

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEmsg.h>
#include <X11/ICE/ICEproto.h>
#include "Rap.h"
#include "RapProto.h"

/************************************************************************/
/*									*/
/* RapAgntHandleHelloReply						*/
/*									*/
/*	When the ICE subprotocol setup is complete, the agent will send	*/
/*	a RapHelloRequest to the client to set up some additional	*/
/*	handshaking.  This routine handles the reply from the client.	*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleHelloReply(RapAgntWidget	raw,
			char		*pdata,
			Window		*window)
{
    char	     *bufptr = pdata;
    CARD32	     pad;
    
    EXTRACT_CARD32(bufptr,False,*window);
    EXTRACT_CARD32(bufptr,False,pad);
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleHelloReply",
	  "window id %x\n",
	  *window);
#endif
    
    return RapSuccess;
    
} /* RapAgntHandleHelloReply */

/************************************************************************/
/*									*/
/* RapAgntHandleAck							*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleAck(RapAgntWidget	raw,
		 CARD16		seq,
		 char		*pdata)
{    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleAck","seq = %d\n",seq);
#endif

    return RapSuccess;

} /* RapAgntHandleAck */

/************************************************************************/
/*									*/
/* RapAgntFindObject							*/
/*									*/
/************************************************************************/
RapAgntObjectPtr
RapAgntFindObject(RapAgntObjectPtr root, RapObjectID *id)
{
    RapAgntObjectPtr found = NULL;
    int i;

    if (root->id.id == id->id)
	found = root;
    else {
	i = 0;
	while ((i < root->num_children) && (found == NULL)) {
	    found = RapAgntFindObject(&(root->children[i]),id);
	    i++;
	}	    
    }
    
    return found;
    
} /* RapAgntFindObject */

/************************************************************************/
/*									*/
/* AddObject								*/
/*									*/
/************************************************************************/
static void
AddObject(RapAgntObjectPtr	root,
	  RapAgntObjectPtr	object,
	  RapObjectID		*parent_id)
{
    RapAgntObjectPtr	parent;
#ifdef USONIX
    int i, j;
#endif

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"AddObject",
	  "object = %lx\n",object->id.id);
    InfoF(__FILE__,__LINE__,"AddObject",
	  "parent = %lx\n",parent_id->id);
    InfoF(__FILE__,__LINE__,"AddObject",
	  "name = %s\n",object->name);
    InfoF(__FILE__,__LINE__,"AddObject",
	  "class_name = %s\n",object->class_name);
    InfoF(__FILE__,__LINE__,"AddObject",
	  "window = %x\n",object->window);
    InfoF(__FILE__,__LINE__,"AddObject",
	  "managed = %d\n",object->managed);
    InfoF(__FILE__,__LINE__,"AddObject",
	  "toolkit = %s\n",object->toolkit);
    InfoF(__FILE__,__LINE__,"AddObject",
	  "num_resources = %d\n",object->num_resources);
#endif

    if (RapAgntFindObject(root,&object->id) != NULL) {      
	ErrorF(__FILE__,__LINE__,"AddObject",
	       "object %lx is already in tree!\n",object->id.id);
	return;
    }
    
    parent = RapAgntFindObject(root,parent_id);
    if (parent == NULL) {
	ErrorF(__FILE__,__LINE__,"AddObject",
	       "Could not find parent %lx in tree!\n",parent_id->id);
	return;
    }
    
    /* Allocate more space if necessary
     */
    if (parent->num_children >= parent->alloc_children) {
	parent->alloc_children += ALLOC_INC;
	parent->children = (RapAgntObjectPtr) 
	    XtRealloc((char *)parent->children,
		      sizeof(RapAgntObject) *
		      parent->alloc_children);

#ifdef USONIX
	/* Added 11/7/97 by megak.  XtRealloc sometimes moves the memory
	   allocated to a different location.  Thus, all of the grandchildren
	   of this parent, must have their parent pointers reset.
	 */
	for (i=0;i<parent->num_children;i++)
	  for (j=0;j<parent->children[i].num_children;j++)
	    parent->children[i].children[j].parent = &parent->children[i];
	/* End 11/7/97 additions by megak */
#endif
    }
    
    /* Save the information
     */
    /* BLS: changed order of next 2 statements */
    object->parent = parent;
    parent->children[parent->num_children] = *object;
    parent->num_children++;

} /* AddObject */

/************************************************************************/
/*									*/
/* RapAgntFreeResourceInfo						*/
/*									*/
/************************************************************************/
void
RapAgntFreeResourceInfo(RapAgntObjectPtr object)
{
    int i;
    
    if (object->resources) {
	for (i = 0; i < object->num_resources; i++) {
	    if (object->resources[i].name)	
		XtFree(object->resources[i].name);
	    if (object->resources[i].class_name)	
		XtFree(object->resources[i].class_name);
	    if (object->resources[i].type)	
		XtFree(object->resources[i].type);
	    if (object->resources[i].value) {
		XtFree((char *)object->resources[i].value->type);
		XtFree((char *)object->resources[i].value->value);
		XtFree((char *)object->resources[i].value);
	    }
	}
	XtFree((char *)object->resources);
    }

    object->resources = NULL;
    object->num_resources = 0;
    
} /* RapAgntFreeResourceInfo */

/************************************************************************/
/*									*/
/* RapAgntFreeObjectInfo						*/
/*									*/
/************************************************************************/
void
RapAgntFreeObjectInfo(RapAgntObjectPtr object)
{
    int i;
    
    /* Remove the children first.
     */
    if (object->children) {
	for (i = 0; i < object->num_children; i++)
	    RapAgntFreeObjectInfo(&object->children[i]);
	XtFree((char *)object->children);	
    }

    if (object->name) 		XtFree(object->name);
    if (object->class_name)	XtFree(object->class_name);
    if (object->toolkit)	XtFree(object->toolkit);

    RapAgntFreeResourceInfo(object);

    memset(object,'\0',sizeof(RapAgntObject));
    
} /* RapAgntFreeObjectInfo */

/************************************************************************/
/*									*/
/* ExtractResource							*/
/*									*/
/************************************************************************/
static void 
ExtractResource(char		   	**bufptr,
		 RapAgntResourcePtr 	resource,
		 Boolean		has_value)
{
    CARD32 pad;
    
    EXTRACT_STRING(*bufptr,False,resource->name);
    EXTRACT_STRING(*bufptr,False,resource->class_name);
    EXTRACT_STRING(*bufptr,False,resource->type);
    EXTRACT_CARD32(*bufptr,False,resource->kind);
    EXTRACT_CARD32(*bufptr,False,pad);

#ifdef DEBUG2
    InfoF(__FILE__,__LINE__,"ExtractResource",
	  "name = %s\n",resource->name);
    InfoF(__FILE__,__LINE__,"ExtractResource",
	  "class_name = %s\n",resource->class_name);
    InfoF(__FILE__,__LINE__,"ExtractResource",
	  "type = %s\n",resource->type);
    InfoF(__FILE__,__LINE__,"ExtractResources",
	  "kind = %d\n",resource->kind);
#endif

    if (has_value) {
	resource->value = (RapValuePtr) XtMalloc(sizeof(RapValue));
	EXTRACT_STRING(*bufptr,False,resource->value->type);
	EXTRACT_BYTES(*bufptr,False,
		      resource->value->size,
		      resource->value->value);
#ifdef DEBUG2
	if (strcmp(resource->value->type,XtRString) == 0)
	    InfoF(__FILE__,__LINE__,"ExtractResource",
		  "value (size = %d) = %s\n",
		  resource->value->size,resource->value->value);
	else
	    InfoF(__FILE__,__LINE__,"ExtractResource",
		  "value (size = %d) = %lx\n",
		  resource->value->size,resource->value->value);
#endif
    }
    
} /* ExtractResource */

/************************************************************************/
/*									*/
/* ExtractResources							*/
/*									*/
/************************************************************************/
static void 
ExtractResources(char		   	**bufptr,
		 int  		   	*num_resources,
		 RapAgntResourcePtr 	*resources,
		 Boolean		has_values)
{
    int 	i;
    CARD32 	pad;
    
    EXTRACT_CARD32(*bufptr,False,*num_resources);
    EXTRACT_CARD32(*bufptr,False,pad);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"ExtractResources",
	  "%d resources\n",*num_resources);
#endif
    
    *resources = (RapAgntResourcePtr) XtMalloc(sizeof(RapAgntResource) *
					       (*num_resources));
    memset(*resources,'\0',sizeof(RapAgntResource) * (*num_resources));
    
    for (i = 0; i < *num_resources; i++)
	ExtractResource(bufptr,&((*resources)[i]),has_values);
    
} /* ExtractResources */

/************************************************************************/
/*									*/
/* ExtractObject							*/
/*									*/
/************************************************************************/
static void 
ExtractObject(char		**bufptr,
	      RapAgntObjectPtr 	object,
	      RapObjectID	*parent_id,
	      Boolean	       	has_values)
{
    RapObjectID		object_id;
    char		*name,*class_name;
    CARD32		window,managed;
    char		*toolkit;
    
    EXTRACT_OBJECTID(*bufptr,False,object->id);
    EXTRACT_OBJECTID(*bufptr,False,*parent_id);
    EXTRACT_STRING(*bufptr,False,object->name);
    EXTRACT_STRING(*bufptr,False,object->class_name);
    EXTRACT_CARD32(*bufptr,False,object->window);
    EXTRACT_CARD32(*bufptr,False,object->managed);
    EXTRACT_STRING(*bufptr,False,object->toolkit);
    
    if (has_values)
	ExtractResources(bufptr,
			 &(object->num_resources),
			 &(object->resources),
			 True);
    
} /* ExtractObject */

/************************************************************************/
/*									*/
/* RapAgntHandleQueryTreeReply						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleQueryTreeReply(RapAgntWidget	raw,
			    RapAgntClientPtr	client,
			    CARD16		seq,
			    char		*pdata,
			    Boolean		has_resources)
{
    char 		*bufptr = pdata;
    RapAgntShellList    *shell_list = NULL;
    CARD32 		pad;
    RapAgntObject	object;
    RapObjectID		parent_id;
    register int 	i,num_objects;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleQueryTreeReply","\n");
#endif

    shell_list = (RapAgntShellList *) 
	XtMalloc(sizeof(RapAgntShellList));
    memset(shell_list,'\0',sizeof(RapAgntShellList));
	
    EXTRACT_CARD32(bufptr,False,shell_list->num_shells);
    EXTRACT_CARD32(bufptr,False,pad);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleQueryTreeReply",
	  "num_shells\n",shell_list->num_shells);
#endif

    /* Now get the shells.
     */
    shell_list->shells = (RapAgntObjectPtr) XtMalloc(sizeof(RapAgntObject) *
						     shell_list->num_shells);
    memset(shell_list->shells,'\0',
	   sizeof(RapAgntObject) * shell_list->num_shells);

    for (i=0; i < shell_list->num_shells; i++) {
	register int j;

	EXTRACT_CARD32(bufptr,False,num_objects);
	EXTRACT_CARD32(bufptr,False,pad);

#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapAgntHandleQueryTreeReply",
	      "num_objects in shell %d = %d\n",i,num_objects);
#endif
	
	/* The first object is the shell
	 */
	ExtractObject(&bufptr,&shell_list->shells[i],&parent_id,has_resources);
	shell_list->shells[i].client = client;
	
	/* The rest are the children
	 */
	for (j=0; j < (num_objects-1); j++) {
	    memset(&object,'\0',sizeof(RapAgntObject));
	    ExtractObject(&bufptr,&object,&parent_id,has_resources);
	    object.client = client;
	    AddObject(&shell_list->shells[i],&object,&parent_id);
	}
    }

    if (raw->rapagnt.waiting_for_reply) {
	raw->rapagnt.reply = (XtPointer) shell_list;
	raw->rapagnt.waiting_for_reply = False;
    }
    else {    
	RapAgntObjectInfoCallbackStruct temp;
	temp.reason         = has_resources ? RapFullQueryTreeReply : 
	                                   RapQueryTreeReply;
	temp.sequence_number= seq;
	temp.client 	    = client;
	temp.error_code     = RapNoError;
	temp.data	    = shell_list;
	
	XtCallCallbacks((Widget)raw,
			RapAgntNobjectInfoCallback,
			&temp);
    }
    
    return RapSuccess;
    
} /* RapAgntHandleQueryTreeReply */

/************************************************************************/
/*									*/
/* RapAgntHandleGetResourcesReply					*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleGetResourcesReply(RapAgntWidget	raw,
			       RapAgntClientPtr	client,
			       CARD16		seq,
			       char		*pdata)
{
    char 		*bufptr = pdata;
    int			error_code,i;
    RapAgntResourceList *res_list=NULL;
    CARD32		pad;
    
    EXTRACT_CARD32(bufptr,False,error_code);
    EXTRACT_CARD32(bufptr,False,pad);

#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapAgntHandleGetResourcesReply",
	      "error_code = %d\n",error_code);
#endif

    if (error_code == RapNoError) {
	res_list = (RapAgntResourceList *) 
	    XtMalloc(sizeof(RapAgntResourceList));
	memset(res_list,'\0',sizeof(RapAgntResourceList));
	
	ExtractResources(&bufptr,
			 &(res_list->num_resources),
			 &(res_list->resources),
			 False);
    }

    if (raw->rapagnt.waiting_for_reply) {
	raw->rapagnt.reply = (XtPointer) res_list;
	raw->rapagnt.waiting_for_reply = False;
    }
    else {    
	RapAgntObjectInfoCallbackStruct temp;

	temp.reason         = RapGetResourcesReply;
	temp.sequence_number= seq;
	temp.client 	    = client;
	temp.error_code	    = error_code;
	temp.data	    = (XtPointer) res_list;
	
	XtCallCallbacks((Widget)raw,
			RapAgntNobjectInfoCallback,
			&temp);
    }
        
    return RapSuccess;
    
} /* RapAgntHandleGetResourcesReply */

#ifdef GC_SUPPORT
/************************************************************************/
/*									*/
/* RapAgntHandleGetGCValuesReply						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleGetGCValuesReply(RapAgntWidget	raw,
			    RapAgntClientPtr	client,
			    CARD16		seq,
			    char		*pdata)
{
    char 			   *bufptr = pdata;
    int				   error_code = RapNoError;
    RapStatus status = RapSuccess;
    register int 		   i;
    RapGCValuesList *reply = NULL;

   /* now the fun part!! */
   reply = (RapGCValuesList*) XtMalloc (sizeof(RapGCValuesList) );
   memset(reply, '\0', sizeof(RapGCValuesList));

   EXTRACT_CARD32(bufptr, FALSE, reply->numGCs);

   reply->GCValues = (RapGCValuesPtr)
       XtMalloc(sizeof(RapGCValues) * reply->numGCs);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetGCValuesReply",
	  "Arrived with numGCs = %d\n",reply->numGCs);
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetGCValuesReply",
	  "allocated %d bytes\n",sizeof(RapGCValues) * reply->numGCs);
#endif

   for (i=0 ; i<reply->numGCs ; i++) {
       EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].gc);
       EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].errorCode);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetGCValuesReply",
	  "%d: gc = %lx, error code = %d \n", i+1, reply->GCValues[i].gc, 
	  reply->GCValues[i].errorCode);
#endif

       if (reply->GCValues[i].errorCode == RapNoError) {
	   long dashes;
 
	/* get the structure!!  */
	
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.function);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.plane_mask);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.foreground);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.background);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.line_width);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.line_style);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.cap_style);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.join_style);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.fill_style);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.fill_rule);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.arc_mode);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.tile);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.stipple);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.ts_x_origin);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.ts_y_origin);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.font);
	   EXTRACT_CARD32(bufptr, FALSE,
			  reply->GCValues[i].data.gcv.subwindow_mode);
	   EXTRACT_CARD32(bufptr, FALSE,
			  reply->GCValues[i].data.gcv.graphics_exposures);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.clip_x_origin);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.clip_y_origin);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.clip_mask);
	   EXTRACT_CARD32(bufptr, FALSE, reply->GCValues[i].data.gcv.dash_offset);
	   EXTRACT_CARD32(bufptr, FALSE, dashes);

	   reply->GCValues[i].data.gcv.dashes = (char) dashes;	
       } else {
	   EXTRACT_STRING(bufptr, FALSE,
			  reply->GCValues[i].data.errorString);

       }
   }

    if (raw->rapagnt.waiting_for_reply) {
	raw->rapagnt.reply = (XtPointer) reply;
	raw->rapagnt.waiting_for_reply = False;
    }
    else {
	RapAgntObjectInfoCallbackStruct temp;

	temp.reason 	    = RapGetGCValuesReply;
	temp.sequence_number= seq;
	temp.client 	    = client;
	temp.error_code	    = error_code;
	temp.data	    = (XtPointer) reply;
    
	XtCallCallbacks((Widget)raw,
			RapAgntNobjectInfoCallback,
			&temp);
    }

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetGCValuesReply",
	  "Leaving...\n",NULL);
#endif

   return status;
} /* RapAgntHandleGetGCValuesReply */

#endif /* GC_SUPPORT */
/************************************************************************/
/*									*/
/* RapAgntHandleGetValuesReply						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleGetValuesReply(RapAgntWidget	raw,
			    RapAgntClientPtr	client,
			    CARD16		seq,
			    char		*pdata)
{
    char 			   *bufptr = pdata;
    int				   error_code;
    int				   num_values;
    RapValuePtr 	   	   values = NULL;
    register int 		   i;
    
    EXTRACT_CARD32(bufptr,False,error_code);
    EXTRACT_CARD32(bufptr,False,num_values);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetValuesReply",
	  "error_code = %d, num_values = %d\n",error_code,num_values);
#endif

    /* Now get the resources if there is no error
     */
    if (error_code == RapNoError) {
	values = (RapValuePtr) XtMalloc(sizeof(RapValue) * num_values);
	memset(values,'\0',sizeof(RapValue) * num_values);
    
	for (i=0; i < num_values; i++) {
	    EXTRACT_STRING(bufptr,False,values[i].type);
	    EXTRACT_BYTES(bufptr,False,values[i].size,values[i].value);

#ifdef DEBUG
	    if (strcmp(values[i].type,XtRString) == 0)
		InfoF(__FILE__,__LINE__,"RapAgntHandleGetValuesReply",
		      "value (size=%d) = %s\n",values[i].size,values[i].value);
	    else
		InfoF(__FILE__,__LINE__,"RapAgntHandleGetValuesReply",
		      "value (size=%d) = %lx\n",values[i].size,values[i].value);
#endif
	}
    }

    if (raw->rapagnt.waiting_for_reply) {
	raw->rapagnt.reply = (XtPointer) values;
	raw->rapagnt.waiting_for_reply = False;
    }
    else {
	RapAgntObjectInfoCallbackStruct temp;

	temp.reason 	    = RapGetValuesReply;
	temp.sequence_number= seq;
	temp.client 	    = client;
	temp.error_code	    = error_code;
	temp.data	    = (XtPointer) values;
    
	XtCallCallbacks((Widget)raw,
			RapAgntNobjectInfoCallback,
			&temp);
    }
    
    return RapSuccess;
    
} /* RapAgntHandleGetValuesReply */

/************************************************************************/
/*									*/
/* RapAgntHandleCreateHook						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleCreateHook(RapAgntWidget		raw,
			RapAgntClientPtr	client,
			char			*pdata)
{
    RapAgntHookCallbackStruct	temp;
    char 			*bufptr = pdata;    
    char			*type;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleCreateHook","\n");
#endif

    /* Get the object id
     */
    memset(&temp,'\0',sizeof(RapAgntHookCallbackStruct));
    temp.object = (RapAgntObjectPtr) XtMalloc(sizeof(RapAgntObject));
    memset(temp.object,'\0',sizeof(RapAgntObject));
    EXTRACT_OBJECTID(bufptr,False,temp.object->id);
    EXTRACT_STRING(bufptr,False,type);
    temp.data = (XtPointer) type;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleCreateHook",
	  "object_id = %lx, type = %s\n",temp.object->id.id,temp.data);
#endif

    temp.reason	= RapAgntHookNotify;
    temp.client = client;
    temp.hook	= RapCreateHook;
    
    XtCallCallbacks((Widget)raw,
		    RapAgntNhookCallback,
		    &temp);
    
    return RapSuccess;
    
} /* RapAgntHandleCreateHook */

/************************************************************************/
/*									*/
/* RapAgntHandleDestroyHook						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleDestroyHook(RapAgntWidget		raw,
			 RapAgntClientPtr	client,
			 char			*pdata)
{
    RapAgntHookCallbackStruct    temp;
    char 			 *bufptr = pdata;    
    char 			 *type;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleDestroyHook","\n");
#endif

    /* Get the object id
     */
    memset(&temp,'\0',sizeof(RapAgntHookCallbackStruct));
    temp.object = (RapAgntObjectPtr) XtMalloc(sizeof(RapAgntObject));
    memset(temp.object,'\0',sizeof(RapAgntObject));
    EXTRACT_OBJECTID(bufptr,False,temp.object->id);
    EXTRACT_STRING(bufptr,False,type);
    temp.data = (XtPointer) type;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleDestroyHook",
	  "object_id = %lx, type = %s\n",temp.object->id.id,temp.data);
#endif

    temp.reason	= RapAgntHookNotify;
    temp.client = client;
    temp.hook	= RapDestroyHook;
    
    XtCallCallbacks((Widget)raw,
		    RapAgntNhookCallback,
		    &temp);
    
    return RapSuccess;
    
} /* RapAgntHandleDestroyHook */

/************************************************************************/
/*									*/
/* RapAgntHandleChangeHook						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleChangeHook(RapAgntWidget		raw,
			RapAgntClientPtr	client,
			char			*pdata)
{
    RapAgntHookCallbackStruct	temp;
    char 			*bufptr = pdata;    
    char			*type;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleChangeHook","\n");
#endif

    /* Get the object id
     */
    memset(&temp,'\0',sizeof(RapAgntHookCallbackStruct));
    temp.object = (RapAgntObjectPtr) XtMalloc(sizeof(RapAgntObject));
    memset(temp.object,'\0',sizeof(RapAgntObject));
    EXTRACT_OBJECTID(bufptr,False,temp.object->id);
    EXTRACT_STRING(bufptr,False,type);
    temp.data = (XtPointer) type;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleChangeHook",
	  "object_id = %lx, type = %s\n",temp.object->id.id,temp.data);
#endif

    temp.reason = RapAgntHookNotify;
    temp.client	= client;
    temp.hook	= RapChangeHook;
    
    XtCallCallbacks((Widget)raw,
		    RapAgntNhookCallback,
		    &temp);
    
    return RapSuccess;
    
} /* RapAgntHandleChangeHook */

/************************************************************************/
/*									*/
/* RapAgntHandleConfigureHook						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleConfigureHook(RapAgntWidget	raw,
			   RapAgntClientPtr	client,
			   char			*pdata)
{
    RapAgntHookCallbackStruct temp;
    char 		      *bufptr = pdata;    
    char		      *type;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleConfigureHook","\n");
#endif

    /* Get the object id
     */
    memset(&temp,'\0',sizeof(RapAgntHookCallbackStruct));
    temp.object = (RapAgntObjectPtr) XtMalloc(sizeof(RapAgntObject));
    memset(temp.object,'\0',sizeof(RapAgntObject));
    EXTRACT_OBJECTID(bufptr,False,temp.object->id);
    EXTRACT_STRING(bufptr,False,type);
    temp.data = (XtPointer) type;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleConfigureHook",
	  "object_id = %lx, type = %s\n",temp.object->id.id,temp.data);
#endif

    temp.reason = RapAgntHookNotify;
    temp.client = client;
    temp.hook	= RapConfigureHook;
    
    XtCallCallbacks((Widget)raw,
		    RapAgntNhookCallback,
		    &temp);
    
    return RapSuccess;
    
} /* RapAgntHandleConfigureHook */

/************************************************************************/
/*									*/
/* RapAgntHandleGeometryHook						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleGeometryHook(RapAgntWidget		raw,
			  RapAgntClientPtr	client,
			  char			*pdata)
{
    RapAgntHookCallbackStruct 	temp;
    char 			*bufptr = pdata;    
    char			*type;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGeometryHook","\n");
#endif

    /* Get the object id
     */
    memset(&temp,'\0',sizeof(RapAgntHookCallbackStruct));
    temp.object = (RapAgntObjectPtr) XtMalloc(sizeof(RapAgntObject));
    memset(temp.object,'\0',sizeof(RapAgntObject));
    EXTRACT_OBJECTID(bufptr,False,temp.object->id);
    EXTRACT_STRING(bufptr,False,type);
    temp.data = (XtPointer) type;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGeometryHook",
	  "object_id = %lx, type = %s\n",temp.object->id.id,temp.data);
#endif

    temp.reason = RapAgntHookNotify;
    temp.client	= client;
    temp.hook	= RapGeometryHook;
    
    XtCallCallbacks((Widget)raw,
		    RapAgntNhookCallback,
		    &temp);
    
    return RapSuccess;
    
} /* RapAgntHandleGeometryHook */

/************************************************************************/
/*									*/
/* RapAgntHandleActionHook						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleActionHook(RapAgntWidget		raw,
			RapAgntClientPtr	client,
			char			*pdata)
{
    RapAgntHookCallbackStruct		temp;
    RapAgntActionHookCallbackStruct	temp_action;
    char 				*bufptr = pdata;
    char				*tempstr;
    CARD32				pad;
    int					i;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleActionHook","\n");
#endif

    /* Get the object id
     */
    memset(&temp,'\0',sizeof(RapAgntHookCallbackStruct));
    temp.object = (RapAgntObjectPtr) XtMalloc(sizeof(RapAgntObject));
    memset(temp.object,'\0',sizeof(RapAgntObject));
    EXTRACT_OBJECTID(bufptr,False,temp.object->id);
    temp.data = (RapAgntActionHookCallbackData) &temp_action;

    /* Get the action information
     */
    EXTRACT_STRING(bufptr,False,tempstr);
    temp_action.action_name = (XtPointer) tempstr;
    temp_action.event = (XEvent *)bufptr;
    bufptr += sizeof(XEvent);
    bufptr += PAD64(4 + sizeof(XEvent));
    EXTRACT_CARD32(bufptr,False,temp_action.num_params);
    EXTRACT_CARD32(bufptr,False,pad);
    
    temp_action.params = (String *) XtMalloc(temp_action.num_params *
					     sizeof(String));
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleActionHook",
	  "object_id = %lx, action = %s\n",temp.object->id.id,
	  temp_action.action_name);
#endif

    for (i = 0; i < temp_action.num_params; i++) {	
	EXTRACT_STRING(bufptr,False,temp_action.params[i]);
#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapAgntHandleActionHook",
	      "param[%d] = %s\n",i,temp_action.params[i]);
#endif
    }
    
    temp.reason = RapAgntHookNotify;
    temp.client	= client;
    temp.hook	= RapActionHook;
    
    XtCallCallbacks((Widget)raw,
		    RapAgntNhookCallback,
		    &temp);

    return RapSuccess;
    
} /* RapAgntHandleActionHook */

/************************************************************************/
/*									*/
/* RapAgntHandleGetHooksReply						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleGetHooksReply(RapAgntWidget	raw,
			   RapAgntClientPtr	client,
			   CARD16		seq,
			   char			*pdata)
{
    char 	*bufptr = pdata;    
    CARD16	hooks;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetHooksReply","\n");
#endif

    EXTRACT_CARD16(bufptr,False,hooks);
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetHooksReply",
	  "hooks = %x\n",hooks);
#endif

    if (raw->rapagnt.waiting_for_reply) {
	raw->rapagnt.reply = (XtPointer) hooks;
	raw->rapagnt.waiting_for_reply = False;
    }
    else {
	RapAgntHookCallbackStruct 	temp;

	memset(&temp,'\0',sizeof(RapAgntHookCallbackStruct));
 	temp.reason 	    = RapAgntGetHooksReply;
	temp.sequence_number= seq;
	temp.client 	    = client;
	temp.hook	    = hooks;
    
	XtCallCallbacks((Widget)raw,
			RapAgntNhookCallback,
			&temp);
    }
    
    return RapSuccess;
    
} /* RapAgntHandleGetHooksReply */

/************************************************************************/
/*									*/
/* RapAgntHandleGetEventsReply						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleGetEventsReply(RapAgntWidget	raw,
			    RapAgntClientPtr	client,
			    CARD16		seq,
			    char		*pdata)
{
    char 	*bufptr = pdata;    
    CARD32	events[4];
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetEventsReply","\n");
#endif

    EXTRACT_CARD32(bufptr,False,events[0]);
    EXTRACT_CARD32(bufptr,False,events[1]);
    EXTRACT_CARD32(bufptr,False,events[2]);
    EXTRACT_CARD32(bufptr,False,events[3]);

    if (raw->rapagnt.waiting_for_reply) {
	raw->rapagnt.reply = (XtPointer) XtMalloc(4 * sizeof(CARD32));
	memcpy(raw->rapagnt.reply,events,4 * sizeof(CARD32));
	raw->rapagnt.waiting_for_reply = False;
    }
    else {
	RapAgntEventCallbackStruct 	temp;

	memset(&temp,'\0',sizeof(RapAgntEventCallbackStruct));
	temp.reason 	    	= RapAgntGetEventsReply;
	temp.sequence_number 	= seq;
	temp.client 	 	= client;
	temp.data		= (XtPointer) events;
    
	XtCallCallbacks((Widget)raw,
			RapAgntNeventCallback,
			&temp);
    }
    
    return RapSuccess;
    
} /* RapAgntHandleGetEventsReply */

/************************************************************************/
/*									*/
/* RapAgntHandleEventNotify						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleEventNotify(RapAgntWidget		raw,
			 RapAgntClientPtr	client,
			 char			*pdata)
{
    RapAgntEventCallbackStruct	temp;
    xEvent			*wire;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleEventNotify","\n");
#endif

    wire = (xEvent *) pdata;

    temp.reason = RapAgntEventNotify;
    temp.client	= client;
    temp.data	= (XtPointer) wire;
    
    XtCallCallbacks((Widget)raw,
		    RapAgntNeventCallback,
		    &temp);
    
    return RapSuccess;
    
} /* RapAgntHandleEventNotify */

/************************************************************************/
/*									*/
/* RapAgntHandleGetRequestsReply	       				*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleGetRequestsReply(RapAgntWidget	raw,
			      RapAgntClientPtr	client,
			      CARD16		seq,
			      char		*pdata)
{
    char 	*bufptr = pdata;    
    CARD32	requests[8];
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleGetRequestsReply","\n");
#endif

    EXTRACT_CARD32(bufptr,False,requests[0]);
    EXTRACT_CARD32(bufptr,False,requests[1]);
    EXTRACT_CARD32(bufptr,False,requests[2]);
    EXTRACT_CARD32(bufptr,False,requests[3]);
    EXTRACT_CARD32(bufptr,False,requests[4]);
    EXTRACT_CARD32(bufptr,False,requests[5]);
    EXTRACT_CARD32(bufptr,False,requests[6]);
    EXTRACT_CARD32(bufptr,False,requests[7]);

    if (raw->rapagnt.waiting_for_reply) {
	raw->rapagnt.reply = (XtPointer) XtMalloc(8 * sizeof(CARD32));
	memcpy(raw->rapagnt.reply,requests,8 * sizeof(CARD32));
	raw->rapagnt.waiting_for_reply = False;
    }
    else {
	RapAgntRequestCallbackStruct 	temp;

	memset(&temp,'\0',sizeof(RapAgntEventCallbackStruct));
	temp.reason 	     	= RapAgntGetRequestsReply;
	temp.sequence_number	= seq;
	temp.client 		= client;
	temp.data		= (XtPointer) requests;

	XtCallCallbacks((Widget)raw,
			RapAgntNrequestCallback,
			&temp);
    }
    
    return RapSuccess;
    
} /* RapAgntHandleGetRequestsReply */

/************************************************************************/
/*									*/
/* RapAgntHandleRequestNotify						*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleRequestNotify(RapAgntWidget	raw,
			   RapAgntClientPtr	client,
			   char			*pdata)
{
    RapAgntEventCallbackStruct	temp;
    xReq			*req;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleRequestNotify","\n");
#endif

    req = (xReq *) pdata;

    temp.reason = RapAgntRequestNotify;
    temp.client = client;
    temp.data	= (XtPointer) req;
    
    XtCallCallbacks((Widget)raw,
		    RapAgntNrequestCallback,
		    &temp);

    return RapSuccess;
    
} /* RapAgntHandleRequestNotify */

/************************************************************************/
/*									*/
/* RapAgntHandleLocateObjectReply					*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHandleLocateObjectReply(RapAgntWidget	raw,
			       RapAgntClientPtr	client,
			       CARD16		seq,
			       char		*pdata)
{
    RapAgntObjectPtr		    object;
    RapObjectID			    parent_id;
    char			    *bufptr = pdata;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntHandleLocateObjectReply","\n");
#endif

    /* Get the object
     */
    object = (RapAgntObjectPtr) XtMalloc(sizeof(RapAgntObject));
    memset(object,'\0',sizeof(RapAgntObject));
    ExtractObject(&bufptr,object,&parent_id,False);
    object->client = client;
    
    if (raw->rapagnt.waiting_for_reply) {
	raw->rapagnt.reply = (XtPointer) object;
	raw->rapagnt.waiting_for_reply = False;
    }
    else {
	RapAgntObjectInfoCallbackStruct temp;

	temp.reason 	    	= RapLocateObjectReply;
	temp.sequence_number    = seq;
	temp.client 	    	= client;
	temp.error_code 	= RapNoError;
	temp.data		= (XtPointer) object;
    	
	XtCallCallbacks((Widget)raw,
			RapAgntNobjectInfoCallback,
			&temp);
    }

    return RapSuccess;
    
} /* RapAgntHandleLocateObjectReply */
