/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
//
//      Copyright 1992, 1993, 1994, 1995
//      Georgia Tech Research Corporation
//      All rights reserved.
//
*/
/*
* $Log: RapAgnt-out.c,v $
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*                                                                      */
/* RapAgnt-out.c  	                                                */
/*                                                                      */
/*      Prepares all outgoing RAP messages from the agent.		*/
/*                                                                      */
/************************************************************************/
#include <X11/Xmd.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#include "RapAgntP.h"

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEmsg.h>
#include <X11/ICE/ICEproto.h>
#include "Rap.h"
#include "RapProto.h"

/************************************************************************/
/*									*/
/* RapAgntHello								*/
/*									*/
/*	When the ICE subprotocol setup is complete, the agent will send	*/
/*	a RapHelloRequest to the client to set up some additional	*/
/*	handshaking.  This routine sends the request.			*/
/*									*/
/************************************************************************/
RapStatus
RapAgntHello(RapAgntWidget	raw,
	     IceConn		ice_conn)
{
    IceSimpleMessage(ice_conn,
		     raw->rapagnt.major_opcode,
		     RapHelloRequest);
    IceFlush(ice_conn);
    return RapSuccess;
    
} /* RapAgntHello */

/************************************************************************/
/*									*/
/* SendRequest								*/
/*									*/
/************************************************************************/
static CARD16
SendRequest(RapAgntWidget	raw,
	    RapAgntClientPtr	client,
	    CARD8		opcode,
	    int			num_bytes,
	    char		*pdata)
{
    RapMessageHeader *header;
    CARD16	     seq;
    int 	     padding = 0;
    int		     total_bytes;

    if (!client->ice_conn)
	return 0;

    seq = raw->rapagnt.last_sequence_num++;
    if (client->ice_conn == (IceConn) RAP_AGENT_SELF_CONNECT) {
	RapClntDispatchMsg(raw->rapagnt.ri,
			   opcode,
			   seq,
			   pdata);
    }
    else {
	IceGetHeader(client->ice_conn,
		     raw->rapagnt.major_opcode,
		     opcode,
		     SIZEOF(RapMessageHeader),
		     RapMessageHeader,
		     header);
	header->sequenceNumber = seq;
	
	if (pdata) {
	    padding = PAD64(num_bytes);
	    total_bytes = num_bytes + padding;
	    header->length += WORD64COUNT(total_bytes);	
	    IceWriteData(client->ice_conn,num_bytes,pdata);
	    if (padding) IceWritePad(client->ice_conn,padding);
	}

	IceFlush(client->ice_conn);
    }

    return seq;
    
} /* SendRequest */

/************************************************************************/
/*									*/
/* RapAgntQueryTree							*/
/*									*/
/************************************************************************/
CARD16
RapAgntQueryTree(Widget			w,
		 RapAgntClientPtr	client,
		 Boolean		get_resources,
		 RapAgntShellList	*shell_list)
{
    RapAgntWidget    raw = (RapAgntWidget) w;
    CARD16	     seq;

    if (shell_list != NULL)
	raw->rapagnt.waiting_for_reply = True;

    /* Send the request
     */
    seq = SendRequest(raw,
		      client,
		      get_resources ? 
		      RapFullQueryTreeRequest : 
		      RapQueryTreeRequest,
		      0,NULL);
    
    /* Wait for a reply if necessary
     */
    if (shell_list != NULL) {
	RapAgntShellList *reply;

	RapAgntWaitForReply(raw,
			    client,
			    seq,
			    get_resources ?
			    RapFullQueryTreeReply : 
			    RapQueryTreeReply);

	reply = (RapAgntShellList *) raw->rapagnt.reply;	
	if (reply != NULL) {
	    shell_list->num_shells  = reply->num_shells;
	    shell_list->shells	= reply->shells;
	}
	else {
	    shell_list->num_shells = 0;
	    shell_list->shells	   = NULL;
	}

	XtFree((char *)raw->rapagnt.reply);
    }

    return seq;

} /* RapAgntQueryTree */

/************************************************************************/
/*									*/
/* RapAgntGetResources							*/
/*									*/
/************************************************************************/
CARD16
RapAgntGetResources(Widget		w,
		    RapAgntClientPtr	client,
		    RapObjectID		*object_id,
		    RapAgntResourceList	*res_list)
{
    RapAgntWidget raw = (RapAgntWidget) w;
    CARD16	  seq;

    if (res_list != NULL)
	raw->rapagnt.waiting_for_reply = True;

    /* Send the request
     */
    seq = SendRequest(raw,client,RapGetResourcesRequest,
		      sizeof(RapObjectID),(char *)object_id);
    
    /* Wait for a reply if necessary
     */
    if (res_list != NULL) {
	RapAgntResourceList *reply = (RapAgntResourceList *) raw->rapagnt.reply;

	RapAgntWaitForReply(raw,client,seq,RapGetResourcesReply);

	reply = (RapAgntResourceList *) raw->rapagnt.reply;	
	if (reply != NULL) {
	    res_list->num_resources = reply->num_resources;
	    res_list->resources	    = reply->resources;
	}
	else {
	    res_list->num_resources = 0;
	    res_list->resources	    = NULL;
	}
	
	XtFree((char *)raw->rapagnt.reply);
    }

    return seq;

} /* RapAgntGetResources */

#ifdef GC_SUPPORT
/************************************************************************/
/*									*/
/* RapAgntGetGCValues							*/
/*									*/
/************************************************************************/
CARD16
RapAgntGetGCValues(Widget			w,
		 RapAgntClientPtr	client,
		   int numGCs,
		   GContext *GCs,
		   RapGCValuesList *reply)
{
  RapAgntWidget raw = (RapAgntWidget)w;
    int totalSizeInBytes = 0;
    int padding = 0;
    register int i;
    char *scratch, *bufptr;
     CARD16	  seq;

#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapAgntGetGCValues",
	      "numGCs = %d\n", numGCs);
#endif

    if (reply != NULL)
	raw->rapagnt.waiting_for_reply = True;
    else
      numGCs = 0;  /* Calcs should depend on GCValuesPtr only /*

    if (client == NULL)
        return 0;
 
    if (numGCs < 0)
        return 0;
 
    /* compute the size of the message */
    totalSizeInBytes += 4;                              /* size of list */
    totalSizeInBytes += numGCs * sizeof(CARD32);
 
    padding = PAD64(totalSizeInBytes);
    totalSizeInBytes += padding;
 
    /* pad to 64 boundary.  Is this necessary? */
    /*
    while (padding--) {
        char zero = '\0';
        IceWriteData(client->iceConn, 1, &zero);
    }
*/
 
    /* Create and send */
    scratch = bufptr = XtMalloc(totalSizeInBytes);

    STORE_CARD32(bufptr, numGCs);
    for(i = 0; i<numGCs; i++)
      STORE_CARD32(bufptr, GCs[i]);

#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapAgntGetGCValues",
	      "totalSizeInBytes = %d\n", totalSizeInBytes);
#endif

    seq = SendRequest(raw,client,RapGetGCValuesRequest, totalSizeInBytes,scratch);
    XtFree(scratch);

    /* Wait for a reply if necessary
     */
    if (reply != NULL) {
      RapGCValuesList *tmpReply = NULL;
	RapAgntWaitForReply(raw,client,seq,RapGetGCValuesReply);
    	tmpReply = (RapGCValuesList*) raw->rapagnt.reply;
	reply->numGCs = (tmpReply==NULL)? 0 : tmpReply->numGCs;
	reply->GCValues = (tmpReply==NULL)? NULL : tmpReply->GCValues;
	  
    }

    return seq;
} /* RapAgntGetGCValuesRequest */

#endif /* GC_SUPPORT */

/************************************************************************/
/*									*/
/* RapAgntGetValues							*/
/*									*/
/************************************************************************/
CARD16
RapAgntGetValues(Widget			w,
		 RapAgntClientPtr	client,
		 RapObjectID		*object_id,
		 char			**resource_names,
		 int			num_resources,
		 RapValuePtr		*values)
{
    RapAgntWidget raw = (RapAgntWidget) w;
    int 	  total_bytes = 0;
    char	  *scratch,*bufptr;
    CARD32	  pad = 0;
    int		  i;
    CARD16	  seq;

    if (values != NULL)
	raw->rapagnt.waiting_for_reply = True;

    /* Compute the size of the request
     */
    total_bytes += sizeof(RapObjectID); /* ObjectID      */
    total_bytes += 4;		     	/* NumStrings    */
    total_bytes += 4; 		     	/* addtl padding */
    for (i = 0; i < num_resources; i++)
	total_bytes += STRING_BYTES(strlen(resource_names[i]));

    /* Create the request and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    STORE_OBJECTID(bufptr,*object_id);
    STORE_CARD32(bufptr,num_resources);
    STORE_CARD32(bufptr,pad);
    for (i = 0; i < num_resources; i++)
	STORE_STRING(bufptr,strlen(resource_names[i]),resource_names[i]);
    
    seq = SendRequest(raw,client,RapGetValuesRequest,total_bytes,scratch);
    XtFree(scratch);

    /* Wait for a reply if necessary
     */
    if (values != NULL) {
	RapAgntWaitForReply(raw,client,seq,RapGetValuesReply);
    	*values = (RapValuePtr) raw->rapagnt.reply;
    }

    return seq;

} /* RapAgntGetValues */

/************************************************************************/
/*									*/
/* RapAgntSetValues							*/
/*									*/
/************************************************************************/
CARD16
RapAgntSetValues(Widget			w,
		 RapAgntClientPtr	client,
		 RapObjectID		*object_id,
		 RapArgList		args,
		 int			num_args)
{
    RapAgntWidget raw = (RapAgntWidget) w;
    int 	  total_bytes = 0;
    char	  *scratch,*bufptr;
    CARD32	  pad = 0;
    int		  i;
    CARD16	  seq;
    
    /* Compute the size of the request
     */
    total_bytes += sizeof(RapObjectID); /* ObjectID      */
    total_bytes += 4;		     /* Num arguments */
    total_bytes += 4; 		     /* addtl padding */
    for (i = 0; i < num_args; i++) {
	total_bytes += STRING_BYTES(strlen(args[i].name));
	total_bytes += STRING_BYTES(strlen(args[i].value->type));
	total_bytes += STRING_BYTES(args[i].value->size);
    }
    
    /* Create the request and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    STORE_OBJECTID(bufptr,*object_id);
    STORE_CARD32(bufptr,num_args);
    STORE_CARD32(bufptr,pad);
    
    for (i = 0; i < num_args; i++) {
	STORE_STRING(bufptr,strlen(args[i].name),args[i].name);
	STORE_STRING(bufptr,strlen(args[i].value->type),args[i].value->type);
	STORE_BYTES(bufptr,args[i].value->size,args[i].value->value);
    }

    seq = SendRequest(raw,client,RapSetValuesRequest,total_bytes,scratch);
    XtFree(scratch);

    return seq;

} /* RapAgntSetValues */

/************************************************************************/
/*									*/
/* RapAgntGetHooks							*/
/*									*/
/************************************************************************/
CARD16
RapAgntGetHooks(Widget			w,
		RapAgntClientPtr	client,
		CARD16			*hooks)
{
    RapAgntWidget    raw = (RapAgntWidget) w;
    CARD16	     seq;
    
    if (hooks != NULL)
	raw->rapagnt.waiting_for_reply = True;

    /* Send the request
     */
    seq = SendRequest(raw,
		      client,
		      RapGetHooksRequest,
		      0,NULL);
    
    /* Wait for a reply if necessary
     */
    if (hooks != NULL) {
	RapAgntWaitForReply(raw,client,seq,RapGetHooksReply);
    	*hooks = (CARD16) raw->rapagnt.reply;
    }

    return seq;

} /* RapAgntGetHooks */

/************************************************************************/
/*									*/
/* RapAgntSelectHooks							*/
/*									*/
/************************************************************************/
CARD16
RapAgntSelectHooks(Widget		w,
		   RapAgntClientPtr	client,
		   CARD16		hooks,
		   CARD16		mask)
{
    RapAgntWidget raw = (RapAgntWidget) w;
    int 	  total_bytes = 0;
    char	  *scratch,*bufptr;
    CARD32	  pad = 0;
    CARD16	  seq;
    
    /* Compute the size of the request 
     */
    total_bytes += sizeof(CARD16);   /* hooks */
    total_bytes += sizeof(CARD16);   /* mask  */

    /* Create the request and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    STORE_CARD16(bufptr,hooks);
    STORE_CARD16(bufptr,mask);

    seq = SendRequest(raw,client,RapSelectHooksRequest,total_bytes,scratch);
    XtFree(scratch);

    return seq;

} /* RapAgntSelectHooks */


/************************************************************************/
/*									*/
/* RapAgntGetEvents							*/
/*									*/
/************************************************************************/
CARD16
RapAgntGetEvents(Widget			w,
		 RapAgntClientPtr	client,
		 CARD32			events[4])
{
    RapAgntWidget    raw = (RapAgntWidget) w;
    CARD16	     seq;
    CARD32	     *ev_reply;
    
    if (events != NULL)
	raw->rapagnt.waiting_for_reply = True;

    /* Send the request
     */
    seq = SendRequest(raw,
		      client,
		      RapGetEventsRequest,
		      0,NULL);
    
    /* Wait for a reply if necessary
     */
    if (events != NULL) {
	RapAgntWaitForReply(raw,client,seq,RapGetEventsReply);
    	ev_reply = (CARD32 *) raw->rapagnt.reply;
	memcpy(events,ev_reply,4 * sizeof(CARD32));
	XtFree((char *)ev_reply);
    }
    
    return seq;

} /* RapAgntGetEvents */

/************************************************************************/
/*									*/
/* RapAgntSelectEvents							*/
/*									*/
/************************************************************************/
CARD16
RapAgntSelectEvents(Widget		w,
		    RapAgntClientPtr	client,
		    CARD32		events[4],
		    CARD32		mask[4])
{
    RapAgntWidget raw = (RapAgntWidget) w;
    int 	  i,total_bytes = 0;
    char	  *scratch,*bufptr;
    CARD32	  pad = 0;
    CARD16	  seq;
    
    /* Compute the size of the request 
     */
    total_bytes += (4 * sizeof(CARD32));   /* events */
    total_bytes += (4 * sizeof(CARD32));   /* mask   */

    /* Create the request and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    for (i = 0; i < 4; i++) {
	STORE_CARD32(bufptr,events[i]);
	STORE_CARD32(bufptr,mask[i]);
#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapAgentSelectEvents",
	      "events[%d] = %x, mask[%d] = %x\n",i,events[i],i,mask[i]);
#endif
    }
    
    seq = SendRequest(raw,client,RapSelectEventsRequest,total_bytes,scratch);
    XtFree(scratch);

    return seq;

} /* RapAgntSelectEvents */


/************************************************************************/
/*									*/
/* RapAgntGetRequests							*/
/*									*/
/************************************************************************/
CARD16
RapAgntGetRequests(Widget		w,
		   RapAgntClientPtr	client,
		   CARD32		requests[8])
{
    RapAgntWidget    raw = (RapAgntWidget) w;
    CARD16	     seq;
    CARD32	     *req_reply;
    
    if (requests != NULL)
	raw->rapagnt.waiting_for_reply = True;

    /* Send the request
     */
    seq = SendRequest(raw,
		      client,
		      RapGetRequestsRequest,
		      0,NULL);
    
    if (requests != NULL) {
	RapAgntWaitForReply(raw,client,seq,RapGetRequestsReply);
    	req_reply = (CARD32 *) raw->rapagnt.reply;
	memcpy(requests,req_reply,8 * sizeof(CARD32));
	XtFree((char *)req_reply);
    }
    
    return seq;

} /* RapAgntGetRequests */

/************************************************************************/
/*									*/
/* RapAgntSelectRequests						*/
/*									*/
/************************************************************************/
CARD16
RapAgntSelectRequests(Widget		w,
		      RapAgntClientPtr	client,
		      CARD32		requests[8],
		      CARD32		mask[8])
{
    RapAgntWidget raw = (RapAgntWidget) w;
    int 	  i,total_bytes = 0;
    char	  *scratch,*bufptr;
    CARD32	  pad = 0;
    CARD16	  seq;
    
    /* Compute the size of the request 
     */
    total_bytes += (8 * sizeof(CARD32));   /* requests */
    total_bytes += (8 * sizeof(CARD32));   /* mask     */

    /* Create the request and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    for (i = 0; i < 8; i++) {
	STORE_CARD32(bufptr,requests[i]);
	STORE_CARD32(bufptr,mask[i]);
#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapAgentSelectRequests",
	      "requests[%d] = %x, mask[%d] = %x\n",i,requests[i],i,mask[i]);
#endif
    }
    
    seq = SendRequest(raw,client,RapSelectRequestsRequest,total_bytes,scratch);
    XtFree(scratch);

    return seq;

} /* RapAgntSelectRequests */


/************************************************************************/
/*									*/
/* RapAgntLocateObject							*/
/*									*/
/************************************************************************/
CARD16
RapAgntLocateObject(Widget		w,
		    RapAgntClientPtr	client,
		    RapObjectID		*object_id,
		    Position		xpos,
		    Position		ypos,
		    RapAgntObjectPtr	*object)
{
    RapAgntWidget raw = (RapAgntWidget) w;
    int 	  total_bytes = 0;
    char	  *scratch,*bufptr;
    CARD32	  pad = 0;
    CARD16	  seq;
    
    if (object != NULL)
	raw->rapagnt.waiting_for_reply = True;

    /* Compute the size of the request
     */
    total_bytes += sizeof(RapObjectID);
    total_bytes += sizeof(CARD32);
    total_bytes += sizeof(CARD32);

    /* Create the request and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    STORE_OBJECTID(bufptr,*object_id);
    STORE_CARD32(bufptr,xpos);
    STORE_CARD32(bufptr,ypos);

    seq = SendRequest(raw,client,RapLocateObjectRequest,total_bytes,scratch);
    XtFree(scratch);

    /* Wait for a reply if necessary
     */
    if (object != NULL) {
	RapAgntWaitForReply(raw,client,seq,RapLocateObjectReply);
    	*object = (RapAgntObjectPtr) raw->rapagnt.reply;
    }
    
    return seq;

} /* RapAgntLocateObject */

