#define RAPAGNT_C
/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
* $Log: RapAgnt.c,v $
* Revision 0.2  1997/11/20 21:57:38  sellden
* Added Marty's pointer-copy fix to ProcessMsg
*
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*									*/
/* RapAgnt.c								*/
/*									*/
/*      Rap is a simple ICE subprotocol that uses Xt hooks to share	*/
/*	share client information with external agents.	This file	*/
/*	handles the RapAgnt object information for agents.		*/
/*									*/
/************************************************************************/
#include <sys/time.h>
#include <X11/Xmd.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEmsg.h>
#include <X11/ICE/ICEproto.h>

#include "Xtea.h"
#include "Rap.h"
#include "RapProto.h"

#include "RapAgntP.h"

RapAgntGetProc _RapAgntGet = NULL;

/************************************************************************/
/*									*/
/* RapAgnt Resources							*/
/*									*/
/************************************************************************/
static XtResource resources[] = {
    { RapAgntNclientCallback,XtCCallback,XtRCallback,sizeof(XtCallbackList),
      XtOffset(RapAgntWidget,rapagnt.client_callback),
      XtRCallback, (XtPointer) NULL },
    { RapAgntNobjectInfoCallback,XtCCallback,XtRCallback,sizeof(XtCallbackList),
      XtOffset(RapAgntWidget,rapagnt.object_info_callback),
      XtRCallback, (XtPointer) NULL },
    { RapAgntNhookCallback,XtCCallback,XtRCallback,sizeof(XtCallbackList),
      XtOffset(RapAgntWidget,rapagnt.hook_callback),
      XtRCallback, (XtPointer) NULL },
    { RapAgntNeventCallback,XtCCallback,XtRCallback,sizeof(XtCallbackList),
      XtOffset(RapAgntWidget,rapagnt.event_callback),
      XtRCallback, (XtPointer) NULL },
    { RapAgntNrequestCallback,XtCCallback,XtRCallback,sizeof(XtCallbackList),
      XtOffset(RapAgntWidget,rapagnt.request_callback),
      XtRCallback, (XtPointer) NULL },
    { RapAgntNknownClients,RapAgntCKnownClients,XtRPointer,sizeof(XtPointer),
      XtOffset(RapAgntWidget,rapagnt.known_clients),
      XtRImmediate, (XtPointer) NULL },
    { RapAgntNautoInitiate,XtCBoolean,XtRBoolean,sizeof(Boolean),
      XtOffset(RapAgntWidget,rapagnt.auto_initiate),
      XtRImmediate, (XtPointer) False },
};

/************************************************************************/
/*									*/
/* RapAgnt Class Record Declaration					*/
/*									*/
/************************************************************************/
/* Forward Declarations
 */
static void 	Initialize();
static Boolean	SetValues();
static void	Destroy();
static void     DispatchMsg(RapAgntWidget 	raw,
			    IceConn		ice_conn,
			    CARD8		opcode,
			    CARD16		seq,
			    char		*pdata);

externaldef(rapagntclassrec) RapAgntClassRec rapAgntClassRec = {
    { /* core class record      */
        /* superclass           */	(WidgetClass) &coreClassRec,
        /* class_name           */      "RapAgnt",
        /* widget_size          */      sizeof(RapAgntRec),
        /* class_initialize     */      NULL,
        /* class init part proc */      NULL,
        /* class_inited         */      False,
        /* initialize           */      (XtInitProc)Initialize,
        /* initialize hook      */      NULL,
        /* realize              */      XtInheritRealize,
        /* actions              */      NULL,
        /* num_actions          */      0,
        /* resources            */      resources,
        /* num_resources        */      XtNumber(resources),
        /* xrm_class            */      NULLQUARK,
        /* compress_motion      */      True,
        /* compress_exposure    */      True,
        /* compress-enter/exit  */      True,
        /* visible_interest     */      False,
        /* destroy              */      (XtWidgetProc)Destroy,
        /* resize               */      XtInheritResize,
        /* expose               */      XtInheritExpose,
        /* set_values           */      (XtSetValuesFunc)SetValues,
        /* set_values hook      */      NULL,
        /* set_values almost    */      XtInheritSetValuesAlmost,
        /* widget get val hook  */      NULL,
        /* accept_focus         */      XtInheritAcceptFocus,
        /* version              */      XtVersion,
        /* callback offset      */      NULL,
        /* default translations */      XtInheritTranslations,
        /* Query Geometry proc  */      XtInheritQueryGeometry,
        /* disp accelerator     */      NULL,
        /* extension            */      NULL,
    },
    { /* RapAgnt class record*/
        /* extension            */      NULL,
    }
};

externaldef(rapagntwidgetclass) WidgetClass rapAgntWidgetClass
                = (WidgetClass) &rapAgntClassRec;


/************************************************************************/
/*									*/
/* AddClient								*/
/*									*/
/*	Add the client to the list of known clients.			*/
/*									*/
/************************************************************************/
RapAgntClientPtr
AddClient(RapAgntWidget	raw,
	  CARD32	window,
	  IceConn	ice_conn,
	  XtPointer	client_data)
{
    RapAgntClientPtr		new,current,prev;
    RapAgntClientCallbackStruct temp;

    /* Make sure we have all the info we need.
     */
    if (!window)
	return NULL;

    /* Create and initialize a table node for the new initializer.
     */
    if ((new = XtNew(RapAgntClient)) == NULL)
	return NULL;
    new->next 		= NULL;
    new->client_data	= client_data;

    /* Install it in the table.
     */
    if (raw->rapagnt.known_clients == NULL) {
	raw->rapagnt.known_clients = new;
    }
    else {
	/* Search the list for a matching node.  If the client
	 * window is already in the table, it means we've just
	 * formed an ICE connection with the client.  If the
	 * client is not already in the table, it means we're
	 * going to try to initiate a connection using Xtea.
	 */
	current = raw->rapagnt.known_clients;
	while ((current != NULL) &&
	       (current->window != window)) {
	    prev = current;
	    current = current->next;
	}

	if (current == NULL)
	    prev->next = new;
	else {
	    XtFree((char *)new);
	    new = current;
	}
    }

    new->window   	= window;
    new->ice_conn 	= ice_conn;    

    temp.client = new;   
    if (new->ice_conn)
	temp.reason = RapAgntClientConnect;		
    else
	temp.reason = RapAgntNewClient;

    XtCallCallbacks((Widget)raw,
		    RapAgntNclientCallback,
		    &temp);

    return new;

} /* AddClient */

/************************************************************************/
/*									*/
/* RapAgntFindClient							*/
/*									*/
/*	Find the client in the list of known clients.  This is done by	*/
/*	a match of either the ice_conn ID *or* the window ID.  If the	*/
/*      either are NULL, they are not used in the search.		*/
/*									*/
/************************************************************************/
RapAgntClientPtr
RapAgntFindClient(Widget 	w,
		  CARD32 	window,
		  IceConn	ice_conn)
{
    RapAgntWidget	raw = (RapAgntWidget) w;
    RapAgntClientPtr 	current = raw->rapagnt.known_clients;
    RapAgntClientPtr 	prev = NULL;

    while (current) {
	if (((window != 0) && (current->window == window)) ||
	    ((ice_conn != NULL) && (current->ice_conn == ice_conn)))
	    break;
	prev = current;
	current = current->next;
    }

    return current;

} /* RapAgntFindClient */

/************************************************************************/
/*									*/
/* RemoveClient								*/
/*									*/
/*	Remove the client from the list of known clients.		*/
/*									*/
/************************************************************************/
static Boolean
RemoveClient(RapAgntWidget	raw,
	     CARD32 		window,
	     IceConn		ice_conn,
	     int		reason)
{
    RapAgntClientPtr 	   	current = raw->rapagnt.known_clients;
    RapAgntClientPtr 	   	prev = NULL;
    RapAgntClientCallbackStruct temp;
    Boolean 			removed = False;
    
    while (current != NULL) {
	if (((current->window == window) || (window == NULL)) &&
	    (current->ice_conn == ice_conn)) {
	    if (prev) 
		prev->next = current->next;
	    else 
	        raw->rapagnt.known_clients = current->next;

	    temp.reason = reason;
	    temp.client	= current;
    	    XtCallCallbacks((Widget)raw,
			    RapAgntNclientCallback,
			    &temp);

	    XtFree((char *)current);
	    removed = True;
	    break;
	}
	prev = current;
	current = current->next;
    }

    return removed;

} /* RemoveClient */

/************************************************************************/
/*									*/
/* RapAgntCheckMessage							*/
/*									*/
/************************************************************************/
static void 
RapAgntCheckMessages(Widget	w,
		     XtPointer 	client_data, 
		     XEvent 	*event, 
		     Boolean 	*cont)
{
    RapAgntWidget	raw = (RapAgntWidget) w;
    Display 		*dpy = XtDisplay(w);
    XClientMessageEvent *cme = (XClientMessageEvent *) event;
    
    /* Make sure this is a ClientMessage event before doing anything.
     */
    if (event->type != ClientMessage) 
	return;

    /* If the message type is not an ICE_PROTOCOLS message,
     * then we know not to process this event any further.
     */
    if ((cme->message_type != 
	 XInternAtom(dpy,ICE_INITIATE_FAILED,False)) ||
	(cme->format != 32)) 
        return;

    RemoveClient(raw,cme->data.l[2],NULL,RapAgntInitiateFailed);
    /*cme->data.l[3] = reason;*/
    
} /* RapAgntCheckMessyages */

/************************************************************************/
/*									*/
/* AddIceConn								*/
/*									*/
/*	Add the ice connection to the list of known connections.	*/
/*									*/
/************************************************************************/
RapAgntIceConnPtr
AddIceConn(RapAgntWidget	raw,
	   IceConn		ice_conn)
{
    RapAgntIceConnPtr new,current,prev;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"AddIceConn","\n");
#endif

    /* Make sure we have all the info we need.
     */
    if (!ice_conn)
	return NULL;

    /* Create and initialize a table node for the new initializer.
     */
    if ((new = XtNew(RapAgntIceConn)) == NULL)
	return NULL;
    new->next = NULL;

    /* Install it in the table.
     */
    if (raw->rapagnt.known_ice_conn == NULL) {
	raw->rapagnt.known_ice_conn = new;
    }
    else {
	/* Search the list for a matching node.  If the connection
	 * is already in the table, just return it.  Otherwise, add
	 * a the new connection and return it.
	 */
	current = raw->rapagnt.known_ice_conn;
	while ((current != NULL) &&
	       (current->ice_conn != ice_conn)) {
	    prev = current;
	    current = current->next;
	}
	if (current == NULL)
	    prev->next = new;
	else {
	    XtFree((char *)new);
	    new = current;
	}
    }

    new->w	  = (Widget) raw;
    new->ice_conn = ice_conn;
    
    return new;

} /* AddIceConn */

/************************************************************************/
/*									*/
/* RemoveIceConn							*/
/*									*/
/*	Remove the ICE connection and all associated clients from the	*/
/*	list of known ICE connections.					*/
/*									*/
/************************************************************************/
static Boolean
RemoveIceConn(RapAgntWidget	raw,
	      IceConn		ice_conn)
{
    RapAgntIceConnPtr current = raw->rapagnt.known_ice_conn;
    RapAgntIceConnPtr prev = NULL;
    Boolean 	      client_removed = True, removed = False;

    while (current) {
	if (current->ice_conn == ice_conn) {
	    if (prev) 
		prev->next = current->next;
	    else 
	        raw->rapagnt.known_ice_conn = current->next;

	    while (client_removed)
		client_removed = RemoveClient(raw,NULL,ice_conn,
					      RapAgntClientDisconnect);

	    XtRemoveInput(current->input_id);
	    XtFree((char *)current);
	    removed = True;
	    break;
	}

	prev = current;
	current = current->next;
    }

    return removed;

} /* RemoveIceConn */

/************************************************************************/
/*									*/
/* RapAgntInitiateConnection						*/
/*									*/
/************************************************************************/
Boolean 
RapAgntInitiateConnection(RapAgntWidget raw, Window win)
{
    Display	*dpy   = XtDisplay(raw);
    Window 	client = XmuClientWindow(dpy,win);
    
    if (client == None)
	return False;
    
    /* Make sure this client will initiate a connection.
     */
    if (!XteaWillInitiate(dpy,client,raw->rapagnt.init_atom))
	return False;

    /* Do not initiate ICE connections with self, but set self
     * up for local transport.
     */
    if (client == raw->rapagnt.network_ids_window) {
	raw->rapagnt.ri = (RapInfo)XtMalloc(sizeof(RapInfoRec));
	memset(raw->rapagnt.ri,'\0',sizeof(RapInfoRec));
	raw->rapagnt.ri->top_level	= (Widget) raw;
	raw->rapagnt.ri->ice_conn 	= (IceConn) RAP_AGENT_SELF_CONNECT;
	raw->rapagnt.ri->dispatch_proc	= (void (*)())DispatchMsg;
	raw->rapagnt.ri->private	= (XtPointer) raw;
	RapClntAddAgent(dpy,raw->rapagnt.ri);
	AddClient(raw,(CARD32)client,(IceConn)RAP_AGENT_SELF_CONNECT,NULL);
	return True;
    }
    
    /* Do not initiate with clients we already know about.
     * Otherwise, put the client in the list of known clients
     * and initiate a connection.
     */
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntInitiateConnection","window=%x\n", client);
#endif
    
    if (RapAgntFindClient((Widget)raw,(CARD32)client,NULL) != NULL)
	return False;
    else
	AddClient(raw,(CARD32)client,NULL,NULL);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntInitiateConnection",
	  "initiating connection\n",client);
#endif
    
    XteaInitiateConnection(dpy,client,raw->rapagnt.network_ids_window,
			   raw->rapagnt.init_atom);
    return True;
    
} /* RapAgntInitiateConnection */

/************************************************************************/
/*									*/
/* NewWindow								*/
/*									*/
/************************************************************************/
static Boolean
NewWindow(XEvent *event)
{
    RapAgntWidget raw = (RapAgntWidget) _RapAgntGet();
    
    if ((event->type == CreateNotify) &&
        (event->xcreatewindow.parent == DefaultRootWindow(XtDisplay(raw))))
	RapAgntInitiateConnection(raw,event->xcreatewindow.window);
    
    return True;

} /* NewWindow */

/************************************************************************/
/*									*/
/* RapAgntContactExistingClients					*/
/*									*/
/************************************************************************/
void 
RapAgntContactExistingClients(Widget w)
{    
    RapAgntWidget	raw = (RapAgntWidget) w;
    int			screen,child,numchildren;
    Window 		dummy, *children = NULL;
    Display		*dpy = XtDisplay(w);

    for (screen = 0; screen < ScreenCount(dpy); screen++) {
	if (XQueryTree(dpy, 
		       RootWindow(dpy,screen),
		       &dummy, &dummy, 
		       &children, (unsigned int *)&numchildren)) {
	    for (child = 0; child < numchildren; child++) {
		RapAgntInitiateConnection(raw,children[child]);
	    }
	    XFree(children);
        }
    }

    return;
    
} /* RapAgntContactExistingClients */

/************************************************************************/
/*									*/
/* ProtoAuthorize							*/
/*									*/
/*	Disables authentication.					*/
/*									*/
/************************************************************************/
Bool
ProtoAuthorize(char *host_name)
{
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"ProtoAuthorize","%s\n",host_name);
#endif
    return True;

} /* ProtoAuthorize */

/************************************************************************/
/*									*/
/* ProtoSetup								*/
/*									*/
/* 	Called after authentication succeeds and before the Protocol	*/
/*	Reply message is sent.						*/
/*									*/
/************************************************************************/
static Status 
ProtoSetup(IceConn	ice_conn,
	   int		major_version,
	   int		minor_version,
	   char		*vendor_str,
	   char        	*release_str,
	   IcePointer	*client_data_ret,
	   char		**failure_reason_ret)
{
    RapAgntWidget raw = (RapAgntWidget) _RapAgntGet();

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"ProtoSetup",
	  "Version: %1d.%1d, Vendor: %s, Release: %s\n",
	   major_version,minor_version,vendor_str,release_str);
#endif

    *client_data_ret = (IcePointer) raw;

    return 1; /* 0 == failure */
    
} /* ProtoSetup */

/************************************************************************/
/*									*/
/* ProtoActivate							*/
/*									*/
/* 	Optional callback to generate a message immediately following	*/
/*	a Protocol Reply.						*/
/*									*/
/************************************************************************/
static void 
ProtoActivate(IceConn		ice_conn,
	      IcePointer	client_data)
{
    RapAgntWidget raw = (RapAgntWidget) client_data;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"ProtoActivate","\n");
#endif

    /* Now that the protocol is active, tell the client to send us 
     * its client window.  The reply from the client will give us both 
     * the ice_conn ID and the client's window ID, via a RapHello.  This 
     * will allowing us to uniquely identify it and add it to our 
     * list of known clients.
     */
    RapAgntHello(raw,ice_conn);
        
} /* ProtoActivate */

/************************************************************************/
/*									*/
/* IOErrorProc								*/
/*									*/
/* 	Called if the ICE connection unexpectedly breaks.		*/
/*									*/
/************************************************************************/
static void 
IOErrorProc(IceConn ice_conn)
{
    ErrorF(__FILE__,__LINE__,"IOErrorProc","\n");

} /* IOErrorProc */

/************************************************************************/
/*									*/
/* IOErrorHandler							*/
/*									*/
/* 	Called if ICE detects an error.					*/
/*									*/
/************************************************************************/
static void 
IOErrorHandler(IceConn ice_conn)
{
    ErrorF(__FILE__,__LINE__,"IOErrorHandler","\n");

} /* IOErrorHandler */

/************************************************************************/
/*									*/
/* ProcessIceMsg							*/
/*									*/
/* 	Responsible for processing the set of messages that can	        */
/*	be received from ICE.						*/
/*									*/
/************************************************************************/
static void 
ProcessIceMsg(XtPointer client_data, 
	      int	*source, 
	      XtInputId	*id)
{
    Bool			reply_ready;
    IceProcessMessagesStatus	status;
    RapAgntIceConnPtr		ici = (RapAgntIceConnPtr) client_data;

    /* IceProcessMessages ends up calling RapAgntProcessMsg
     * because it is the message handler for the RAP protocol.
     *
     */
    status = IceProcessMessages(ici->ice_conn,
				NULL,
				&reply_ready);
    
    /* Check the status to see if there were any problems.
     */
    if (status != IceProcessMessagesSuccess) {
	switch(status) {
	    case IceProcessMessagesIOError: {
		ErrorF(__FILE__,__LINE__,"ProcessIceMsg",
		       "IceProcessMessagesIOError\n");
		IceCloseConnection(ici->ice_conn);
		break;
	    }
	    case IceProcessMessagesConnectionClosed: {
		ErrorF(__FILE__,__LINE__,"ProcessIceMsg",
		       "IceProcessMessagesConnectionClosed\n");
		break;
	    }
	    default: {
		ErrorF(__FILE__,__LINE__,"ProcessIceMsg",
		       "Unknown status %1d\n",status);
		break;
	    }
	}
	RemoveIceConn((RapAgntWidget)ici->w,ici->ice_conn);
    }
    	
} /* ProcessIceMsg */

/************************************************************************/
/*									*/
/* DispatchMsg								*/
/*									*/
/* 	Responsible for handling/dispatching the set of messages that 	*/
/*	can be received from the protocol.				*/
/*									*/
/************************************************************************/
static void
DispatchMsg(RapAgntWidget 	raw,
	    IceConn		ice_conn,
	    CARD8		opcode,
	    CARD16		seq,
	    char		*pdata)
{
    RapAgntClientPtr	client = RapAgntFindClient((Widget)raw,0,ice_conn);
    
    switch(opcode) {

	/* The RapHelloReply message is the client's response to our 
	 * RapHelloRequest message.  Once we have this, save the
	 * client in the knownClients list.  This allows us to
	 * make sure we don't keep trying to initiate a connection
	 * with a client should someone do something like restart
	 * the window manager.
	 */
	case RapHelloReply: {
	    Window client_window;
	    RapAgntHandleHelloReply(raw,pdata,&client_window);
	    AddClient(raw,client_window,ice_conn,NULL);
	    break;
	}

	case RapAcknowledge: {
	    RapAgntHandleAck(raw,seq,pdata);
	    break;
	}
	
	case RapQueryTreeReply: {
	    RapAgntHandleQueryTreeReply(raw,client,seq,pdata,False);
	    break;
	}
	
	case RapFullQueryTreeReply: {
	    RapAgntHandleQueryTreeReply(raw,client,seq,pdata,True);
	    break;
	}
	
	case RapGetResourcesReply: {
	    RapAgntHandleGetResourcesReply(raw,client,seq,pdata);
	    break;
	}
	
#ifdef GC_SUPPORT
	case RapGetGCValuesReply: {
	    RapAgntHandleGetGCValuesReply(raw,client,seq,pdata);
	    break;
	}
	
#endif
	case RapGetValuesReply: {
	    RapAgntHandleGetValuesReply(raw,client,seq,pdata);
	    break;
	}
	
	case RapGetHooksReply: {
	    RapAgntHandleGetHooksReply(raw,client,seq,pdata);
	    break;
	}
	
	case RapCreateHookNotify: {
	    RapAgntHandleCreateHook(raw,client,pdata);
	    break;
	}
	
	case RapDestroyHookNotify: {
	    RapAgntHandleDestroyHook(raw,client,pdata);
	    break;
	}
	
	case RapChangeHookNotify: {
	    RapAgntHandleChangeHook(raw,client,pdata);
	    break;
	}
	
	case RapConfigureHookNotify: {
	    RapAgntHandleConfigureHook(raw,client,pdata);
	    break;
	}

	case RapGeometryHookNotify: {
	    RapAgntHandleGeometryHook(raw,client,pdata);
	    break;
	}
		
	case RapActionHookNotify: {
	    RapAgntHandleActionHook(raw,client,pdata);
	    break;
	}
	
	case RapGetEventsReply: {
	    RapAgntHandleGetEventsReply(raw,client,seq,pdata);
	    break;
	}
	
	case RapEventNotify: {
	    RapAgntHandleEventNotify(raw,client,pdata);
	    break;
	}
	
	case RapGetRequestsReply: {
	    RapAgntHandleGetRequestsReply(raw,client,seq,pdata);
	    break;
	}
	
	case RapRequestNotify: {
	    RapAgntHandleRequestNotify(raw,client,pdata);
	    break;
	}
	
	case RapLocateObjectReply: {
	    RapAgntHandleLocateObjectReply(raw,client,seq,pdata);
	    break;

	}
	
	default: {
	    ErrorF(__FILE__,__LINE__,"ProcessMsg",
		   "unknown opcode %d\n",opcode);
	    break;
	}
	
    } /* switch(opcode) */

} /* DispatchMsg */

/************************************************************************/
/*									*/
/* ProcessMsg								*/
/*									*/
/* 	Responsible for processing the set of messages that can	        */
/*	be received from the protocol.					*/
/*									*/
/************************************************************************/
static void 
ProcessMsg(IceConn		ice_conn,
	   IcePointer 		client_data,
	   int			opcode,
	   unsigned long	length,
	   Bool			swap)
{
    RapAgntWidget 	raw = (RapAgntWidget) client_data;
    RapMessageHeader 	*header;
    char		*pdata;
    CARD16		seq;
    
    IceReadCompleteMessage(ice_conn,
			   SIZEOF(RapMessageHeader),
			   RapMessageHeader,
			   header,
			   pdata);
    if (!IceValidIO(ice_conn)) {
	ErrorF(__FILE__,__LINE__,"ProcessMsg",
	       "IceValidIO returned False on connection.\n");
    }
    seq = header->sequenceNumber;

    /* Dispatch the message right away if we are not waiting
     * for a reply, or if this is the reply we're waiting for.
     * Otherwise, add the message to the end of the event queue.
     */
    if (!raw->rapagnt.waiting_for_reply ||
	(raw->rapagnt.waiting_for_reply &&
	 (raw->rapagnt.wait_opcode == opcode) &&
	 (raw->rapagnt.wait_sequence_num == header->sequenceNumber)))
	DispatchMsg(raw,ice_conn,opcode,header->sequenceNumber,pdata);
    else {
	RapAgntEventQueuePtr new,current,prev;

	new = XtNew(RapAgntEventQueue);
	new->ice_conn = ice_conn;
	new->header = *header;
#ifdef USONIX 
/* Change suggested by Marty */
/* Fixes a  mystery problem he was having. */
	new->pdata = XtMalloc(8 * header->length +4);
	new->next = NULL;
	memcpy(new->pdata,pdata,(8 * header->length +4));
#else
	new->pdata = XtMalloc(8 * header->length);
	new->next = NULL;
	memcpy(new->pdata,pdata,(8 * header->length));
#endif

	if (raw->rapagnt.event_queue == NULL)
	    raw->rapagnt.event_queue = new;
	else {
	    current = raw->rapagnt.event_queue;
	    while (current != NULL) {
		prev = current;
		current = current->next;
	    }
	    prev->next = new;
	}
    }

    IceDisposeCompleteMessage(ice_conn,pdata);

} /* ProcessMsg */

/************************************************************************/
/*									*/
/* WorkProc - process the event queue.					*/
/*									*/
/************************************************************************/
static Boolean
WorkProc(XtPointer client_data)
{
    RapAgntWidget 	 raw = (RapAgntWidget) client_data;
    RapAgntEventQueuePtr current;
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"WorkProc","\n");
#endif

    while (raw->rapagnt.event_queue) {
	current = raw->rapagnt.event_queue;
	DispatchMsg(raw,
		    current->ice_conn,
		    current->header.minorOpcode,
		    current->header.sequenceNumber,
		    current->pdata);
	raw->rapagnt.event_queue = current->next;
	XtFree(current->pdata);
	XtFree((char *)current);
    }

    return True;
    
} /* WorkProc */

/************************************************************************/
/*									*/
/* RapAgntWaitForReply							*/
/*									*/
/************************************************************************/
void
RapAgntWaitForReply(Widget		w,
		    RapAgntClientPtr	client,
		    CARD16		seq,
		    CARD8		opcode)
{
    RapAgntWidget	raw = (RapAgntWidget) w;
    fd_set 		readfds;
    RapAgntIceConn	ici;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapAgntWaitForReply","\n");
#endif


    raw->rapagnt.waiting_for_reply = True;
    raw->rapagnt.wait_sequence_num = seq;
    raw->rapagnt.wait_opcode       = opcode;

    if (client->ice_conn == (IceConn) RAP_AGENT_SELF_CONNECT) {
	raw->rapagnt.waiting_for_reply = False;
	return;
    }
    
    ici.ice_conn = client->ice_conn;
    ici.w	 = w;
    
    FD_ZERO(&readfds);
    FD_SET(IceConnectionNumber(client->ice_conn),&readfds);

    while (raw->rapagnt.waiting_for_reply) {
	select(FD_SETSIZE,&readfds,NULL,NULL,NULL);
	ProcessIceMsg((XtPointer)&ici,NULL,NULL);
    }
    
    /* Add a handler for any events that might have queued up 
     * while waiting for the reply.
     */
    if (raw->rapagnt.event_queue != NULL)
	XtAppAddWorkProc(XtWidgetToApplicationContext((Widget)raw),
			 WorkProc,
			 (XtPointer)raw);

} /* RapAgntWaitForReply */

/************************************************************************/
/*									*/
/* NewConnection							*/
/*									*/
/************************************************************************/
void 	
NewConnection(XtPointer	client_data, 
	      int	*source, 
	      XtInputId	*id)
{
    RapAgntWidget	raw = (RapAgntWidget) _RapAgntGet();
    IceConn		ice_conn;
    IceAcceptStatus	status;
    RapAgntIceConnPtr	ici;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"NewConnection","\n");
#endif

    /* Accept the connection
     */
    ice_conn = IceAcceptConnection((IceListenObj) client_data,
				   &status);
    
    if (status != IceAcceptSuccess) {
	printf("agent: newConnectionProc accept connection failed due to ");
	switch(status)
	{
	    case IceAcceptFailure: {
		ErrorF(__FILE__,__LINE__,"NewConnection",
		       "IceAcceptFailure\n");
		break;
	    }
	    case IceAcceptBadMalloc: {
		ErrorF(__FILE__,__LINE__,"NewConnection",
		       "IceAcceptBadMalloc\n");
		break;
	    }
	    default: {
		ErrorF(__FILE__,__LINE__,"NewConnection",
		       "unknown status %d\n",status);
		break;
	    }
	}
	exit(1);
    }

    /* Add an input handler for this ICE connection.
     */
    ici = AddIceConn(raw,ice_conn);
    ici->input_id = XtAppAddInput(XtWidgetToApplicationContext((Widget)raw),
				  IceConnectionNumber(ice_conn),
				  (XtPointer)XtInputReadMask,
				  ProcessIceMsg,(XtPointer)ici);

} /* NewConnection */

/************************************************************************/
/*									*/
/* Initialize								*/
/*									*/
/************************************************************************/
static void
Initialize(Widget 	rw, 
	   Widget 	nw, 
	   ArgList 	args, 
	   Cardinal	*num_args)
{
    RapAgntWidget	raw = (RapAgntWidget) nw;
    IcePaVersionRec	protocol_versions[1];
    Status		status;
    XtAppContext	app_context = XtWidgetToApplicationContext((Widget)raw);
    char		error_str[256];
    int			i;
    Display		*dpy = XtDisplay(nw);
    
    /* Set the agent up for protocol reply.
     */
    protocol_versions[0].major_version = RapMajorVersion;
    protocol_versions[0].minor_version = RapMinorVersion;
    protocol_versions[0].process_msg_proc = ProcessMsg;
        
    raw->rapagnt.init_atom  = XInternAtom(dpy,
					  ICE_INITIATE RapName,False);
    raw->rapagnt.major_opcode = 
        IceRegisterForProtocolReply(RapName,		/* Protocol Name */
				    "X Consortium",	/* Vendor Name	 */
				    "1.0", 		/* Release	 */
				    1,			/* Num Versions	 */
				    protocol_versions,	/* Versions	 */
				    0,			/* Num Auth's	 */
				    NULL,		/* Auth Names	 */
				    NULL,		/* Auth Procs	 */
				    ProtoAuthorize,	/* Host Auth Proc*/
				    ProtoSetup,		/* Before Reply	 */
				    ProtoActivate,	/* After Reply	 */
				    IOErrorProc);	/* Error Handler */

    if (raw->rapagnt.major_opcode == -1) {
	ErrorF(__FILE__,__LINE__,"Initialize",
	       "IceRegisterForProtocolReply failed.\n");
	exit(1);
    }
    else
	IceSetIOErrorHandler(IOErrorHandler);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"Initialize","rapagnt.major_opcode=%1d\n",
	  raw->rapagnt.major_opcode);
#endif

    /* Get the listen objects and set the agent up to listen
     * on the selected objects.
     */
    status = IceListenForConnections(&raw->rapagnt.num_listen_objs,
				     &raw->rapagnt.listen_objs,
				     255,error_str);

    if (!status) {
        ErrorF(__FILE__,__LINE__,"Initialize",
	       "IceListenForConnections failed: %s\n",error_str);
	exit(1);
    }

    raw->rapagnt.input_ids = (XtInputId *)XtMalloc(sizeof(XtInputId) *
						   raw->rapagnt.num_listen_objs);
    
    for (i = 0; i < raw->rapagnt.num_listen_objs; i++) {
	raw->rapagnt.input_ids[i] = 
	    XtAppAddInput(app_context,
			  IceGetListenConnectionNumber(raw->rapagnt.listen_objs[i]),
			  (XtPointer)XtInputReadMask,
			  NewConnection,(XtPointer)raw->rapagnt.listen_objs[i]);
	IceSetHostBasedAuthProc(raw->rapagnt.listen_objs[i],ProtoAuthorize);
    }
    
    /* Setup the agent so clients can find the ICE Network IDs.
     */
    raw->rapagnt.network_ids = 
	IceComposeNetworkIdList(raw->rapagnt.num_listen_objs,
				raw->rapagnt.listen_objs);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"Initialize","rapagnt.network_ids='%s'\n",
	  raw->rapagnt.network_ids);
#endif

    raw->rapagnt.network_ids_window = XtWindow(XtParent(raw));
    XteaSetupAgent(dpy,raw->rapagnt.network_ids_window,
		   raw->rapagnt.network_ids);

    /* Set up the ClientMessage handler for failed attempts to initiate
     * a connection.
     */
    XtAddEventHandler((Widget)raw,		/* Widget to rcv events */
                      (EventMask) 0,		/* EventMask		*/
		      TRUE,			/* nonmaskable events	*/
	              RapAgntCheckMessages,  	/* The handler.         */
	              NULL);		        /* Client data.         */

    /* Select for the SubstructureNotify event on the root window.
     * This will tell us when a child of the root window has a
     * StructureNotify and is a way for us to determine when a new
     * client has connected to the display.
     */
    if (raw->rapagnt.auto_initiate) {	
	XSelectInput(dpy,
		     DefaultRootWindow(dpy),
		     SubstructureNotifyMask);

	XtSetEventDispatcher(dpy,CreateNotify,NewWindow);

	RapAgntContactExistingClients((Widget)raw);
    }

    raw->rapagnt.known_ice_conn    = NULL;
    raw->rapagnt.last_sequence_num = 0;
    raw->rapagnt.waiting_for_reply = False;
    raw->rapagnt.event_queue 	   = NULL;
    
} /* Initialize */

/************************************************************************/
/*									*/
/* SetValues								*/
/*									*/
/************************************************************************/
static Boolean
SetValues(Widget 	current, 
	  Widget 	request, 
	  Widget 	new, 
	  ArgList 	args, 
	  Cardinal 	*num_args)
{
    RapAgntWidget 	raw = (RapAgntWidget) new;

    return False;
    
} /* SetValues */

/************************************************************************/
/*									*/
/* Destroy								*/
/*									*/
/************************************************************************/
static void
Destroy(Widget rw) 
{
    RapAgntWidget 	raw = (RapAgntWidget) rw;
    int			i;

    /* Get rid of the known clients
     */
    while (raw->rapagnt.known_clients)
	RemoveClient(raw,
		     raw->rapagnt.known_clients->window,
		     raw->rapagnt.known_clients->ice_conn,
		     RapAgntDestroy);
    
    /* Close and free the listen objects
     */
    IceFreeListenObjs(raw->rapagnt.num_listen_objs,
		      raw->rapagnt.listen_objs);
    
    /* Remove the input handlers for the listen objects
     */
    for (i = 0; i < raw->rapagnt.num_listen_objs; i++)
	XtRemoveInput(raw->rapagnt.input_ids[i]);
    XtFree((char *)raw->rapagnt.input_ids);
    
    /* Remove the CheckMessages message handler.
     */
    XtRemoveEventHandler((Widget)raw,		/* Widget to rcv events */
			 (EventMask) 0,		/* EventMask		*/
			 TRUE,			/* nonmaskable events	*/
			 RapAgntCheckMessages,	/* The handler.         */
			 NULL);		        /* Client data.         */

    if (raw->rapagnt.auto_initiate) {	
	XtSetEventDispatcher(XtDisplay(rw),CreateNotify,NULL);
    }

} /* Destroy */

/************************************************************************/
/*									*/
/* RapAgntCreate							*/
/*									*/
/************************************************************************/
Widget 
RapAgntCreate(Widget 		parent, 
	      char 		*name, 
	      ArgList	   	args, 
	      Cardinal	   	num_args,
	      RapAgntGetProc 	get_proc)
{
    _RapAgntGet = get_proc;
    
    return XtCreateWidget(name,rapAgntWidgetClass,
			  parent, args, num_args);
    
} /* RapAgntCreate */
