/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
//
//      Copyright 1992, 1993, 1994, 1995
//      Georgia Tech Research Corporation
//      All rights reserved.
//
*/
/*
* $Log: RapClnt-in.c,v $
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*                                                                      */
/* RapClnt-in.c                                                         */
/*                                                                      */
/*      Handles all incoming RAP messages to the client.		*/
/*                                                                      */
/************************************************************************/
#include <X11/Xmd.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEmsg.h>
#include <X11/ICE/ICEproto.h>
#include "Rap.h"
#include "RapProto.h"
#include "RapClnt.h"

/************************************************************************/
/*									*/
/* RapClntHandleGetResourcesRequest					*/
/*									*/
/************************************************************************/
RapStatus
RapClntHandleGetResourcesRequest(RapInfo ri,
				 CARD16	 seq,
				 char	 *pdata)
{
    char	     *bufptr = pdata;
    RapObjectID      object_id;
    RapStatus	     status;
    int		     i;
    
    EXTRACT_OBJECTID(bufptr,False,object_id);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleGetResourcesRequest",
	  "object_id = %lx\n",object_id.id);
#endif

    status = RapClntGetResourcesReply(ri,seq,&object_id);

    return status;

} /* RapClntHandleGetResourcesRequest */

#ifdef GC_SUPPORT
/************************************************************************/
/*									*/
/* RapClntHandleGetGCValuesRequest  */
/*									*/
/************************************************************************/
RapStatus
RapClntHandleGetGCValuesRequest(RapInfo ri,
			      CARD16  seq,
			      char    *pdata)
{
    CARD32 numGCs;
    GContext *GCs;
    int ret;
 
    EXTRACT_CARD32(pdata, FALSE, numGCs);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleGetGCValuesRequest",
	  "numGCs = %d\n", numGCs);
#endif
 
    GCs = (numGCs <1)? (GContext *)NULL : (GContext *) pdata;
 
    ret = RapClntGetGCValuesReply(ri, seq, numGCs, GCs);

    return ret;
} /* RapClntHandleGetGCValuesRequest */

#endif /* GC_SUPPORT */

/************************************************************************/
/*									*/
/* RapClntHandleGetValuesRequest					*/
/*									*/
/************************************************************************/
RapStatus
RapClntHandleGetValuesRequest(RapInfo ri,
			      CARD16  seq,
			      char    *pdata)
{
    char	     *bufptr = pdata;
    RapObjectID	     object_id;
    CARD32	     num_resources;
    char	     **resources;
    CARD32	     pad;
    RapStatus	     status;
    int		     i;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleGetValuesRequest","\n");
#endif

    EXTRACT_OBJECTID(bufptr,False,object_id);
    EXTRACT_CARD32(bufptr,False,num_resources);
    EXTRACT_CARD32(bufptr,False,pad);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleGetValuesRequest",
	  "object_id = %lx\n",object_id.id);
    InfoF(__FILE__,__LINE__,"RapClntHandleGetValuesRequest",
	  "num_resources = %lx\n",num_resources);
#endif

    resources = (char **)XtMalloc(sizeof(char *) * num_resources);
    for (i = 0; i < num_resources; i++) {
	EXTRACT_STRING(bufptr,False,resources[i]);
#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapClntHandleGetValuesRequest",
	      "resource[%d] = %s\n",i,resources[i]);
#endif
    }
    status = RapClntGetValuesReply(ri,seq,&object_id,resources,num_resources);

    for (i = 0; i < num_resources; i++)
	XtFree(resources[i]);
    XtFree((char *)resources);

    return status;

} /* RapClntHandleGetValuesRequest */

/************************************************************************/
/*									*/
/* RapClntHandleSetValueRequest						*/
/*									*/
/************************************************************************/
RapStatus
RapClntHandleSetValuesRequest(RapInfo ri,
			      CARD16  seq,
			      char    *pdata)
{
    char	     *bufptr = pdata;
    RapObjectID	     object_id;
    RapArgList	     args;
    int		     num_args;
    CARD32	     pad;
    RapStatus	     status;
    int		     i;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleSetValuesRequest","\n");
#endif
    
    /* Read the message and send a reply.
     */
    EXTRACT_OBJECTID(bufptr,False,object_id);
    EXTRACT_CARD32(bufptr,False,num_args);
    EXTRACT_CARD32(bufptr,False,pad);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleSetValuesRequest",
	  "object_id = %lx\n",object_id.id);
    InfoF(__FILE__,__LINE__,"RapClntHandleSetValuesRequest",
	  "num_args = %d\n",num_args);
#endif
    
    args = (RapArgList)XtMalloc(sizeof(RapArg) * num_args);
    for (i = 0; i < num_args; i++) {
	EXTRACT_STRING(bufptr,False,args[i].name);
	args[i].value = (RapValuePtr)XtMalloc(sizeof(RapValue));
	EXTRACT_STRING(bufptr,False,args[i].value->type);
	EXTRACT_BYTES(bufptr,False,args[i].value->size,args[i].value->value);

#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapClntHandleSetValuesRequest",
	      "args[%d].name = %s\n",i,args[i].name);
	InfoF(__FILE__,__LINE__,"RapClntHandleSetValuesRequest",
	      "args[%d].value->type = %s\n",i,args[i].value->type);	
	InfoF(__FILE__,__LINE__,"RapClntHandleSetValuesRequest",
	      "args[%d].value->size = %s\n",i,args[i].value->size);
	InfoF(__FILE__,__LINE__,"RapClntHandleSetValuesRequest",
	      "args[%d].value->value = %s\n",i,args[i].value->value);
#endif	
    }

    status = RapClntSetValuesReply(ri,seq,&object_id,args,num_args);

    /* Free the memory we used.
     */
    for (i = 0; i < num_args; i++) {
	XtFree(args[i].name);
	XtFree(args[i].value->type);
	XtFree(args[i].value->value);
	XtFree((char *)args[i].value);
    }
    XtFree((char *)args);

    return status;

} /* RapClntHandleSetValuesRequest */

/************************************************************************/
/*									*/
/* RapClntHandleSelectHooksRequest					*/
/*									*/
/************************************************************************/
RapStatus
RapClntHandleSelectHooksRequest(RapInfo ri,
				CARD16  seq,
				char    *pdata)
{
    char	     	*bufptr = pdata;
    CARD16		hooks,mask;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleSelectHooksRequest","\n");
#endif

    EXTRACT_CARD16(bufptr,False,hooks);
    EXTRACT_CARD16(bufptr,False,mask);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleSelectHooksRequest",
	  "hooks = %x, mask = %x\n",hooks,mask);
#endif

    return RapClntSelectHooksReply(ri,seq,hooks,mask);

} /* RapClntHandleSelectHooksRequest */


/************************************************************************/
/*									*/
/* RapClntHandleSelectEventsRequest					*/
/*									*/
/************************************************************************/
RapStatus
RapClntHandleSelectEventsRequest(RapInfo ri,
				 CARD16  seq,
				 char    *pdata)
{
    char	     	*bufptr = pdata;
    CARD32		events[4],mask[4];
    int			i;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleSelectEventsRequest","\n");
#endif

    for (i = 0; i < 4; i++) {
	EXTRACT_CARD32(bufptr,False,events[i]);
	EXTRACT_CARD32(bufptr,False,mask[i]);

#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapClntHandleSelectHooksRequest",
	      "events[%d] = %x, mask[%d] = %x\n",i,events[i],i,mask[i]);
#endif
    }

    return RapClntSelectEventsReply(ri,seq,events,mask);

} /* RapClntHandleSelectEventsRequest */


/************************************************************************/
/*									*/
/* RapClntHandleSelectRequestsRequest					*/
/*									*/
/************************************************************************/
RapStatus
RapClntHandleSelectRequestsRequest(RapInfo ri,
				   CARD16  seq,
				   char    *pdata)
{
    char	     	*bufptr = pdata;
    CARD32		requests[8],mask[8];
    int			i;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleSelectRequestsRequest","\n");
#endif

    for (i = 0; i < 8; i++) {
	EXTRACT_CARD32(bufptr,False,requests[i]);
	EXTRACT_CARD32(bufptr,False,mask[i]);
    
#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapClntHandleSelectRequestsRequest",
	      "requests[%d] = %x, mask[%d] = %x\n",i,requests[i],i,mask[i]);
#endif
    }

    return RapClntSelectRequestsReply(ri,seq,requests,mask);

} /* RapClntHandleSelectRequestsRequest */

/************************************************************************/
/*									*/
/* RapClntHandleLocateObjectRequest					*/
/*									*/
/************************************************************************/
RapStatus
RapClntHandleLocateObjectRequest(RapInfo ri,
				 CARD16  seq,
				 char    *pdata)
{
    char	     *bufptr = pdata;
    RapObjectID      object_id;
    RapStatus	     status;
    CARD32	     x,y;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleLocateObjectRequest","\n");
#endif

    EXTRACT_OBJECTID(bufptr,False,object_id);
    EXTRACT_CARD32(bufptr,False,x);
    EXTRACT_CARD32(bufptr,False,y);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHandleLocateObjectRequest",
	  "object = %lx, x = %d, y = %d\n",object_id.id,x,y);
#endif

    status = RapClntLocateObjectReply(ri,seq,&object_id,(Position)x,(Position)y);

    return status;

} /* RapClntHandleLocateObjectRequest */
