/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
//
//      Copyright 1992, 1993, 1994, 1995
//      Georgia Tech Research Corporation
//      All rights reserved.
//
*/
/*

Copyright (c) 1989  X Consortium

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from the X Consortium.

*/
/*
* $Log: RapClnt-out.c,v $
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*                                                                      */
/* RapClnt-out.c                                                        */
/*                                                                      */
/*      Prepares all outgoing RAP messages from the client.		*/
/*                                                                      */
/************************************************************************/

#ifdef GC_SUPPORT
/* this is 4*23 - cause there are 23 attributes send over
   as CARD32s each */
#define GCVALUES_BUF_SIZE 92
#define GC_ID_SIZE 4
#define ERR_CODE_SIZE 4

#define ValidGCValuesBits (GCFunction | GCPlaneMask | GCForeground | \
                           GCBackground | GCLineWidth | GCLineStyle | \
                           GCCapStyle | GCJoinStyle | GCFillStyle | \
                           GCFillRule | GCTile | GCStipple | \
                           GCTileStipXOrigin | GCTileStipYOrigin | \
                           GCFont | GCSubwindowMode | GCGraphicsExposures | \
                           GCClipXOrigin | GCClipYOrigin | GCDashOffset | \
                           GCArcMode)
#endif

#define NEED_EVENTS
#ifdef HAS_MOTIF
#include<Xm/Xm.h>
#endif
#include <X11/Xmd.h>
#include <X11/Xlibint.h>
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#include <X11/ShellP.h>		/* for ApplicationShellWidgetClass */

#include <X11/Xproto.h>

#include <X11/ICE/ICElib.h>
#include <X11/ICE/ICEmsg.h>
#include <X11/ICE/ICEproto.h>
#include "Rap.h"
#include "RapProto.h"
#include "RapClnt.h"

#ifdef GC_SUPPORT
#include "XGCList.h"
#endif

extern XExtCodes *rap_ext_codes;

/************************************************************************/
/*									*/
/* WriteToAgent								*/
/*									*/
/************************************************************************/
static RapStatus
WriteToAgent(RapInfo ri, 
	     CARD8   opcode,
	     CARD16  seq,
	     int     num_bytes,
	     char    *pdata)
{
    RapMessageHeader *header;
    int 	     padding = 0;
    int		     total_bytes;
    
    if (ri->ice_conn == (IceConn) RAP_AGENT_SELF_CONNECT) {
	(*ri->dispatch_proc)((Widget)ri->private,
			     (IceConn)RAP_AGENT_SELF_CONNECT,
			     (CARD8)opcode,
			     (CARD16)seq,
			     (char *)pdata);
    }
    else {
	IceGetHeader(ri->ice_conn,
		     ri->major_opcode,
		     opcode,
		     SIZEOF(RapMessageHeader),
		     RapMessageHeader,
		     header);

	if (header) {
	    header->sequenceNumber = seq;

	    if (pdata) {
		padding = PAD64(num_bytes);
		total_bytes = num_bytes + padding;
		header->length += WORD64COUNT(total_bytes);	
		IceWriteData(ri->ice_conn,num_bytes,pdata);
		if (padding) IceWritePad(ri->ice_conn,padding);
	    }
	    IceFlush(ri->ice_conn);

	    return RapSuccess;
	}
	else {
	    /* [[[WDW - We probably should handle malloc problems
	     * 	    a little more gracefully.]]]
	     */
	    ErrorF(__FILE__,__LINE__,"WriteToAgent",
		   "Unable to do IceGetHeader\n");
	    return RapBadAlloc;
	}
    }
    
} /* WriteToAgent */

/************************************************************************/
/*									*/
/* RapClntHelloReply							*/
/*									*/
/*	Send a RapHelloReply to the agent.				*/
/*									*/
/************************************************************************/
RapStatus
RapClntHelloReply(RapInfo ri)
{
    RapMessageHeader *reply = NULL;
    int		     total_bytes = 0;
    CARD32	     window = (CARD32) XtWindow(ri->top_level);
    CARD32	     pad = 0;
    int		     padding;
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntHelloReply","\n");
#endif
    
    IceGetHeader(ri->ice_conn, 
		 ri->major_opcode, 
		 RapHelloReply,
		 sizeof(RapMessageHeader), 
		 RapMessageHeader, 
		 reply);

    if (reply) {
	total_bytes += 4; /* window */
	total_bytes += 4; /* pad    */

	/* Pad the total to 64-bit boundaries 
	 */
	padding = PAD64(total_bytes);
	total_bytes = total_bytes + padding;

	/* Set the size of the message 
	 */
	reply->length += WORD64COUNT(total_bytes);
	
	IceWriteData32(ri->ice_conn,sizeof(CARD32),&window);
	IceWriteData32(ri->ice_conn,sizeof(CARD32),&pad);

	/* Pad to 64 bits and flush
	 */
	IceWritePad(ri->ice_conn,padding);
	IceFlush(ri->ice_conn);

	return RapSuccess;
    }
    else
    {
	/* [[[WDW - We probably should handle malloc problems
	 * 	    a little more gracefully.]]]
	 */
	ErrorF(__FILE__,__LINE__,"RapClntHelloReply",
	       "Unable to do IceGetHeader\n");
	return RapBadAlloc;
    }
    
} /* RapClntHelloReply */

/************************************************************************/
/*									*/
/* RapClntAck								*/
/*									*/
/************************************************************************/
RapStatus
RapClntAck(RapInfo ri, CARD16 seq)
{
    return WriteToAgent(ri,RapAcknowledge,seq,0,NULL);

} /* RapClntAck */

/************************************************************************/
/*									*/
/* FindChildren								*/
/*									*/
/************************************************************************/
static int
FindChildren(Widget parent, Widget **children, Boolean normal, Boolean popup)
{
    CompositeWidget cw = (CompositeWidget) parent;
    int i, num_children = 0, current = 0;

    if (XtIsWidget(parent) && popup)
	num_children += parent->core.num_popups;

    if (XtIsComposite(parent) && normal)
	num_children += cw->composite.num_children;

    if (num_children == 0) {
	*children = NULL;
	return 0;
    }

    *children = (Widget *) XtMalloc((Cardinal) sizeof(Widget) * num_children);

    if (XtIsComposite(parent) && normal)
	for (i=0 ; i<cw->composite.num_children; i++, current++)
	    (*children)[current] = cw->composite.children[i];

    if (XtIsWidget(parent) && popup)
	for (i=0 ; i<parent->core.num_popups ; i++, current++)
	    (*children)[current] = parent->core.popup_list[i];

    return num_children;

} /* FindChildren */

/************************************************************************/
/*									*/
/* IsApplicationShell							*/
/*									*/
/* This is a trick/kludge.  To make shared libraries happier (linking	*/
/* against Xmu but not linking against Xt, and apparently even work as 	*/
/* we desire on SVR4, we need to avoid an explicit data reference to	*/
/* applicationShellWidgetClass.  XtIsTopLevelShell is known		*/
/* (implementation dependent assumption!) to use a bit flag.  So we go	*/
/* that far.  Then, we test whether it is an applicationShellWidget	*/
/* class by looking for an explicit class name.  Seems pretty safe.	*/
/*									*/
/************************************************************************/
static Bool
IsApplicationShell(Widget w)
{
    register WidgetClass c;
 
    if (!XtIsTopLevelShell(w))
        return False;
    for (c = XtClass(w); c; c = c->core_class.superclass) {
        if (!strcmp(c->core_class.class_name, "ApplicationShell"))
            return True;
    }
    return False;

} /* Is ApplicationShell */

/************************************************************************/
/*									*/
/* CvtWarningHandler							*/
/*									*/
/************************************************************************/
static void 
CvtWarningHandler(String 	name, 
		  String 	type, 
		  String 	class, 
		  String 	def, 
		  String 	*params, 
		  Cardinal	*num_params)
{
  /* Just ignore the warning 
   */
  return;

} /* CvtWarningHandler */

/************************************************************************/
/*									*/
/* ComputeResourceSize							*/
/*									*/
/************************************************************************/
static int
ComputeResourceSize(Widget 		w, 
		    XtResourceList	resource,
		    Boolean		get_value,
		    int			*val_size,
		    char 		**val_ret)
{
    int 		num_bytes = 0;
    CARD32 		value_size = resource->resource_size;
    String 		string = "";
    char		buffer[100];
    CARD8 		card8;
    CARD16 		card16;
    CARD32 		card32;
#if (LONG_BIT == 64)
    long 		card64;
#endif
    long		value=0;
    int			bytes_required;
    XtErrorMsgHandler	old_handler;
    XrmValue 	        src,dest,dest_color;
    
    /* Storage needed for resource name, class, type, and kind
     */
    num_bytes += STRING_BYTES(strlen(resource->resource_name));
    num_bytes += STRING_BYTES(strlen(resource->resource_class));
    num_bytes += STRING_BYTES(strlen(resource->resource_type));
    num_bytes += 4; /* kind:  normal or constraint */
    num_bytes += 4; /* pad to 64-bit boundary */
    
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"ComputeResourceSize",
	  "name = %s, type = %s\n",
	  resource->resource_name,
	  resource->resource_type);
#endif
    
    /* Storage needed to hold the value
     */
    if (get_value) {

	/* NOTE:  Special case for strings.
	 */
	if (strcmp(resource->resource_type,XtRString) == 0) {
	    XtVaGetValues(w,resource->resource_name,&string,NULL);
#ifdef HAS_MOTIF
	} else if (strcmp(resource->resource_type, "XmString") == 0) {
	    XmString xmStr;
	    XtVaGetValues(w, resource->resource_name, &xmStr, NULL);
	    XmStringGetLtoR(xmStr, XmSTRING_DEFAULT_CHARSET, &string);
	    /* Motif hands back a pointer to a copy of the value */
	     	    	    XmStringFree(xmStr); 
#endif
	} 	else 	  {
	    switch (value_size) {
		/* CARD8 */
		case 1: {
		    XtVaGetValues(w,resource->resource_name,&card8,NULL);
		    value = (CARD8) card8;
		    break;
		}
		
		/* CARD16 */
		case 2: {
		    XtVaGetValues(w,resource->resource_name,&card16,NULL);
		    value = (CARD16) card16;
		    break;
		}
		
		/* CARD32 */
		case 4: {
		    XtVaGetValues(w,resource->resource_name,&card32,NULL);
		    value = (CARD32) card32;
		    break;
		}	
#if (LONG_BIT == 64)
		/* CARD64 */
		case 8: {
		    XtVaGetValues(w,resource->resource_name,&card64,NULL);
		    value = card64;
		    break;
		}
#endif
		default: {
		    ErrorF(__FILE__,__LINE__,"ComputeResourceSize",
			   "invalid size %d\n",value_size);
		    break;
		}
	    }

	    src.size = value_size;

	    /* This is a check because our XmString->String converter
	     * is busted. [[[WDW - We need to fix our converter]]]
	     */
		 if ((strcmp(resource->resource_type,"XmString") == 0) && value)
		src.addr = (caddr_t) value;
		else
		src.addr = (caddr_t) &value;

	    dest.addr = NULL;
	    dest.size = 0;
	    dest_color.addr = NULL;
	    dest_color.size = 0;
	    
	    /* Ignore conversion warnings.
	     */
	    old_handler = 
		XtAppSetWarningMsgHandler(XtWidgetToApplicationContext(w),
					  CvtWarningHandler);

	    if (!strcmp(resource->resource_type, XtRPixel) &&
		XtConvertAndStore(w, XtRPixel, &src, XtRColor, &dest) &&
		XtConvertAndStore(w, XtRColor, &src, XtRString, &dest_color)) {
		string = dest_color.addr;
	    }
	    else if (XtConvertAndStore(w,resource->resource_type, &src, 
				       XtRString, &dest)) {
		string = dest.addr;
	    }
	    else {
		/* Conversion failed, fall back to representing it as integer.
		 */
		switch (resource->resource_size) {
		    case 1: {
			sprintf(buffer,"%d",card8);
			string = buffer;
			break;
		    }
		    
		    case 2: {
			sprintf(buffer,"%d",card16);
			string = buffer;
			break;
		    }
		    
		    case 4: {
			sprintf(buffer,"%d",card32);
			string = buffer;
			break;
		    }
		    
#if (LONG_BIT == 64)
		    case 8: {
			sprintf(buffer,"%ld",card64);
			string = buffer;
			break;
		    }
#endif
		    default:
		        break;
		}
		strcat(buffer, " (integer fallback conversion)");
	    }
	    XtAppSetWarningMsgHandler(XtWidgetToApplicationContext(w), 
				      old_handler);
	}
	
	if (string)
	    value_size = strlen(string);
	else
	    value_size = 0;

	bytes_required = STRING_BYTES(strlen(XtRString));
	bytes_required += STRING_BYTES(value_size);
	num_bytes += bytes_required;
	
	if (val_size) *val_size = bytes_required;
	if (val_ret) {
	    char *bufp;
	    *val_ret = XtMalloc(bytes_required);
	    memset(*val_ret,'\0',bytes_required);
	    bufp = *val_ret;
	    STORE_STRING(bufp,strlen(XtRString),XtRString);
	    STORE_STRING(bufp,value_size,string);
	}
    } /* get_value */

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"ComputeResourceSize",
	  "value= %s\n",string);
#endif
    
    return num_bytes;

} /* ComputeResourceSize */

/************************************************************************/
/*									*/
/* ComputeResourceSizes							*/
/*									*/
/************************************************************************/
static int
ComputeResourceSizes(Widget 	w,
		     Boolean 	get_values)
{
    int			num_bytes = 0;
    register int 	i;
    XtResourceList 	normals, constraints = NULL;
    Cardinal 		num_normals, num_constraints, size;
    char 		*native, *name;
 
    num_bytes += 4; /* number of resources */
    num_bytes += 4; /* pad to 64-bit boundary */
    
    /* Do the resources of the widget itself
     */
    XtGetResourceList(XtClass(w), &normals, &num_normals);
    for (i=0; i < num_normals; i++)
        num_bytes += ComputeResourceSize(w,&(normals[i]),get_values,NULL,NULL);
    XtFree((char *) normals);

    /* Now do the constraint resources if this is a child.
     */
    if (XtParent(w) != NULL) {
        XtGetConstraintResourceList(XtClass(XtParent(w)), &constraints,
                                    &num_constraints);
	for (i=0; i < num_constraints; i++)
	    num_bytes += ComputeResourceSize(w,&(constraints[i]),get_values,
					     NULL,NULL);
	XtFree((char *) constraints);
    }

    return num_bytes;
    
} /* ComputeResourceSizes */

/************************************************************************/
/*									*/
/* ComputeObjectSize							*/
/*									*/
/************************************************************************/
static int
ComputeObjectSize(Widget 	w,
		  Boolean 	get_resources)
{
    int		num_bytes = 0;
    char 	*class;

    if (IsApplicationShell(w))
	class = ((ApplicationShellWidget) w)->application.class;
    else
	class = XtClass(w)->core_class.class_name;

    num_bytes += sizeof(RapObjectID);			/* our widget id */
    num_bytes += sizeof(RapObjectID);			/* our parent's id */
    num_bytes += STRING_BYTES(strlen(XtName(w)));	/* our name */
    num_bytes += STRING_BYTES(strlen(class));		/* our class */
    num_bytes += sizeof(CARD32);			/* our window id */
    num_bytes += sizeof(CARD32);			/* manage state */
    num_bytes += STRING_BYTES(strlen(TOOLKIT));		/* our toolkit */
    if (get_resources)
	num_bytes += ComputeResourceSizes(w,True);

    return num_bytes;
    
} /* ComputeObjectSize */

/************************************************************************/
/*									*/
/* ComputeTreeSizes							*/
/*									*/
/* Determine how many widgets are in the tree and how many bytes it	*/
/* will take to send them over the wire.       				*/
/*									*/
/************************************************************************/
static void
ComputeTreeSizes(Widget w, 
		 CARD32 *num_bytes, 
		 CARD32 *num_widgets, 
		 Boolean get_resources)
{
    Widget	 *children;
    int 	 num_children;
    register int i;

    (*num_widgets)++;

    (*num_bytes) += ComputeObjectSize(w,get_resources);

    num_children = FindChildren(w, &children, TRUE, TRUE);
    for (i=0; i < num_children; i++) {
	ComputeTreeSizes(children[i],num_bytes,num_widgets,
			 get_resources);
    }

    XtFree((char *) children);

} /* ComputeTreeSizes */


/************************************************************************/
/*									*/
/* StoreResource							*/
/*									*/
/************************************************************************/
static void
StoreResource(Widget 		w, 
	      XtResourceList	resource,
	      RapResourceKind	kind,
	      char 		**bufptr,
	      Boolean		get_value)
{
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"StoreResource",
	  "name = %s, class = %s, type = %s\n",
	  resource->resource_name,
	  resource->resource_class,
	  resource->resource_type);
#endif

    /* Store the name, class, type, and kind
     */
    STORE_STRING(*bufptr,strlen(resource->resource_name),
		 resource->resource_name);
    STORE_STRING(*bufptr,strlen(resource->resource_class),
		 resource->resource_class);
    STORE_STRING(*bufptr,strlen(resource->resource_type),
		 resource->resource_type);
    STORE_CARD32(*bufptr,kind);
    STORE_CARD32(*bufptr,0); /* Pad to 64-bit boundary */

    /* If we're getting the values, store the return type and
     * the actual value itself.
     */
    if (get_value) {
	int	val_size;
	char 	*val_buf;

	ComputeResourceSize(w,resource,True,&val_size,&val_buf);
	memcpy(*bufptr,val_buf,val_size);
	(*bufptr) += val_size;

	XtFree(val_buf);
    }

} /* StoreResource */

/************************************************************************/
/*									*/
/* StoreResources							*/
/*									*/
/************************************************************************/
static void
StoreResources(Widget 	w, 
	       char 	**bufptr,
	       Boolean	get_values)
{
    register int 	i;
    XtResourceList 	normals, constraints=NULL;
    Cardinal 		num_normals, num_constraints;
    
    /* Get the list of all the resources
     */
    XtGetResourceList(XtClass(w), &normals, &num_normals);
    if (XtParent(w) != NULL)
        XtGetConstraintResourceList(XtClass(XtParent(w)), &constraints,
                                    &num_constraints);
    else	
        num_constraints = 0;

    /* Store the number of resources 
     */
    STORE_CARD32(*bufptr,num_normals + num_constraints);
    STORE_CARD32(*bufptr,0); /* Pad to 64-bit boundary */

    /* Store the normal resources
     */
    for (i = 0; i < num_normals; i++)
	StoreResource(w,&(normals[i]),RapNormalResource,bufptr,get_values);
    XtFree((char *) normals);

    /* Store the contraint resources
     */
    if (constraints != NULL) {
	for (i = 0; i < num_constraints; i++)
	    StoreResource(w,&(constraints[i]),RapConstraintResource,
			  bufptr,get_values);
	XtFree((char *) constraints);
    }

} /* StoreResources */

/************************************************************************/
/*									*/
/* StoreObject								*/
/*									*/
/************************************************************************/
static void
StoreObject(Widget w, char **bufptr, Boolean get_resources)
{
    char 	*class;
    Window 	window;
    Widget 	parent;
    CARD32 	managed;
    RapObjectID	obj;

    if (IsApplicationShell(w))
	class = ((ApplicationShellWidget) w)->application.class;
    else
	class = XtClass(w)->core_class.class_name;

    if (XtIsWidget(w)) {
	if (XtIsRealized(w))
	    window = XtWindow(w);
	else
	    window = RapUnrealized;
    } 
    else {
	window = RapGadget;
    }

    parent = XtParent(w);
    managed = (CARD32) XtIsManaged(w);

    WidgetToRapObjectID(w,obj);
    WidgetToRapObjectID(parent,obj);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"StoreObject",
	  "widget = %lx %lx\n",w,obj.id);
    InfoF(__FILE__,__LINE__,"StoreObject",
	  "parent = %lx %lx\n",parent,obj.id);
    InfoF(__FILE__,__LINE__,"StoreObject",
	  "name = %s\n",XtName(w));
    InfoF(__FILE__,__LINE__,"StoreObject",
	  "class_name = %s\n",class);
    InfoF(__FILE__,__LINE__,"StoreObject",
	  "window = %x\n",window);
    InfoF(__FILE__,__LINE__,"StoreObject",
	  "managed = %d\n",managed);
    InfoF(__FILE__,__LINE__,"StoreObject",
	  "toolkit = %s\n",TOOLKIT);
#endif

    WidgetToRapObjectID(w,obj);
    STORE_OBJECTID(*bufptr,obj);
    WidgetToRapObjectID(parent,obj);
    STORE_OBJECTID(*bufptr,obj);
    STORE_STRING(*bufptr,strlen(XtName(w)),XtName(w));
    STORE_STRING(*bufptr,strlen(class),class);
    STORE_CARD32(*bufptr,window);
    STORE_CARD32(*bufptr,managed);
    STORE_STRING(*bufptr,strlen(TOOLKIT),TOOLKIT);
    if (get_resources)
	StoreResources(w,bufptr,True);

} /* StoreObject */

/************************************************************************/
/*									*/
/* StoreTree								*/
/*									*/
/************************************************************************/
static void
StoreTree(Widget w, char **bufptr, Boolean get_resources)
{
    Widget 	*children;
    int 	i,num_children;

    StoreObject(w,bufptr,get_resources);
    
    num_children = FindChildren(w, &children, TRUE, TRUE);
    for (i=0; i < num_children; i++) {
	StoreTree(children[i],bufptr,get_resources);
    }

    XtFree((char *) children);

} /* StoreTree */

/************************************************************************/
/*									*/
/* RapClntQueryTreeReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntQueryTreeReply(RapInfo ri,
		      CARD16  seq,
		      Boolean get_resources)
{
    Widget	     	hooks_widget;
    Arg 		args[2];
    Cardinal	     	num_shells;
    WidgetList		shells;
    CARD32		*num_widgets;
    char		*bufptr, *scratch;
    CARD32		total_bytes = 0;
    CARD32		pad = 0;
    register int 	i;
    RapStatus		status;
    
    /* Find out how many shells the client has
     */
    hooks_widget = XtHooksOfDisplay(XtDisplay(ri->top_level));
    XtSetArg(args[0], XtNnumShells, &num_shells);
    XtSetArg(args[1], XtNshells,    &shells);
    XtGetValues(hooks_widget,args,2);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntQueryTreeReply",
	  "num_shells = %d\n",num_shells);
#endif

    /* Compute the size of the reply
     */
    total_bytes += 4; /* num shells */
    total_bytes += 4; /* pad        */

    num_widgets = (CARD32 *) malloc(sizeof(CARD32) * num_shells);
    for (i=0; i < num_shells; i++) {
	total_bytes += 4; /* count */
	total_bytes += 4; /* pad   */
	num_widgets[i] = 0;
	ComputeTreeSizes(shells[i],&total_bytes,&(num_widgets[i]),
			 get_resources);
#ifdef DEBUG
	InfoF(__FILE__,__LINE__,"RapClntQueryTreeReply",
	      "shell %x has %d widgets\n",
	      shells[i],num_widgets[i]);
#endif
    }
    
    /* Write out the reply
     */
    bufptr = scratch = XtMalloc(total_bytes);
    STORE_CARD32(bufptr,num_shells);
    STORE_CARD32(bufptr,pad);
    
    for (i=0; i < num_shells; i++) {
	/* Write the number of elems in the list 
	 */
	STORE_CARD32(bufptr,num_widgets[i]);
	STORE_CARD32(bufptr,pad);

	/* Write the scratch buffer for this tree 
	 */
	StoreTree(shells[i],&bufptr,get_resources);
    }
    XtFree((char *)num_widgets);

    /* Send the data to the client
     */
    status = WriteToAgent(ri,
			  get_resources ? 
			  RapFullQueryTreeReply : 
			  RapQueryTreeReply,
			  seq,total_bytes,scratch);
    XtFree(scratch);

    return status;

} /* RapClntQueryTreeReply */

/************************************************************************/
/*									*/
/* RapClntGetResourcesReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntGetResourcesReply(RapInfo 	ri,
			 CARD16		seq,
			 RapObjectID	*object_id)
{
    Widget		w;
    char 		*bufptr, *scratch;
    CARD32		total_bytes = 0;
    CARD32		pad = 0;
    RapStatus		status;
    
    /* [[[Here we should do some check to see if the widget 
     * actually exists ]]]
     */
    RapObjectIDToWidget(*object_id,w);

    /* First compute the size of the reply.
     */
    total_bytes += 4;	/* error code 	 */
    total_bytes += 4;	/* pad	 	 */
    total_bytes += ComputeResourceSizes(w,False);

    /* Write out the reply
     */
    bufptr = scratch = XtMalloc(total_bytes);
    STORE_CARD32(bufptr,RapNoError);
    STORE_CARD32(bufptr,pad);
    StoreResources(w,&bufptr,False);

    /* Send the data to the client
     */
    status = WriteToAgent(ri,RapGetResourcesReply,seq,total_bytes,scratch);
    XtFree(scratch);

    return status;

} /* RapClntGetResourcesReply */

#ifdef GC_SUPPORT
/************************************************************************/
/*									*/
/* StoreGCValues						*/
/*									*/
/************************************************************************/
static void
StoreGCValues(GC gc, XGCValues *gcv, char **bufPointer)
{

   long dashes = 0L ;
	dashes = (long) gcv->dashes;

   /* store the fields, one at a time */
    STORE_CARD32(*bufPointer, gcv->function);
    STORE_CARD32(*bufPointer, gcv->plane_mask);
    STORE_CARD32(*bufPointer, gcv->foreground);
    STORE_CARD32(*bufPointer, gcv->background);
    STORE_CARD32(*bufPointer, gcv->line_width);
    STORE_CARD32(*bufPointer, gcv->line_style);
    STORE_CARD32(*bufPointer, gcv->cap_style);
    STORE_CARD32(*bufPointer, gcv->join_style);
    STORE_CARD32(*bufPointer, gcv->fill_style);
    STORE_CARD32(*bufPointer, gcv->fill_rule);
    STORE_CARD32(*bufPointer, gcv->arc_mode);
    STORE_CARD32(*bufPointer, gcv->tile);
    STORE_CARD32(*bufPointer, gcv->stipple);
    STORE_CARD32(*bufPointer, gcv->ts_x_origin);
    STORE_CARD32(*bufPointer, gcv->ts_y_origin);
    STORE_CARD32(*bufPointer, gcv->font);
    STORE_CARD32(*bufPointer, gcv->subwindow_mode);
    STORE_CARD32(*bufPointer, gcv->graphics_exposures);
    STORE_CARD32(*bufPointer, gcv->clip_x_origin);
    STORE_CARD32(*bufPointer, gcv->clip_y_origin);
    STORE_CARD32(*bufPointer, gcv->clip_mask);
    STORE_CARD32(*bufPointer, gcv->dash_offset);
    STORE_CARD32(*bufPointer, dashes);

}

/************************************************************************/
/*									*/
/* RapWriteAllGCValues						*/
/*									*/
/************************************************************************/
int
RAPWriteAllGCValues(RapInfo ri, char **scratch)
{
    XGCListPtr head = XGCGetListHead();
    XGCListPtr temp = head;
char *bufptr;
    int  i, padding;
    CARD32 errCode = RapNoError;
    Display *dpy = XtDisplay(ri->top_level);
 int numGCs = 0;
    int gcBufsize = GCVALUES_BUF_SIZE + GC_ID_SIZE + ERR_CODE_SIZE;

     while (temp)
     {
        if (dpy == temp->dpy)
                numGCs++;
 
        temp = temp->next;
     }
 
     /* a gcbuf per gc plus the total number of entries, which is
       a 4 byte entry */
    gcBufsize = (gcBufsize*numGCs) + 4;
 
    padding = PAD64(gcBufsize);
    gcBufsize += padding;
 
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapWriteAllGCValues","numGCs = %d\n",
      numGCs);
#endif
 
*scratch = bufptr = malloc(gcBufsize);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapWriteAllGCValues","gcBufsize = %d, scratch = %lx\n",
      gcBufsize, *scratch);
#endif

        STORE_CARD32(bufptr, numGCs);
 
     temp = head;
    while (temp)
    {
        XGCValues gcv;
        GContext gcId;
 
        if (dpy == temp->dpy)
        {
 
           gcId = XGContextFromGC(temp->gc);
 
           XGetGCValues(dpy, temp->gc,
                                ValidGCValuesBits, &gcv);
 
           STORE_CARD32(bufptr, gcId);
            STORE_CARD32(bufptr, errCode);
 
           /* write the fields, one at a time longs, ints and even dashes,
            which is a char, are sent as card32 ... is this the right
            way to do things?  I know I can't use WRITE_CARD32 in its
            current state */

StoreGCValues(temp->gc, &gcv, &bufptr);

}
 
      temp = temp->next;
 }
 
        /* write out padding */
    for (i=0 ; i<padding ; i++){
        char zero = '\0';
        memcpy(bufptr, &zero, 1);
        bufptr++;
    }

return gcBufsize;
} /* RapWriteAllGCValues */

/************************************************************************/
/*									*/
/* RapClntGetGCValueesReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntGetGCValuesReply(RapInfo 		ri,
		      CARD16		seq,
  int numGCs,
  GContext *GCs)
{
    register int i;
    int totalSizeInBytes = 0;
    long *numBytes;
    CARD32 *errorCodes;
    char **scratch=NULL, *bufptr=NULL, *outgoing=NULL;
    int padding;
    RapStatus status = RapSuccess;

if (numGCs < 1) {
/* Return all Gcs */
        totalSizeInBytes = RAPWriteAllGCValues(ri, &outgoing);

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntGetGCValuesReply","outgoing = %lx\n",
      outgoing);
#endif

    status = WriteToAgent(ri,RapGetGCValuesReply,seq,totalSizeInBytes, outgoing);

    if(totalSizeInBytes > 0)
      free(outgoing);  

     } else {  /* you actually have some GCs to work with */
    /* increment our total size to account for the number of entries field */
     totalSizeInBytes += 4;
 
    numBytes = (long *) malloc(sizeof(long) * numGCs);
    errorCodes = (CARD32 *) malloc(sizeof(CARD32) * numGCs);
    scratch = (char **) malloc(sizeof(char *) * numGCs);
 
    for (i=0 ; i<numGCs ; i++) {
        char *bufPointer;
	Status success;
	XGCValues gcv;
	GC gcptr = (GC) NULL;
 
        numBytes[i] = 0;
 
        /* Here we should do some check to see if the GC is valid */
        /* XGCFromGContext will return null if the gcid isn't valid */
	/* Hack alert - I made this change to get things */
	/* to compile.  Using top_level works if all is on the */
	/* same display.  This function used to take a widget */
	/* as an argument, from which the correct display */
	/* was extracted.  Should we go back to that? - BLS */
	/*        gcptr = XGCFromGContext(XtDisplay(w), GCs[i]); */
        gcptr = XGCFromGContext(XtDisplay(ri->top_level), GCs[i]);
         if (gcptr)
        {
        	errorCodes[i] = RapNoError;

		XGetGCValues(XtDisplay(ri->top_level), gcptr,
                                ValidGCValuesBits, &gcv);
 
		/*		ComputeGCValuesSize(gcptr, &(numBytes[i])); */
		numBytes[i] = GCVALUES_BUF_SIZE;
        	scratch[i] = (char *) malloc(numBytes[i]);
        	memset(scratch[i], '\0', numBytes[i]);
 
        	bufPointer = scratch[i];

#ifdef RAP_DEBUG
        printf("GC %lx, total size to store = %ld\n",
               gcptr, numBytes[i]);
#endif
 
        	StoreGCValues(gcptr, &gcv, &bufPointer); 

 
        	totalSizeInBytes += 4;                  /* GC id */
        	totalSizeInBytes += 4;                  /* error code */
        	totalSizeInBytes += numBytes[i];        /* gc values */
	} else {
		/* invalid GC */
        	errorCodes[i] = RapNoSuchGraphicsContext;

		/* just put the string in the scratch area!! */
		 numBytes[i] += STRING_BYTES(strlen("GC doesn't exist"));
		 scratch[i] = (char *) malloc(numBytes[i]);
        	 bufPointer = scratch[i];
		 STORE_STRING(bufPointer, strlen("GC doesn't exist"),
                         "GC doesn't exist");
        	totalSizeInBytes += 4;                  /* GC id */
        	totalSizeInBytes += 4;                  /* error code */
        	totalSizeInBytes += numBytes[i];        
		}
		
	}
 
    padding = PAD64(totalSizeInBytes);
    totalSizeInBytes += padding;
 
    outgoing = bufptr = (char*)malloc(totalSizeInBytes);

    /* Write out the number of entries we're returning */
    STORE_CARD32(bufptr, numGCs);
#ifdef RAP_DEBUG
    printf("WROTE NUM GCS = %ld\n", numGCs);
#endif
 
    for (i=0 ; i<numGCs ; i++) {
        STORE_CARD32(bufptr, GCs[i]);
#ifdef RAP_DEBUG
        printf("\t%d GC = %lx\n", i, GCs[i]);
#endif
        STORE_CARD32(bufptr, errorCodes[i]);
#ifdef RAP_DEBUG
        printf("\t%d ERROR CODE = %ld\n", i, errorCodes[i]);
#endif
 
        STORE_BYTES(bufptr, numBytes[i], scratch[i]);
#ifdef RAP_DEBUG
        printf("\t%d WROTE %ld bytes scratch data\n", i, numBytes[i]);
#endif
 
        free(scratch[i]);
    }
 
    for (i=0 ; i<padding ; i++) {
        char zero = '\0';
        STORE_SIZED_BYTES(bufptr, 1, &zero);
    }
#ifdef RAP_DEBUG
    printf("FLUSHING GET GCS, %ld bytes, %d WORD64s.\n",
           totalSizeInBytes, reply->length);
#endif

    status = WriteToAgent(ri,RapGetGCValuesReply,seq,totalSizeInBytes, outgoing);
    free(outgoing);
    }

    return status;
} /* RapClntGetGCValuesReply */

#endif /* GC_SUPPORT */

/************************************************************************/
/*									*/
/* RapClntGetValuesReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntGetValuesReply(RapInfo 		ri,
		      CARD16		seq,
		      RapObjectID	*object_id,
		      char		**resources,
		      int		num_resources)
{
    Widget		w;
    XtResourceList 	normals,constraints=NULL,requested;
    Cardinal 		num_normals,num_constraints;
    int			*kind;
    register int 	i,j;
    Boolean		found = False;
    CARD32		total_bytes = 0;
    char		*scratch,*bufptr,*val_buf;
    int 		val_size;
    RapStatus		status;
    
    /* Get the list of all the resources the widget has
     */
    RapObjectIDToWidget(*object_id,w);
    XtGetResourceList(XtClass(w), &normals, &num_normals);
    if (XtParent(w) != NULL)
        XtGetConstraintResourceList(XtClass(XtParent(w)), &constraints,
                                    &num_constraints);
    else	
        num_constraints = 0;

    /* Find the resources the agent is requesting
     */
    requested = (XtResourceList)XtMalloc(sizeof(XtResource) * num_resources);
    kind = (int *)XtMalloc(sizeof(int) * num_resources);
    
    memset(requested,'\0',sizeof(XtResource) * num_resources);
    for (i = 0; i < num_resources; i++) {
	found = False;
	j = 0;
	while (!found && (j < num_normals)) {
	    if (!strcmp(resources[i],normals[j].resource_name)) {
		requested[i] = normals[j];
		kind[i] = RapNormalResource;
		found = True;
	    }
	    else
		j++;
	}
	j = 0;
	while (!found && (j < num_constraints)) {
	    if (!strcmp(resources[i],constraints[j].resource_name)) {
		requested[i] = constraints[j];
		kind[i] = RapConstraintResource;
		found = True;
	    }
	    else
		j++;
	}

	if (!found)
	    break;
    }

    /* Calculate the sizes
     */
    total_bytes += 4;			/* Error Code */
    total_bytes += 4;			/* Num Values */
    
    /* Only send resource values if all the resources requested
     * are valid.
     */
    if (found) {
	total_bytes += 4;		/* Num Resources */
	for (i = 0; i < num_resources; i++) {
	    ComputeResourceSize(w,&(requested[i]),True,&val_size,NULL);
	    total_bytes += val_size;
	}
    }
    
    /* Store the object ID, error code, and num resources
     */
    bufptr = scratch = XtMalloc(total_bytes);
    if (found) {
	STORE_CARD32(bufptr,RapNoError);
	STORE_CARD32(bufptr,num_resources);

	/* Store the resources
	 */
	for (i = 0; i < num_resources; i++) {
	    ComputeResourceSize(w,&(requested[i]),True,&val_size,&val_buf);
	    memcpy(bufptr,val_buf,val_size);
	    bufptr += val_size;
	    XtFree(val_buf);
	}
    }
    else {
	STORE_CARD32(bufptr,RapNoSuchResource);
	STORE_CARD32(bufptr,0);
    }

    /* Send the data to the client
     */
    status = WriteToAgent(ri,RapGetValuesReply,seq,total_bytes,scratch);

    /* Free the memory we used
     */
    XtFree(scratch);
    XtFree((char *)normals);
    if (constraints) XtFree((char *)constraints);
    XtFree((char *)requested);
    XtFree((char *)kind);
    
    return status;

} /* RapClntGetValuesReply */

/************************************************************************/
/*									*/
/* RapClntSetValuesReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntSetValuesReply(RapInfo 		ri,
		      CARD16		seq,
		      RapObjectID	*object_id,
		      RapArgList	args,
		      int		num_args)
{
    Widget		w;
    register int 	i;

    RapObjectIDToWidget(*object_id,w);

    /* Set the values for each resource.
     */
    for (i = 0; i < num_args; i++)
	XtVaSetValues(w,
		      XtVaTypedArg,
		      args[i].name,
		      args[i].value->type,
		      args[i].value->value,
		      args[i].value->size,
		      NULL);

    return RapClntAck(ri,seq);

} /* RapClntSetValuesReply */

/************************************************************************/
/*									*/
/* RapCreateHookProc							*/
/*									*/
/*	The Xt-type callback procedure that gets called by the		*/
/*	Hooks object when a widget is created.  The create callback	*/
/*	just before returning from XtCreateWidget, 			*/
/*	XtCreateManagedWidget, XtCreatePopupShell, XtAppCreateShell,	*/
/*	their corresponding varargs versions.				*/
/*								        */
/*      [[[WDW - This is still to be designed.  For now, just send a    */
/*               message that says a hook was called.                   */
/*                                                                      */
/************************************************************************/
void RapCreateHookProc(Widget           hooks_object,
                       XtPointer        client_data,
                       XtCreateHookData callback_data)
{
    RapInfo 		ri = (RapInfo) client_data;
    Widget		w = callback_data->widget;
    RapObjectID		object_id;
    CARD32		total_bytes = 0;
    char		*scratch,*bufptr;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapCreateHookProc","%s\n",
	  callback_data->type);
#endif

    /* Compute the size of the event
     */
    total_bytes += sizeof(RapObjectID);
    total_bytes += STRING_BYTES(strlen(callback_data->type));

    /* Create the event and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    WidgetToRapObjectID(w,object_id);
    STORE_OBJECTID(bufptr,object_id);
    STORE_STRING(bufptr,strlen(callback_data->type),callback_data->type);

    WriteToAgent(ri,RapCreateHookNotify,0,total_bytes,scratch);
    XtFree(scratch);
        
} /* RapCreateHookProc */

/************************************************************************/
/*                                                                      */
/* RapDestroyHookProc                                                   */
/*                                                                      */
/*      The Xt-type callback procedure that gets called when a          */
/*      widget is destroyed as a result of calling XtDestroyWidget.	*/
/*	The callback is called on completion of phase one destroy.      */
/*                                                                      */
/*      [[[WDW - This is still to be designed.  For now, just send a    */
/*               message that says a hook was called.                   */
/*                                                                      */
/************************************************************************/
void RapDestroyHookProc(Widget           	hooks_object,
			XtPointer        	client_data,
			XtDestroyHookData     	callback_data)
{
    RapInfo 		ri = (RapInfo) client_data;
    Widget		w = callback_data->widget;
    RapObjectID		object_id;
    CARD32		total_bytes = 0;
    char		*scratch,*bufptr;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapDestroyHookProc","%s\n",
	  callback_data->type);
#endif

    /* Compute the size of the event
     */
    total_bytes += sizeof(RapObjectID);
    total_bytes += STRING_BYTES(strlen(callback_data->type));

    /* Create the event and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    WidgetToRapObjectID(w,object_id);
    STORE_OBJECTID(bufptr,object_id);
    STORE_STRING(bufptr,strlen(callback_data->type),callback_data->type);

    WriteToAgent(ri,RapDestroyHookNotify,0,total_bytes,scratch);
    XtFree(scratch);
    
} /* RapDestroyHookProc */

/************************************************************************/
/*                                                                      */
/* RapChangeHookProc                                                    */
/*                                                                      */
/*      The Xt-type callback procedure that gets called by the          */
/*      Hooks object when a widget is changed.  The change callback     */
/*      is called from XtSetValues, XtVaSetValues, XtManageChild(ren),  */
/*      XtUnmanageChild, XtUnmanageChildren, XtRealizeWidget,           */
/*      XtUnrealizeWidget, XtAddCallback, XtRemoveCallback,             */
/*      XtAddCallbacks, XtRemoveCallbacks, XtRemoveAllCallbacks,        */
/*      XtAugmentTranslations, XtOverrideTranslations,                  */
/*      XtUninstallTranslations, XtSetKeyboardFocus,                    */
/*      XtSetWMColormapWindows, XtSetMappedWhenManaged, XtMapWidget,    */
/*      XtUnmapWidget, XtPopup, XtPopupSpringLoaded, and XtPopdown.     */
/*                                                                      */
/*      [[[WDW - This is still to be designed.  For now, just send a    */
/*               message that says a hook was called.                   */
/*                                                                      */
/************************************************************************/
void RapChangeHookProc(Widget           hooks_object,
		       XtPointer        client_data,
		       XtChangeHookData callback_data)
{
    RapInfo 		ri = (RapInfo) client_data;
    Widget		w = callback_data->widget;
    RapObjectID		object_id;
    CARD32		total_bytes = 0;
    char		*scratch,*bufptr;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapChangeHookProc","%s\n",
	  callback_data->type);
#endif

    /* If it is XtHsetValues, the client called XtSetValues or XtVaSetValues.
     * The widget is the new widget and the event_data is an 
     * XtChangeHookSetValuesData that holds the set values information.  The 
     * callback is called after the set_values and constraint set_values have 
     * been called.
     */
    if (!strcmp(callback_data->type,XtHsetValues)) {
	XtChangeHookSetValuesData svr;
	svr = (XtChangeHookSetValuesData) callback_data->event_data;
    }
    /* If it is XtHmanageChildren, the client called XtManageChild or 
     * XtManageChildren.  The widget is the parent and the event_data is a
     * WidgetList of the children that were managed.  The count of widgets 
     * in the WidgetList is held in num_event_data.  The callback is called 
     * after the children have been managed.
     */
    else if (!strcmp(callback_data->type,XtHmanageChildren)) {
	WidgetList wl = (WidgetList) callback_data->event_data;
    }
    /* If it is XtHunmanageChildren, the client called XtUnmanageChild or 
     * XtUnmanageChildren.  The widget is the parent and the event_data is a
     * WidgetList of the children that were unmanaged.  The count of widgets 
     * in the WidgetList is held in num_event_data.  The callback is called 
     * after the children have been unmanaged.
     */
    else if (!strcmp(callback_data->type,XtHunmanageChildren)) {
	WidgetList wl = (WidgetList) callback_data->event_data;
    }
    /* If it is XtHunmanageSet, the client called XtChangeManagedSet and
     * this call represents the children being unmanaged.  The widget is 
     * the parent and the event_data is a WidgetList of the children that 
     * were unmanaged.  The count of widgets in the WidgetList is held in 
     * num_event_data.
     */
    else if (!strcmp(callback_data->type,XtHunmanageSet)) {
	WidgetList wl = (WidgetList) callback_data->event_data;
    }
    /* If it is XtHmanageSet, the client called XtChangeManagedSet and
     * this call represents the children being managed.  The widget is 
     * the parent and the event_data is a WidgetList of the children that 
     * were managed.  The count of widgets in the WidgetList is held in 
     * num_event_data.
     */
    else if (!strcmp(callback_data->type,XtHmanageSet)) {
	WidgetList wl = (WidgetList) callback_data->event_data;
    }
    /* If it is XtHrealizeWidget, the client called XtRealizeWidget.
     * The widget is the widget that was realized.  The callback is 
     * called after the widget was realized.
     */
    else if (!strcmp(callback_data->type,XtHrealizeWidget)) {
    }
    /* If it is XtHunrealizeWidget, the client called XtUnrealizeWidget.
     * The widget is the widget that was unrealized.  The callback is 
     * called after the widget was unrealized.
     */
    else if (!strcmp(callback_data->type,XtHunrealizeWidget)) {
    }
    /* If it is XtHaddCallback, the client called XtAddCallback.
     * The widget is the widget to which the callback was added, and
     * event_data is a string that is the name of the callback.  The
     * callback is called after the callback was added.
     */
    else if (!strcmp(callback_data->type,XtHaddCallback)) {
	char *callback_name = (char *)callback_data->event_data;
    }
    /* If it is XtHaddCallbacks, the client called XtAddCallbacks.
     * The widget is the widget to which the callbacks were added, and
     * event_data is a string that is the name of the callback.  The
     * callback is called after the callbacks were added.
     */
    else if (!strcmp(callback_data->type,XtHaddCallbacks)) {
	char *callback_name = (char *)callback_data->event_data;
    }
    /* If it is XtHremoveCallback, the client called XtRemoveCallback.
     * The widget is the widget from which the callback was removed, and
     * event_data is a string that is the name of the callback.  The
     * callback is called after the callback was removed.
     */
    else if (!strcmp(callback_data->type,XtHremoveCallback)) {
	char *callback_name = (char *)callback_data->event_data;
    }
    /* If it is XtHremoveCallbacks, the client called XtRemoveCallbacks.
     * The widget is the widget from which the callbacks were removed, and
     * event_data is a string that is the name of the callback.  The
     * callback is called after the callbacks were removed.
     */
    else if (!strcmp(callback_data->type,XtHremoveCallbacks)) {
	char *callback_name = (char *)callback_data->event_data;
    }
    /* If it is XtHremoveAllCallbacks, the client called XtRemoveAllCallbacks.
     * The widget is the widget from which the callbacks were removed.  The
     * callback is called after the callbacks were removed.
     */
    else if (!strcmp(callback_data->type,XtHremoveAllCallbacks)) {
	char *callback_name = (char *)callback_data->event_data;
    }
    /* If it is XtHremoveAllCallbacks, the client called XtRemoveAllCallbacks.
     * The widget is the widget from which the callbacks were removed.  The
     * callback is called after the callbacks were removed.
     */
    else if (!strcmp(callback_data->type,XtHremoveAllCallbacks)) {
	char *callback_name = (char *)callback_data->event_data;
    }
    /* If it is XtHaugmentTranslations, the client called 
     * XtAugmentTranslations.  The widget is the widget whose translations
     * were modified.  The callback is called after the translations were 
     * modified.
     */
    else if (!strcmp(callback_data->type,XtHaugmentTranslations)) {
    }
    /* If it is XtHoverrideTranslations, the client called 
     * XtOverrideTranslations.  The widget is the widget whose translations
     * were modified.  The callback is called after the translations were 
     * modified.
     */
    else if (!strcmp(callback_data->type,XtHoverrideTranslations)) {
    }
    /* If it is XtHuninstallTranslations, the client called 
     * XtUninstallTranslations.  The widget is the widget whose translations
     * were modified.  The callback is called after the translations were 
     * modified.
     */
    else if (!strcmp(callback_data->type,XtHuninstallTranslations)) {
    }
    /* If it is XtHsetKeyboardFocus, the client called XtSetKeyboardFocus. 
     * The widget is the descendant argument passed to XtSetKeyboardFocus.
     * The callback is called before returning from XtSetKeyboardFocus.
     */
    else if (!strcmp(callback_data->type,XtHsetKeyboardFocus)) {
    }
    /* If it is XtHsetWMColormapWindows, the client called 
     * XtSetWMColormapWindows.  The event_data is a WidgetList of the 
     * children passed to XtSetWMColormapWindows.  The count of widgets 
     * in the WidgetList is held in num_event_data.
     */
    else if (!strcmp(callback_data->type,XtHsetWMColormapWindows)) {
	WidgetList wl = (WidgetList) callback_data->event_data;
	w = NULL;
    }
    /* If it is XtHsetMappedWhenManaged, the client called 
     * XtSetMappedWhenManaged.  The widget is the widget whose 
     * mapped_when_managed field was changed.  The event_data is a 
     * Boolean that is the value of the mapped_when_managed argument passed 
     * to XtSetMappedWhenManaged.  The callback is called after the setting
     * the mapped_when_managed field, but before realizing or unrealizing
     * the widget.
     */
    else if (!strcmp(callback_data->type,XtHsetMappedWhenManaged)) {
	Boolean mwm = (Boolean) callback_data->event_data;
    }
    /* If it is XtHmapWidget, the client called XtMapWidget.  The widget
     * is the widget that was mapped.  The callback is called after the
     * widget was mapped.
     */
    else if (!strcmp(callback_data->type,XtHmapWidget)) {
    }
    /* If it is XtHunmapWidget, the client called XtUnmapWidget.  The widget
     * is the widget that was unmapped.  The callback is called after the
     * widget was unmapped.
     */
    else if (!strcmp(callback_data->type,XtHunmapWidget)) {
    }
    /* If it is XtHpopup, the client called XtPopup.  The widget
     * is the widget that is being popped up.  The event_data is the
     * grab kind passed to XtPopup.  The callback is called before
     * returning from XtPopup.
     */
    else if (!strcmp(callback_data->type,XtHpopup)) {
	XtGrabKind gk = (XtGrabKind) callback_data->event_data;
    }
    /* If it is XtHpopupSpringLoaded, the client called XtPopupSpringLoaded.  
     * The widget is the widget that is being popped up.  The callback is 
     * called before returning from XtPopupSpringLoaded.
     */
    else if (!strcmp(callback_data->type,XtHpopupSpringLoaded)) {
    }
    /* If it is XtHpopdown, the client called XtPopdown.  The widget
     * is the widget that is being popped up.  The event_data is the
     * grab kind passed to XtPopdown.  The callback is called before
     * returning from XtPopdown.
     */
    else if (!strcmp(callback_data->type,XtHpopdown)) {
    }

    /* Compute the size of the event
     */
    total_bytes += sizeof(RapObjectID);
    total_bytes += STRING_BYTES(strlen(callback_data->type));

    /* Create the event and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    WidgetToRapObjectID(w,object_id);
    STORE_OBJECTID(bufptr,object_id);
    STORE_STRING(bufptr,strlen(callback_data->type),callback_data->type);

    WriteToAgent(ri,RapChangeHookNotify,0,total_bytes,scratch);
    XtFree(scratch);
        
} /* RapChangeHookProc */

/************************************************************************/
/*                                                                      */
/* RapConfigureHookProc                                                 */
/*                                                                      */
/*      The Xt-type callback procedure that gets called any time the    */
/*      Intrinsics move, resize, or configure a widget and when         */
/*      XtResizeWindow is called.  The callback is called after the     */
/*      changes have been made to the widget.                           */
/*                                                                      */
/*      [[[WDW - This is still to be designed.  For now, just send a    */
/*               message that says a hook was called.                   */
/*                                                                      */
/************************************************************************/
void RapConfigureHookProc(Widget           	hooks_object,
			  XtPointer        	client_data,
			  XtConfigureHookData   callback_data)
{
    RapInfo 		ri = (RapInfo) client_data;
    Widget		w = callback_data->widget;
    RapObjectID		object_id;
    CARD32		total_bytes = 0;
    char		*scratch,*bufptr;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapConfigureHookProc","%s\n",
	  callback_data->type);
#endif

    /* Compute the size of the event
     */
    total_bytes += sizeof(RapObjectID);
    total_bytes += STRING_BYTES(strlen(callback_data->type));

    /* Create the event and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    WidgetToRapObjectID(w,object_id);
    STORE_OBJECTID(bufptr,object_id);
    STORE_STRING(bufptr,strlen(callback_data->type),callback_data->type);

    WriteToAgent(ri,RapConfigureHookNotify,0,total_bytes,scratch);
    XtFree(scratch);
    
} /* RapConfigureHookProc */

/************************************************************************/
/*                                                                      */
/* RapGeometryHookProc                                                  */
/*                                                                      */
/*      The Xt-type callback procedure that gets called from            */
/*      XtMakeGeometryRequest and XtMakeResizeRequest once before and   */
/*      once after geometry negotiation occurs.                         */
/*                                                                      */
/*      [[[WDW - This is still to be designed.  For now, just send a    */
/*               message that says a hook was called.                   */
/*                                                                      */
/************************************************************************/
void RapGeometryHookProc(Widget           	hooks_object,
			 XtPointer        	client_data,
			 XtGeometryHookData     callback_data)
{
    RapInfo 		ri = (RapInfo) client_data;
    Widget		w = callback_data->widget;
    RapObjectID		object_id;
    CARD32		total_bytes = 0;
    char		*scratch,*bufptr;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapGeometryHookProc","%s\n",
	  callback_data->type);
#endif

    /* Compute the size of the event
     */
    total_bytes += sizeof(RapObjectID);
    total_bytes += STRING_BYTES(strlen(callback_data->type));

    /* Create the event and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    WidgetToRapObjectID(w,object_id);
    STORE_OBJECTID(bufptr,object_id);
    STORE_STRING(bufptr,strlen(callback_data->type),callback_data->type);

    WriteToAgent(ri,RapGeometryHookNotify,0,total_bytes,scratch);
    XtFree(scratch);
    
} /* RapGeometryHookProc */

/************************************************************************/
/*									*/
/* RapActionHookProc							*/
/*									*/
/*	The procedure that gets called just before a widget's action	*/
/*	proc is called.							*/
/*								        */
/************************************************************************/
void RapActionHookProc(Widget 		w, 
		       XtPointer 	client_data,
		       String		action_name,
		       XEvent		*event,
		       String		*params,
		       Cardinal		*num_params)
{
    RapInfo 		ri = (RapInfo) client_data;
    RapObjectID		object_id;
    int			i,total_bytes = 0;
    char		*scratch,*bufptr;
    CARD32		pad = 0;
   
#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapActionHookProc","%s\n",action_name);
#endif

    /* Compute the size of the event
     */
    total_bytes += sizeof(RapObjectID);			/* ObjectID   */
    total_bytes += STRING_BYTES(strlen(action_name));	/* name	      */
    total_bytes += PADDED_BYTES64(sizeof(XEvent));	/* event      */
    total_bytes += 4;	 			        /* num params */
    total_bytes += 4;					/* pad	      */
    for (i = 0; i < *num_params; i++)
	total_bytes += STRING_BYTES(strlen(params[i]));

    /* Create the event and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    WidgetToRapObjectID(w,object_id);
    STORE_OBJECTID(bufptr,object_id);
    STORE_STRING(bufptr,strlen(action_name),action_name);
    STORE_SIZED_BYTES(bufptr,sizeof(XEvent),(char *)event);
    STORE_CARD32(bufptr,*num_params);
    STORE_CARD32(bufptr,pad);
    for (i = 0; i < *num_params; i++)
	STORE_STRING(bufptr,strlen(params[i]),params[i]);

    WriteToAgent(ri,RapActionHookNotify,0,total_bytes,scratch);
    XtFree(scratch);

} /* RapActionHookProc */

/************************************************************************/
/*									*/
/* RapClntGetHooksReply							*/
/*									*/
/************************************************************************/
RapStatus
RapClntGetHooksReply(RapInfo ri,
		     CARD16  seq)
{
    return WriteToAgent(ri,RapGetHooksReply,seq,
			sizeof(CARD16),(char *)&ri->hooks);

} /* RapClntGetHooksReply */

/************************************************************************/
/*									*/
/* RapClntRemoveCallbacks						*/
/*									*/
/* 	Removes all callbacks assigned by an agent.			*/
/*									*/
/************************************************************************/
void
RapClntRemoveCallbacks(RapInfo ri)
{
    Widget hooks_object = XtHooksOfDisplay(XtDisplay(ri->top_level));
    int	   bit;
    
    while (ri->hooks) {
	bit = ffs(ri->hooks) - 1;
	switch(bit) {
	    case RapCreateHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RemoveCallbacks",
		      "removing RapCreateHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNcreateHook,
				 (XtCallbackProc)RapCreateHookProc,
				 ri);
		break;
	    }

	    case RapDestroyHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RemoveCallbacks",
		      "removing RapDestroyHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNdestroyHook,
				 (XtCallbackProc)RapDestroyHookProc,
				 ri);
		break;
	    }

	    case RapChangeHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RemoveCallbacks",
		      "removing RapChangeHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNchangeHook,
				 (XtCallbackProc)RapChangeHookProc,
				 ri);
		break;
	    }

	    case RapConfigureHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RemoveCallbacks",
		      "removing RapConfigureHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNconfigureHook,
				 (XtCallbackProc)RapConfigureHookProc,
				 ri);
		break;
	    }

	    case RapGeometryHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RemoveCallbacks",
		      "removing RapGeometryHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNgeometryHook,
				 (XtCallbackProc)RapGeometryHookProc,
				 ri);
		break;
	    }

	    case RapActionHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RemoveCallbacks",
		      "removing RapActionHook.\n");
#endif
		if (ri->action_hook != NULL)
		    XtRemoveActionHook(ri->action_hook);
		ri->action_hook = NULL;
		break;
	    }
	}
	ri->hooks &= ~(1 << bit);
    }

} /* RapClntRemoveCallbacks */

/************************************************************************/
/*									*/
/* RapClntSelectHooksReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntSelectHooksReply(RapInfo ri,
			CARD16	seq,
			CARD16	hooks,
			CARD16	mask)
{
    CARD16 new_hooks,removed,added;
    int	   bit;
    Widget hooks_object = XtHooksOfDisplay(XtDisplay(ri->top_level));

    new_hooks = (ri->hooks & (~mask)) | (hooks & mask);
    removed   = (ri->hooks ^ (new_hooks)) & ri->hooks;
    added     = (ri->hooks ^ (new_hooks)) & new_hooks;
    
    /* Remove callbacks
     */
    while (removed) {
	bit = ffs(removed) - 1;
	switch(bit) {
	    case RapCreateHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "turning off RapCreateHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNcreateHook,
				 (XtCallbackProc)RapCreateHookProc,
				 ri);
		break;
	    }

	    case RapDestroyHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "turning off RapDestroyHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNdestroyHook,
				 (XtCallbackProc)RapDestroyHookProc,
				 ri);
		break;
	    }

	    case RapChangeHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "turning off RapChangeHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNchangeHook,
				 (XtCallbackProc)RapChangeHookProc,
				 ri);
		break;
	    }

	    case RapConfigureHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "turning off RapConfigureHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNconfigureHook,
				 (XtCallbackProc)RapConfigureHookProc,
				 ri);
		break;
	    }

	    case RapGeometryHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "turning off RapGeometryHook.\n");
#endif
		XtRemoveCallback(hooks_object,
				 XtNgeometryHook,
				 (XtCallbackProc)RapGeometryHookProc,
				 ri);
		break;
	    }

	    case RapActionHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "turning off RapActionHook.\n");
#endif
		if (ri->action_hook != NULL)
		    XtRemoveActionHook(ri->action_hook);
		ri->action_hook = NULL;
		break;
	    }
	}
	removed &= ~(1 << bit);
    }
    
    /* Add callbacks if necessary
     */
    while (added) {
	bit = ffs(added) - 1;
	switch(bit) {
	    case RapCreateHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "adding RapCreateHook.\n");
#endif
		XtAddCallback(hooks_object,
			      XtNcreateHook,
			      (XtCallbackProc)RapCreateHookProc,
			      ri);
		break;
	    }

	    case RapDestroyHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "adding RapDestroyHook.\n");
#endif
		XtAddCallback(hooks_object,
			      XtNdestroyHook,
			      (XtCallbackProc)RapDestroyHookProc,
			      ri);
		break;
	    }

	    case RapChangeHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "adding RapChangeHook.\n");
#endif
		XtAddCallback(hooks_object,
			      XtNchangeHook,
			      (XtCallbackProc)RapChangeHookProc,
			      ri);
		break;
	    }

	    case RapConfigureHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "adding RapConfigureHook.\n");
#endif
		XtAddCallback(hooks_object,
			      XtNconfigureHook,
			      (XtCallbackProc)RapConfigureHookProc,
			      ri);
		break;
	    }

	    case RapGeometryHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "adding RapGeometryHook.\n");
#endif
		XtAddCallback(hooks_object,
			      XtNgeometryHook,
			      (XtCallbackProc)RapGeometryHookProc,
			      ri);
		break;
	    }

	    case RapActionHook: {
#ifdef DEBUG
		InfoF(__FILE__,__LINE__,"RapClntSelectHooksReply",
		      "adding RapActionHook.\n");
#endif
		ri->action_hook = XtAppAddActionHook(
		     XtWidgetToApplicationContext(ri->top_level),
		     RapActionHookProc,
		     ri);
		break;
	    }
	}
	added &= ~(1 << bit);
    }

    ri->hooks = new_hooks;
    
    return RapClntAck(ri,seq);

} /* RapClntSelectHooksReply */


/************************************************************************/
/*									*/
/* RapClntGetEventsReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntGetEventsReply(RapInfo ri,
		      CARD16  seq)
{
    CARD32		total_bytes = 0;
    char		*scratch,*bufptr;
    int			i;
    RapStatus		status;
    
    /* Compute the size of the reply 
     */
    total_bytes += (4 * sizeof(CARD32));	/* events */

    /* Create the reply and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    for (i = 0; i < 4; i++)
        STORE_CARD32(bufptr,ri->events[i]);
    
    status = WriteToAgent(ri,RapGetEventsReply,seq,total_bytes,scratch);
    XtFree(scratch);
    
    return status;

} /* RapClntGetEventsReply */


/************************************************************************/
/*									*/
/* RapClntEventProc							*/
/*									*/
/************************************************************************/
Bool
RapClntEventProc(Display 	*dpy, 
		 XEvent 	*event, 
		 xEvent 	*wire)
{
    XEDataObject	ext_object;
    XExtData		*ext_data;
    RapInfo		ri;
    CARD32		total_bytes = 0;
    Bool		status,called=False;

#ifdef DEBUG
    InfoF(__FILE__,__LINE__,"RapClntEventProc",
	  "just got an event of type %d.\n",wire->u.u.type);
#endif

    ext_object.display = dpy;
    ext_data = XFindOnExtensionList(XEHeadOfExtensionList(ext_object),
				    rap_ext_codes->extension);

    ri = (RapInfo) ext_data->private_data;
    while (ri != NULL) {
	if (ri->event_vec[wire->u.u.type] != NULL) {
	    WriteToAgent(ri,RapEventNotify,0,
			 sizeof(xEvent), (char *)wire);

	    /* Only convert it once per call, not once per agent that
	     * requested the event.
	     */
	    if (!called) {
		called = True;
		status = (*ri->event_vec[wire->u.u.type])(dpy,event,wire);
	    }
	}
	ri = ri->next;
    }
    
    return status;
    
} /* RapClntEventProc */

/************************************************************************/
/*									*/
/* RapClntSelectEventsReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntSelectEventsReply(RapInfo ri,
			 CARD16	seq,
			 CARD32	events[4],
			 CARD32	mask[4])
{
    CARD32  new_events,removed,added;
    int	    i,bit,event;
    Widget  hooks_object = XtHooksOfDisplay(XtDisplay(ri->top_level));
    Display *dpy = XtDisplay(ri->top_level);

    for (i = 0; i < 4; i++) {
	new_events = (ri->events[i] & (~mask[i])) | (events[i] & mask[i]);
	removed    = (ri->events[i] ^ (new_events)) & ri->events[i];
	added      = (ri->events[i] ^ (new_events)) & new_events;
    
	/* Remove callbacks
	 */
	while (removed) {
	    bit = ffs(removed) - 1;
	    event = (i * 32) + bit;
#ifdef DEBUG
	    InfoF(__FILE__,__LINE__,"RapClntSelectEventsReply",
		  "turning off event %d.\n",event);
#endif
	    XESetWireToEvent(dpy,event,ri->event_vec[event]);
	    ri->event_vec[event] = NULL;
	    removed &= ~(1 << bit);
	}
    
	/* Add callbacks if necessary
	 */
	while (added) {
	    bit = ffs(added) - 1;
	    event = (i * 32) + bit;
#ifdef DEBUG
	    InfoF(__FILE__,__LINE__,"RapClntSelectEventsReply",
		  "turning on event %d.\n",event);
#endif
	    ri->event_vec[event] = XESetWireToEvent(dpy,event,RapClntEventProc);
	    added &= ~(1 << bit);
	}

	ri->events[i] = new_events;
    }
    
    return RapClntAck(ri,seq);

} /* RapClntSelectEventsReply */


/************************************************************************/
/*									*/
/* RapClntGetRequestsReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntGetRequestsReply(RapInfo ri,
			CARD16  seq)
{
    CARD32		total_bytes = 0;
    char		*scratch,*bufptr;
    int			i;
    RapStatus		status;
    
    /* Compute the size of the reply 
     */
    total_bytes += (8 * sizeof(CARD32)); /* requests */

    /* Create the reply and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    for (i = 0; i < 8; i++)
        STORE_CARD32(bufptr,ri->requests[i]);

    status = WriteToAgent(ri,RapGetRequestsReply,seq,total_bytes,scratch);
    XtFree(scratch);
    
    return status;

} /* RapClntGetRequestsReply */


/************************************************************************/
/*									*/
/* RapClntFlushProc							*/
/*									*/
/************************************************************************/
void
RapClntFlushProc(Display 	*dpy, 
		 XExtCodes	*ext, 
		 char 		*data, 
		 long 		len)
{
    XEDataObject	ext_object;
    XExtData		*ext_data;
    RapInfo		ri_head,ri;
    xReq 		*request;
    CARD32		total_bytes = 0;
    long		used_len = 0;
    
    ext_object.display = dpy;
    ext_data = XFindOnExtensionList(XEHeadOfExtensionList(ext_object),
				    rap_ext_codes->extension);
    ri_head = (RapInfo) ext_data->private_data;

    while (used_len < len) {	
	request = (xReq *) (data + used_len);
	used_len += request->length * 4;

	/* sanity check 
	 */
	if (request->length < 1) break; 

#ifdef DEBUG2 /* blsx: removing this for now */
	InfoF(__FILE__,__LINE__,"RapClntFlushProc",
	      "just got a request of type %d.\n",request->reqType);
#endif

	/* Send the requests to all interested agents, one request at a 
	 * time.
	 */
	ri = ri_head;
	while (ri) {
	    if (ri->requests[request->reqType/32] & 
		(1 << (request->reqType % 32))) {
		WriteToAgent(ri,RapRequestNotify,0,
			     request->length * 4, (char *)request);
	    }
	    ri = ri->next;
	}
    }

} /* RapClntFlushProc */


/************************************************************************/
/*									*/
/* RapClntSelectRequestsReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntSelectRequestsReply(RapInfo ri,
			   CARD16  seq,
			   CARD32  requests[8],
			   CARD32  mask[8])
{
    int	   i;

    for (i = 0; i < 8; i++)
	ri->requests[i] = (ri->requests[i] & (~mask[i])) | (requests[i] & mask[i]);
    
    return RapClntAck(ri,seq);

} /* RapClntSelectRequestsReply */

/************************************************************************/
/*									*/
/* PositionInChild -  returns true if this location is in the child.	*/
/*									*/
/************************************************************************/
static Boolean
PositionInChild(Widget 		child, 
		Position	x, 
		Position	y)
{
    Arg 	args[6];
    Cardinal 	num;
    Dimension 	width, height, border_width;
    Position 	child_x, child_y;
    Boolean 	mapped_when_managed;

    /* It must be at least a rect object.
     */
    if (!XtIsRectObj(child))
	return(FALSE);

    num = 0;
    XtSetArg(args[num], XtNmappedWhenManaged, 	&mapped_when_managed); 	num++;
    XtSetArg(args[num], XtNwidth, 		&width); 		num++;
    XtSetArg(args[num], XtNheight, 		&height); 		num++;
    XtSetArg(args[num], XtNx, 			&child_x); 		num++;
    XtSetArg(args[num], XtNy, 			&child_y);		num++;
    XtSetArg(args[num], XtNborderWidth,        	&border_width); 	num++;
    XtGetValues(child, args, num);
 
    /* The only way we will know of the widget is mapped is to see if
     * mapped when managed is True and this is a managed child.  Otherwise
     * we will have to ask the server if this window is mapped.
     */
    if (XtIsWidget(child) && !(mapped_when_managed && XtIsManaged(child)) ) {
	XWindowAttributes attrs;

	if (XGetWindowAttributes(XtDisplay(child), 
				 XtWindow(child), &attrs) != 0) {
	    /* oops */
	}
	else if (attrs.map_state != IsViewable)
	    return(FALSE);
    }

    return ((x >= child_x) &&
	    (x <= (child_x + (Position)width + 2 * (Position)border_width)) &&
	    (y >= child_y) &&
	    (y <= (child_y + (Position)height + 2 * (Position)border_width)));

} /* PositionInChild */

/************************************************************************/
/*									*/
/* FindChild - Finds the child that actually contains the point shown.	*/
/*									*/
/************************************************************************/
static Widget 
FindChild(Widget	parent, 
	  Position	x,
	  Position	y)
{
    Widget 	*children;
    int 	i = FindChildren(parent, &children, TRUE, FALSE);

    while (i > 0) {
	i--;
	if (PositionInChild(children[i], x, y)) {
	    Widget child = children[i];
	    XtFree((char *)children);
	    return(FindChild(child, x - child->core.x, y - child->core.y));
	}
    }

    XtFree((char *)children);
    return(parent);

} /* FindChild */

/************************************************************************/
/*									*/
/* DoFindChild - Finds the child that contains the location specified.	*/
/*									*/
/************************************************************************/
static Widget
DoFindChild(Widget	w, 
	    Position	x,
	    Position	y)
{
    Position 		parent_x, parent_y;
    
    XtTranslateCoords(w, 
		      (Position) 0, (Position) 0,
		      &parent_x, &parent_y);
    
    return(FindChild(w, 
		     x - parent_x,
		     y - parent_y));
   
} /* DoFindChild */

/************************************************************************/
/*									*/
/* RapClntLocateObjectReply						*/
/*									*/
/************************************************************************/
RapStatus
RapClntLocateObjectReply(RapInfo 	ri,
			 CARD16		seq,
			 RapObjectID	*object_id,
			 Position	x,
			 Position	y)
{
    Widget		w,found_w;
    int			total_bytes = 0;
    char		*scratch,*bufptr;
    RapStatus		status;
    
    /* Find the object
     */
    RapObjectIDToWidget(*object_id,w);
    found_w = DoFindChild(w,x,y);

    /* Compute the size of the reply
     */
    total_bytes += ComputeObjectSize(found_w,False);

    /* Create the reply and send it
     */
    bufptr = scratch = XtMalloc(total_bytes);
    StoreObject(found_w,&bufptr,False);

    status = WriteToAgent(ri,RapLocateObjectReply,seq,total_bytes,scratch);
    XtFree(scratch);

    return status;

} /* RapClntLocateObjectReply */

