/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
* $Log: RapUtils.c,v $
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*									*/
/* RapUtils.c								*/
/*									*/
/*	Utilities common to both agents and clients.			*/
/*									*/
/************************************************************************/
#include <stdio.h>
#include <stdarg.h>
#include <sys/time.h>

/************************************************************************/
/*									*/
/* GetTimeInMillis							*/
/*									*/
/* 	Returns time of day in milliseconds.				*/
/*									*/
/************************************************************************/
int 
GetTimeInMillis()
{
    struct timeval  tp;
    gettimeofday(&tp, 0);
    return(tp.tv_sec * 1000) + (tp.tv_usec / 1000);

} /* GetTimeInMillis */

/************************************************************************/
/*									*/
/* ErrorF								*/
/*									*/
/*	Prints out an error message.					*/
/*									*/
/************************************************************************/
void    
ErrorF(char *file, int line, char *func, char *fmt, ...)
{
    va_list args;
    
    /* Display the name of the function that called error
     */
    fprintf(stderr,"ERROR: (%s:%d) %x %s: ",
	    file,line,GetTimeInMillis(),func);

    /* Display the remainder of the message
     */
    va_start(args,fmt);
    vfprintf(stderr,fmt,args);
    va_end(args);

    fflush(stderr);
    
} /* ErrorF */

/************************************************************************/
/*									*/
/* InfoF								*/
/*									*/
/*	Prints out an information message.				*/
/*									*/
/************************************************************************/
void        
InfoF(char *file, int line, char *func, char *fmt, ...)
{
    va_list args;
    
    /* Display the name of the function that called error
     */
    fprintf(stdout,"INFO: (%s:%d) %x %s: ",
	    file,line,GetTimeInMillis(),func);

    /* Display the remainder of the message
     */
    va_start(args,fmt);
    vfprintf(stdout,fmt,args);
    va_end(args);

    fflush(stdout);
    
} /* InfoF */
