#ifndef RAP_H
#define RAP_H
/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
* $Log: Rap.h,v $
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*									*/
/* Rap.h								*/
/*									*/
/*	Header file for a simple protocol.				*/
/*     									*/
/************************************************************************/

/************************************************************************/
/*									*/
/* RapObjectID								*/
/*									*/
/*	Abstraction of Widget ID.					*/
/*									*/
/************************************************************************/
typedef struct _RapObjectID {
#if defined(LONG64) || defined(WORD64)
    unsigned long id;
#else
    unsigned long id;
    CARD32	  pad;
#endif
} RapObjectID, *RapObjectList;

#ifdef GC_SUPPORT
/************************************************************************/
/*									*/
/* RapGCValues								*/
/*									*/
/************************************************************************/
typedef struct _RapGCValues{
    CARD32              gc;
    CARD32              errorCode;
    union {
        char            *errorString;
        XGCValues       gcv;
    } data;
} RapGCValues, *RapGCValuesPtr;
 
 /************************************************************************/
/*									*/
/* RapGCValuesList								*/
/*									*/
/************************************************************************/
typedef struct _RapGCValuesList {
    CARD32              numGCs;
    RapGCValuesPtr        GCValues;
} RapGCValuesList;
 
#endif /* GC_SUPPORT */

/************************************************************************/
/*									*/
/* RapValue								*/
/*									*/
/************************************************************************/
typedef struct _RapValue 
{
    char  	*type;
    int		size;
    XtPointer	value;
} RapValue, *RapValuePtr;

/************************************************************************/
/*									*/
/* RapArg								*/
/*									*/
/************************************************************************/
typedef struct _RapArg
{
    char  	*name;
    RapValuePtr	value;
} RapArg, *RapArgList;

/************************************************************************/
/*									*/
/* RapStatus codes							*/
/*									*/
/************************************************************************/
typedef enum {
    RapSuccess			= 0,
    RapInvalidIO		= 1,
    RapBadAlloc			= 2,
    RapUnknownResource		= 3
} RapStatus;

/************************************************************************/
/*									*/
/* Rap error codes -- INCOMPLETE					*/
/*									*/
/************************************************************************/
typedef enum {
    RapNoError			= 0,
    RapNoSuchObject		= 1,
    RapNoSuchResource		= 2,
    RapNoSuchGraphicsContext	= 3,
    RapCannotConvertType	= 4,
    RapErrorAddNotify		= 5
} RapErrorTypes;

/************************************************************************/
/*									*/
/* Boolean types for resource kinds (see RapGetResourcesReply)		*/
/*									*/
/************************************************************************/
typedef enum {
    RapNormalResource 		= 0,
    RapConstraintResource 	= 1
} RapResourceKind;


/************************************************************************/
/*									*/
/* Special tokens for windows						*/
/*									*/
/************************************************************************/
typedef enum {
    RapUnrealized		= 0,
    RapGadget			= 2
} RapWindowTypes;

/************************************************************************/
/*									*/
/* Enumerations for the hook callbacks					*/
/*									*/
/************************************************************************/
typedef enum {
    RapCreateHook		= 1,
    RapDestroyHook		= 2,
    RapChangeHook		= 3,
    RapConfigureHook		= 4,
    RapGeometryHook		= 5,
    RapActionHook		= 6
} HookCallbacks;

typedef enum {
    RapCreateHookMask		= 1 << RapCreateHook,
    RapDestroyHookMask		= 1 << RapDestroyHook,
    RapChangeHookMask		= 1 << RapChangeHook,
    RapConfigureHookMask	= 1 << RapConfigureHook,
    RapGeometryHookMask		= 1 << RapGeometryHook,
    RapActionHookMask		= 1 << RapActionHook
} HookCallbackMasks;

#endif /* RAP_H */


