#ifndef RAPAGNT_H
#define RAPAGNT_H
/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
* $Log: RapAgnt.h,v $
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*									*/
/* RapAgnt.h								*/
/*									*/
/*      Rap is a simple ICE subprotocol that uses Xt hooks to share	*/
/*	share client information with external agents.	This file	*/
/*	defines the RapAgnt widget for use by agents.			*/
/*									*/
/************************************************************************/
#include <X11/ICE/ICElib.h>
#include "Rap.h"
#include "RapProto.h"

/* New resources
 */
#define RapAgntNclientCallback		"RapAgntclientCallback"
#define RapAgntNobjectInfoCallback	"RapAgntobjectInfoCallback"
#define RapAgntNhookCallback		"RapAgnthookCallback"
#define RapAgntNeventCallback		"RapAgnteventCallback"
#define RapAgntNrequestCallback		"RapAgntrequestCallback"
#define RapAgntNknownClients		"RapAgntknownClients"
#define RapAgntCKnownClients		"RapAgntKnownClients"
#define RapAgntNautoInitiate		"RapAgntautoInitiate"

/* Class record pointer
 */
externalref     WidgetClass             rapAgntObjectClass;

/************************************************************************/
/*									*/
/* RapAgntClient							*/
/*									*/
/*	A table to hold the list of known clients.  Each entry contains	*/
/*	the client's top level window as well as the ice_connection for	*/
/*	the client.							*/
/*									*/
/************************************************************************/
typedef struct _RapAgntClient
{
    CARD32			window;
    IceConn			ice_conn;
    XtPointer			client_data;
    struct _RapAgntClient	*next;
} RapAgntClient, *RapAgntClientPtr;

/************************************************************************/
/*									*/
/* RapAgntResource							*/
/*									*/
/************************************************************************/
typedef struct _RapAgntResource
{
    char			*name;
    char			*class_name;
    char			*type;
    RapResourceKind		kind;
    RapValuePtr			value;
} RapAgntResource, *RapAgntResourcePtr;

/************************************************************************/
/*									*/
/* RapAgntResourceList							*/
/*									*/
/************************************************************************/
typedef struct _RapAgntResourceList
{
    int				num_resources;
    RapAgntResourcePtr		resources;
} RapAgntResourceList;

/************************************************************************/
/*									*/
/* RapAgntObject							*/
/*									*/
/************************************************************************/
#define ALLOC_INC 6
typedef struct _RapAgntObject
{
    RapObjectID			id;
    struct _RapAgntObject	*parent;
    RapAgntClientPtr		client;
    char			*name;
    char			*class_name;
    CARD32			window;
    CARD32			managed;
    char			*toolkit;
    int				num_resources;
    RapAgntResourcePtr		resources;
    int				alloc_children;
    int				num_children;
    struct _RapAgntObject	*children;
    XtPointer			client_data;
} RapAgntObject, *RapAgntObjectPtr;

/************************************************************************/
/*									*/
/* RapAgntShellList							*/
/*									*/
/************************************************************************/
typedef struct _RapAgntShellList
{
    int				num_shells;
    RapAgntObjectPtr		shells;
} RapAgntShellList;

/************************************************************************/
/*									*/
/* Callback Reasons							*/
/*									*/
/************************************************************************/
typedef enum {
    RapAgntNewClient	 	= 0,
    RapAgntClientConnect	= 1,
    RapAgntInitiateFailed	= 2,
    RapAgntClientDisconnect  	= 3,
    RapAgntDestroy		= 4,
    RapAgntGetHooksReply	= 5,
    RapAgntHookNotify		= 6,
    RapAgntEventNotify		= 7,
    RapAgntGetEventsReply	= 8,
    RapAgntRequestNotify	= 9,
    RapAgntGetRequestsReply	= 10,
    RapAgntLocateObjectReply	= 11
} RapAgntCallbackReasons;
    
/************************************************************************/
/*									*/
/* RapAgntClientCallbackStruct						*/
/*									*/
/************************************************************************/
typedef struct {
    int		 	reason;	
    RapAgntClientPtr	client;
} RapAgntClientCallbackStruct;

/************************************************************************/
/*									*/
/* RapAgntObjectInfoCallbackStruct					*/
/*									*/
/************************************************************************/
typedef struct {
    int			reason;	
    CARD16		sequence_number;
    RapAgntClientPtr	client;
    int			error_code;
    XtPointer		data;
} RapAgntObjectInfoCallbackStruct;

/************************************************************************/
/*									*/
/* RapAgntActionHookCallbackStruct					*/
/*									*/
/************************************************************************/
typedef struct {
    String	action_name;
    XEvent	*event;
    int		num_params;
    String	*params;
} RapAgntActionHookCallbackStruct, *RapAgntActionHookCallbackData;

/************************************************************************/
/*									*/
/* RapAgntHookCallbackStruct						*/
/*									*/
/************************************************************************/
typedef struct {
    int 		reason;	
    CARD16		sequence_number;
    RapAgntClientPtr	client;
    RapAgntObjectPtr	object;
    CARD16		hook;
    XtPointer		data;
} RapAgntHookCallbackStruct;

/************************************************************************/
/*									*/
/* RapAgntEventCallbackStruct						*/
/*									*/
/************************************************************************/
typedef struct {
    int 		reason;	
    CARD16		sequence_number;
    RapAgntClientPtr	client;
    XtPointer		data;
} RapAgntEventCallbackStruct;

/************************************************************************/
/*									*/
/* RapAgntRequestCallbackStruct						*/
/*									*/
/************************************************************************/
typedef struct {
    int 		reason;	
    CARD16		sequence_number;
    RapAgntClientPtr	client;
    XtPointer		data;
} RapAgntRequestCallbackStruct;

/************************************************************************/
/*									*/
/* RapAgntGetProc							*/
/*									*/
/* 	Really hacky procedure to get the RapAgntObject ID from the	*/
/* 	client.  This is necessary because ICE does not allow us to set	*/
/* 	up the client data the way we need it set up (or at least I	*/
/* 	haven't figured out a way).					*/
/*									*/
/************************************************************************/
typedef Widget (*RapAgntGetProc)(void);

/************************************************************************/
/*									*/
/* Function Prototypes							*/
/*		 							*/
/************************************************************************/
Widget 
RapAgntCreate(Widget parent, char *name, 
	      ArgList args, Cardinal num_args,
	      RapAgntGetProc get_proc);

void 
RapAgntContactExistingClients(Widget w);

RapAgntClientPtr
RapAgntFindClient(Widget 		w,
		     CARD32 		window,
		     IceConn		ice_conn);

/* Object Utilities
 */
RapAgntObjectPtr
RapAgntFindObject(RapAgntObjectPtr root, RapObjectID *id);

void
RapAgntFreeObjectInfo(RapAgntObjectPtr object);

void
RapAgntFreeResourceInfo(RapAgntObjectPtr object);


/* Requests
 */
CARD16
RapAgntQueryTree(Widget			w,
		 RapAgntClientPtr	client,
		 Boolean		get_resources,
		 RapAgntShellList	*shell_list);

CARD16
RapAgntGetResources(Widget		w,
		    RapAgntClientPtr	client,
		    RapObjectID		*object_id,
		    RapAgntResourceList	*res_list);

#ifdef GC_SUPPORT
CARD16
RapAgntGetGCValues(Widget			w,
		 RapAgntClientPtr	client,
		   int numGCs,
		   GContext *GCs,
		   RapGCValuesList *reply);
#endif

CARD16
RapAgntGetValues(Widget			w,
		 RapAgntClientPtr	client,
		 RapObjectID		*object_id,
		 char			**resource_names,
		 int			num_resources,
		 RapValuePtr		*values);

CARD16
RapAgntSetValues(Widget			w,
		 RapAgntClientPtr	client,
		 RapObjectID		*object_id,
		 RapArgList		args,
		 int			num_args);

CARD16
RapAgntGetHooks(Widget	       		w,
		RapAgntClientPtr	client,
		CARD16			*hooks);

CARD16
RapAgntSelectHooks(Widget		w,
		   RapAgntClientPtr	client,
		   CARD16		hooks,
		   CARD16		mask);

CARD16
RapAgntGetEvents(Widget	       		w,
		 RapAgntClientPtr	client,
		 CARD32			events[4]);

CARD16
RapAgntSelectEvents(Widget		w,
		    RapAgntClientPtr	client,
		    CARD32		events[4],
		    CARD32		mask[4]);

CARD16
RapAgntGetRequests(Widget	       	w,
		   RapAgntClientPtr	client,
		   CARD32		requests[8]);

CARD16
RapAgntSelectRequests(Widget		w,
		      RapAgntClientPtr	client,
		      CARD32		requests[8],
		      CARD32		mask[8]);


CARD16
RapAgntLocateObject(Widget		w,
		    RapAgntClientPtr	client,
		    RapObjectID		*object_id,
		    Position		xpos,
		    Position		ypos,
		    RapAgntObjectPtr	*object);

#endif /* RAPAGNT_H */

