#ifndef RAPPROTO_H
#define RAPPROTO_H
/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
//
//      Copyright 1992, 1993, 1994, 1995
//      Georgia Tech Research Corporation
//      All rights reserved.
//
*/
/*
* $Log: RapProto.h,v $
* Revision 0.1  1997/09/02 20:49:03  rcsg
* Initial revision
*
*/

/************************************************************************/
/*									*/
/* RapProto.h								*/
/*									*/
/*	Header file for the RAP protocol.				*/
/*     									*/
/************************************************************************/
#define RapName			"RAP"
#define RapVendor		"X Consortium"
#define RapRelease		"0.1"
#define RapMajorVersion		0
#define RapMinorVersion		1

/* Rap minor opcodes:
 */
typedef enum {
    RapError			= 0,
    RapAcknowledge		= 1,

    RapHelloRequest		= 2,
    RapQueryTreeRequest		= 3,
    RapFullQueryTreeRequest	= 4,
    RapGetResourcesRequest	= 5,
    RapGetValuesRequest		= 6,
    RapGetGCValuesRequest	= 7,
    RapSetValuesRequest		= 8,
    RapWindowToObjectRequest	= 9,
    RapObjectToWindowRequest	= 10,
    RapLocateObjectRequest	= 11,
    RapSelectEventsRequest	= 12,
    RapSelectRequestsRequest	= 13,
    RapGetHooksRequest		= 14,
    RapSelectHooksRequest	= 15,
    RapCloseConnectionRequest	= 16,

    RapHelloReply		= 17,
    RapQueryTreeReply		= 18,
    RapFullQueryTreeReply	= 19,	
    RapGetResourcesReply	= 20,
    RapGetValuesReply		= 21,
    RapGetGCValuesReply		= 22,
    RapWindowToObjectReply	= 23,
    RapObjectToWindowReply	= 24,
    RapLocateObjectReply	= 25,

    RapCreateHookNotify		= 26,
    RapDestroyHookNotify	= 27,
    RapChangeHookNotify		= 28,
    RapConfigureHookNotify	= 29,
    RapGeometryHookNotify	= 30,
    RapActionHookNotify		= 31,

    RapEventNotify		= 32,
    RapRequestNotify		= 33,
    RapGetEventsRequest		= 34,
    RapGetRequestsRequest	= 35,
    RapGetHooksReply		= 36,
    RapGetEventsReply		= 37,
    RapGetRequestsReply		= 38
} RapMessageTypes;

/* RapMessageHeader 
 *
 * This is the header for every Rap message.
 */
typedef struct _RapMessageHeader {
    CARD8	majorOpcode;
    CARD8	minorOpcode;
    CARD16	sequenceNumber;
    CARD32	length;
} RapMessageHeader;
#define sz_RapMessageHeader	8


/* RapErrMessage
 *
 * NOTE:  There is no need to pad here since there is going be a
 *        variable length string attached to end of the msg.
 */
typedef struct _RapErrMessage
{
    CARD8       majorOpcode;
    CARD8       minorOpcode;
    CARD16      sequenceNumber;
    CARD32      length B32;
    CARD16      reqSequenceNumber;   /* seq num of requesting message */
    CARD16      errCode;
} RapErrMessage;

/************************************************************************/
/*									*/
/* PAD64								*/
/* PADDED_BYTES64	        Macros for dealing with 64-bit		*/
/* WORD64COUNT			boundaries.				*/
/* STRING_BYTES								*/
/*									*/
/************************************************************************/
/* compute the 64-bit-boundary padding for the specified number of bytes 
 */
#define PAD64(_bytes)		((8 - ((unsigned int) (_bytes) % 8)) % 8)

/* compute the total size, including padding, for the specified number 
 * of bytes 
 */
#define PADDED_BYTES64(_bytes)	((_bytes) + PAD64(_bytes))

/* Number of 8-byte units in _bytes 
 */
#define WORD64COUNT(_bytes)	(((unsigned int) ((_bytes) + 7)) >> 3)

/* Number of bytes needed to store a string of length _len and its
 * length.
 */
#define STRING_BYTES(_len)	(4 + (_len) + PAD64(4 + (_len)))

/************************************************************************/
/*									*/
/* STORE_OBJECTID							*/
/* STORE_CARD32								*/
/* STORE_CARD16			Macros for storing information into	*/
/* STORE_STRING			a character buffer.			*/
/* STORE_BYTES								*/
/* STORE_SIZED_BYTES							*/
/*									*/
/************************************************************************/
/* Store an RapObjectID
 */
#define STORE_OBJECTID(_buf, _obj) { \
                    *((RapObjectID *) (_buf)) = (_obj); \
		    (_buf) += sizeof(RapObjectID); \
		}

/* Store a CARD32
 */
#define STORE_CARD32(_buf, _val) \
		{ \
	            *((CARD32 *) (_buf)) = (CARD32) (_val); \
		    (_buf) += 4; \
		}

/* Store a CARD16
 */
#define STORE_CARD16(_buf, _val) \
		{ \
	            *((CARD16 *) (_buf)) = (CARD16) (_val); \
		    (_buf) += 2; \
		}

/* Store a character string and its length
 */
#define STORE_STRING(_buf, _len, _str) \
		{ \
		    STORE_CARD32((_buf), (_len)); \
		    memcpy((_buf), (_str), (_len)); \
		    (_buf) += (_len); \
		    (_buf) += PAD64(4 + (_len)); \
		}

/* Store array of bytes and its length 
 */
#define STORE_BYTES(_buf, _len, _array) \
		{ \
      		    STORE_CARD32((_buf), (_len)); \
		    memcpy((_buf), (_array), (_len)); \
		    (_buf) += (_len); \
		    (_buf) += PAD64(4 + (_len)); \
		}
/* Store an array of bytes EXCLUDING its length 
 */
#define STORE_SIZED_BYTES(_buf, _len, _array) \
		{ \
		    memcpy((_buf), (_array), (_len)); \
		    (_buf) += (_len); \
		    (_buf) += PAD64(_len); \
		}

/************************************************************************/
/*									*/
/* LSWAPL		Macros for byte swapping literal values.	*/
/* LSWAPS	       							*/
/*									*/
/************************************************************************/
/* byte swap a long literal 
 */
#define LSWAPL(_val) ((((_val) & 0xff) << 24) |\
                   (((_val) & 0xff00) << 8) |\
                   (((_val) & 0xff0000) >> 8) |\
                   (((_val) >> 24) & 0xff))
 
/* byte swap a short literal 
 */
#define LSWAPS(_val) ((((_val) & 0xff) << 8) | (((_val) >> 8) & 0xff))

/************************************************************************/
/*									*/
/* EXTRACT_OBJECTID							*/
/* EXTRACT_CARD32							*/
/* EXTRACT_CARD16		Macros for extracting information from	*/
/* EXTRACT_STRING		a character buffer.			*/
/* EXTRACT_BYTES							*/
/* EXTRACT_SIZED_BYTES							*/
/*									*/
/************************************************************************/
/* Extract an OBJECTID from a character buffer [[[WDW - Dunno about this]]]
 */
#define EXTRACT_OBJECTID(_buf, _swap, _obj) \
		{ \
      		    (_obj) = *((RapObjectID *) (_buf)); \
		    (_buf) += sizeof(RapObjectID); \
		    if (_swap) {\
			(_obj).id = LSWAPL((_obj).id); \
		    } \
		}

/* Extract a CARD32 from a character buffer
 */    
#define EXTRACT_CARD32(_buf, _swap, _val) \
		{ \
      		    (_val) = *((CARD32 *) (_buf)); \
		    (_buf) += 4; \
		    if (_swap) \
			(_val) = LSWAPL(_val); \
		}

/* Extract a CARD16 from a character buffer 
 */
#define EXTRACT_CARD16(_buf, _swap, _val) \
		{ \
      		    (_val) = *((CARD16 *) (_buf)); \
		    (_buf) += 2; \
		    if (_swap) \
			(_val) = LSWAPS(_val); \
		}

/* Extract a string from a character buffer 
 */
#define EXTRACT_STRING(_buf, _swap, _string) \
		{ \
		    CARD32 _len; \
		    EXTRACT_CARD32(_buf, _swap, _len); \
		    (_string) = (char *) XtMalloc((_len) + 1); \
		    memcpy((_string), (_buf), (_len)); \
		    (_string)[(_len)] = '\0'; \
		    (_buf) += (_len); \
		    (_buf) += PAD64(4 + (_len)); \
		}

/* Extract an array of bytes from a character buffer, reading its length
 */
#define EXTRACT_BYTES(_buf, _swap, _len, _array) \
		{ \
		    char *_temp; \
      		    EXTRACT_CARD32((_buf), (_swap), (_len)); \
		    _temp = (char *) XtMalloc((_len) + 1); \
		    memcpy(_temp, (_buf), (_len)); \
		    _temp[(_len)] = '\0'; \
		    (_buf) += (_len); \
		    (_buf) += PAD64(4 + (_len)); \
		    (_array) = _temp; \
		}

/* Extract an array of bytes from a character buffer, GIVEN its length
 */
#define EXTRACT_SIZED_BYTES(_buf, _swap, _len, _array) \
		{ \
		    (_array) = (char *) XtMalloc((_len) + 1); \
		    memcpy((_array), (_buf), (_len)); \
		    (_array)[(_len)] = '\0'; \
		    (_buf) += (_len); \
		    (_buf) += PAD64(4 + (_len)); \
		}

#define WRITE_STRING(conn, str) \
		{ \
		    CARD32 _size = strlen(str); \
		    register int _j; \
		    char zero = '\0'; \
		    IceWriteData32((conn), sizeof(CARD32), &_size); \
		    IceWriteData((conn), _size, (str)); \
		    for (_j=0 ; _j<PAD64(4 + _size) ; _j++) \
		        IceWriteData((conn), 1, &zero); \
		}

#define WRITE_BYTES(conn, size, bytes) \
		{ \
		    register int _j; \
		    char zero = '\0'; \
		    IceWriteData32((conn), sizeof(CARD32), &(size)); \
		    IceWriteData((conn), (size), (bytes)); \
		    for (_j=0 ; _j<PAD64(4 + (size)) ; _j++) \
		        IceWriteData((conn), 1, &zero); \
		}

#define WRITE_OBJECTID(conn,object_id) \
		{ \
		    IceWriteData32((conn),sizeof(RapObjectID),&(object_id)); \
		}

#endif /* RAPPROTO_H */
