/************************************************************************/
/*									*/
/*   Copyright 1996 by Digital Equipment Corporation, Maynard, Mass.	*/
/*									*/
/*                        All Rights Reserved				*/
/*								        */
/* Permission to use, copy, modify, and distribute this software and    */
/* its documentation for any purpose and without fee is hereby granted,	*/
/* provided that the above copyright notice appear in all copies and 	*/
/* that both that copyright notice and this permission notice appear in */
/* supporting documentation, and that the name Digital not be used in	*/
/* advertising or publicity pertaining to distribution of the software	*/
/* without specific, written prior permission.				*/
/*									*/
/* DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,	*/
/* INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN  */
/* NO EVENT SHALL DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR	*/
/* CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS	*/
/* OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, 		*/
/* NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 		*/
/* CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.		*/
/*								        */
/************************************************************************/
/*
* $Log: Xtea.c,v $
* Revision 0.1  1997/09/02 19:05:02  rcsg
* Initial revision
*
*/

/************************************************************************/
/*									*/
/* Xtea.c								*/
/*									*/
/* Xt external agent enabling code.  This contains the following	*/
/* public entry points:							*/
/*									*/
/* Status								*/
/* XteaSetICEProtocols(Display *, Window, Atom *, int)			*/
/*									*/
/*	Display	*dpy		- The display.				*/
/*	Window	w		- Window with ICE_PROTOCOLS property.	*/
/*	Atom	*prot		- New value for ICE_PROTOCOLS.		*/
/*	int	count		- Number of protocols in prot.		*/
/*									*/
/*	Sets the ICE_PROTOCOLS property on Window.  A return value of	*/
/*	BadAtom indicates the routine could not create or get the	*/
/*	ICE_PROTOCOLS Atom.  Otherwise, the ICE_PROTOCOLS property	*/
/*	is set and the return value is Success.  If count is 0, the 	*/
/*	ICE_PROTOCOLS property is deleted.				*/
/*									*/
/*									*/
/* Status								*/
/* XteaGetICEProtocols(Display *,Window,Atom **, int *)			*/
/*								        */
/*	Display	*dpy		- The display.				*/
/*	Window	w		- Window with ICE_PROTOCOLS property.	*/
/*	Atom	**prot		- Will be populated with ICE_PROTOCOLS.	*/
/*	int	*count		- Will be populated with count.		*/
/*									*/
/*	Gets the ICE_PROTOCOLS property on Window.  If the property	*/
/*	doesn't exist, the return value is BadAtom.  If the property 	*/
/*	is not in the expected format, the return value is BadValue.  	*/
/*	Otherwise, the list and count are populated and the return 	*/
/*	value is Success.  To free the list, use XFree.			*/
/*									*/
/*									*/
/* Boolean								*/
/* XteaWillInitiate(Display *, Window, Atom)				*/
/*									*/
/* 	Display *dpy		- The display.				*/
/* 	Window	w		- The client window.			*/
/* 	Atom	prot		- The protocol.				*/
/*									*/
/*	Determines if the window contains an ICE_PROTOCOLS property 	*/
/*	that contains the protocol.  Returns False if it doesn't and	*/
/*	True if it does.						*/
/*									*/
/*									*/
/* Status								*/
/* XteaUpdateICE_PROTOCOLS(Display *, Window, Atom, Boolean)		*/
/*									*/
/*	Display *dpy		- The display.				*/
/*	Window	w		- The client window.			*/
/*	Atom	prot		- The protocol.				*/
/*	Boolean add		- Add or remove the protocol.		*/
/*									*/
/*	Adds or removes the protocol to/from the ICE_PROTOCOLS property	*/
/*	of the top level window of the app.  Will return BadValue if	*/
/*	there is nothing to remove and add = False.  Otherwise, it will	*/
/*	return the status of XteaSetICE_PROTOCOLS.			*/
/*									*/
/*									*/
/* Status								*/
/* XteaAddProtocol(Widget, Atom, XTeaProtoInitializer, XtPointer)	*/
/*			       						*/
/*	Widget  w		- The widget.				*/
/*	Atom	prot		- The protocol.				*/
/*	XteaProtocolInitializer proc - The init proc.			*/
/*	XtPointer closure 	- Client Data.				*/
/*									*/
/*	Add the protocol, designated by the unique name, prot, and its 	*/
/*	initialization routine to the list of known protocols.  The 	*/
/*	client data is passed onto the init proc.  If the prototocol 	*/
/*	atom is already in the list, the proc and closure entries are	*/
/*	overridden.  Returns BadValue if the protocol or proc are 	*/
/*	missing.  Returns BadAlloc if the new entry could not be	*/
/*	created.  Otherwise, it returns the result of the call to	*/
/*	XteaUpdateICE_PROTOCOLS.					*/
/*									*/
/*									*/
/* status								*/
/* XteaRemoveProtocol(Widget, Atom)					*/
/*									*/
/*	Widget  w	- The widget.					*/
/*	Atom prot	- The atomized name of the protocol.		*/
/*									*/
/*	Remove the protocol from the list of known protocols.  Returns	*/
/*	BadValue if the protocol is not in the list of known protocols.	*/
/*	Otherwise, it returns the status of the call to			*/
/*	XteaUpdateICE_PROTOCOLS.					*/
/*									*/
/*									*/
/* Status								*/
/* XteaSetupClient(Widget)						*/
/*									*/
/*	Widget w       		- The client's top level shell.		*/
/*									*/
/*	Sets up a ClientMessage event handler that listens only for	*/
/*	ClientMessages of type ICE_PROTOCOLS.	Always returns Success.	*/
/*									*/
/*									*/
/* Status								*/
/* XteaSetupAgent(Display *, Window, char *)				*/
/*									*/
/*	Display *dpy		- The display.				*/
/*	Window w		- The agent's window.			*/
/*	char *ids		- The agent's ICE network IDs.		*/
/*									*/
/*	Sets up the ICE_NETWORK_IDS text property on the window passed 	*/
/*	by the agent.  Returns BadValue of the ids property is not	*/
/*	passed.  Otherwise it returns Success.				*/
/*									*/
/*									*/
/* Status								*/
/* XteaInitiateConnection(Display *, Window, Window, Atom)		*/
/*									*/
/*	Display	*dpy		- The display.				*/
/*	Window cw		- Top level window of the client.	*/
/*	Window aw		- Agent window with network ID's prop.  */
/*	Atom prot		- Atomized name of the protocol.	*/
/*									*/
/*	Sets up a ClientMessage on behalf of the agent and sends it to	*/
/*	the client window.  Returns BadMatch if the client does not 	*/
/*	speak the protocol.  Otherwise, it sends the message and 	*/
/*	returns Success.						*/
/*									*/
/*									*/
/* Status								*/
/* XteaSendFailure(Display *, XClientMessageEvent *, Window, int);	*/
/*									*/
/*	Display	*dpy		- The display.				*/
/*	XClientMessageEvent * 	- Original initiate client 		*/
/*				  message event.			*/
/*	Window cw		- Top level window of the client that	*/
/*				  received the ICE_PROTOCOLS message.	*/
/*	int reason 		- Reason for initiate failure.		*/
/*								        */
/*	Sets up a ClientMessage on behalf of the client and sends it to	*/
/*	the agent saying the ICE initiate attempt failed.		*/
/*									*/
/************************************************************************/
#include <X11/IntrinsicP.h>		/* for usual Xt function decls 	*/
#include <X11/Xos.h>			/* for strdup declaration 	*/
#include <X11/Xatom.h>			/* for XA_STRING 		*/
#include<stdio.h>
#include "Xtea.h"


/************************************************************************/
/*									*/
/* Table to hold list of known protocols and their initialization	*/
/* routines.  This is a linked list -- lookup is through a sequential	*/
/* search.								*/
/*									*/
/************************************************************************/
typedef struct _XteaProtoTableEntry 
{
    Widget			topLevel;
    Atom 			protoAtom;
    XteaProtoInitializer	initializerProc;
    XtPointer	    		clientData;  
    struct _XteaProtoTableEntry	*next;
} XteaProtoTableEntry, *XteaProtoTable;


/************************************************************************/
/*									*/
/* Local global variables.  						*/
/*									*/
/* knownProtocols is a table of XteaProtoTable with each entry 		*/
/* containing the atomized name of the protocol as well	as the 		*/
/* protocol's initialization procedure.					*/
/*									*/
/* [[[WDW - Is static information bad for multithreaded apps?]]]	*/
/*									*/
/************************************************************************/
static XteaProtoTable  	knownProtocols=NULL;	


/************************************************************************/
/*									*/
/* Status								*/
/* XteaSetICEProtocols(Display *, Window, Atom *, int)			*/
/*									*/
/*	Display	*dpy		- The display.				*/
/*	Window	w		- Window with ICE_PROTOCOLS property.	*/
/*	Atom	*prot		- New value for ICE_PROTOCOLS.		*/
/*	int	count		- Number of protocols in prot.		*/
/*									*/
/*	Sets the ICE_PROTOCOLS property on Window.  A return value of	*/
/*	BadAtom indicates the routine could not create or get the	*/
/*	ICE_PROTOCOLS Atom.  Otherwise, the ICE_PROTOCOLS property	*/
/*	is set and the return value is Success.  If count is 0, the 	*/
/*	ICE_PROTOCOLS property is deleted.				*/
/*									*/
/************************************************************************/
Status
XteaSetICEProtocols(Display *dpy, 
		    Window  w, 
		    Atom    *proto, 
		    int     count)
{
    Atom prop;

/* See if we can make/get the atom.
     */
    prop = XInternAtom(dpy,ICE_PROTOCOLS,False);
    if (prop == None) return BadAtom;

    /* If the new count is 0, remove the property altogether.
     * Otherwise, set the property to the new value.
     */
    if (count == 0)
	XDeleteProperty(dpy,w,prop);
    else
	XChangeProperty(dpy,w,prop,XA_ATOM,32,
			PropModeReplace,(unsigned char *)proto,count);

    return Success;

} /* XteaSetICEProtocols */


/************************************************************************/
/*									*/
/* Status								*/
/* XteaGetICEProtocols(Display *,Window,Atom **, int *)			*/
/*								        */
/*	Display	*dpy		- The display.				*/
/*	Window	w		- Window with ICE_PROTOCOLS property.	*/
/*	Atom	**prot		- Will be populated with ICE_PROTOCOLS.	*/
/*	int	*count		- Will be populated with count.		*/
/*									*/
/*	Gets the ICE_PROTOCOLS property on Window.  If the property	*/
/*	doesn't exist, the return value is BadAtom.  If the property 	*/
/*	is not in the expected format, the return value is BadValue.  	*/
/*	Otherwise, the list and count are populated and the return 	*/
/*	value is Success.  To free the list, use XFree.			*/
/*									*/
/************************************************************************/
Status
XteaGetICEProtocols(Display *dpy, 
		    Window  w,
		    Atom    **prot, 
		    int     *count)
{
    Atom 	  *data = NULL;
    Atom 	  actual_type;
    Atom 	  prop;
    int 	  actual_format;
    unsigned long leftover, nitems;

/* See if we can even get the atom.
     */
    prop = XInternAtom(dpy, ICE_PROTOCOLS, True);
    if (prop == None) return BadAtom;

    /* Now try to get the property.
     */
    if (XGetWindowProperty(dpy, w, prop,
			   0L, 1000000L, False,
			   XA_ATOM, &actual_type, &actual_format,
			   &nitems, &leftover, (unsigned char **) &data)
	!= Success)
      return BadValue;

    /* Make sure the format is what we expect.
     */
    if (actual_type != XA_ATOM || actual_format != 32) {
	if (data) 
	    XFree((char *) data);
	return BadValue;
    }

    if (prot != NULL)
	*prot = (Atom *) data;
    
    if (count != NULL)
	*count = (int) nitems;
    
    return Success;

} /* XteaGetICEProtocols */


/************************************************************************/
/*									*/
/* Boolean								*/
/* XteaWillInitiate(Display *, Window, Atom)				*/
/*									*/
/* 	Display *dpy		- The display.				*/
/* 	Window	w		- The client window.			*/
/* 	Atom	prot		- The protocol.				*/
/*									*/
/*	Determines if the window contains an ICE_PROTOCOLS property 	*/
/*	that contains the protocol.  Returns False if it doesn't and	*/
/*	True if it does.						*/
/*									*/
/************************************************************************/
Boolean
XteaWillInitiate(Display *dpy,
		 Window	w,
		 Atom	prot)
{
    Atom 	*protocols;
    int	 	numProtocols;
    Boolean 	found=False;
    
if (XteaGetICEProtocols(dpy,w,&protocols,&numProtocols) == Success) {
	while ((numProtocols > 0) && (!found)) {
	    numProtocols--;
	    found = (protocols[numProtocols] == prot);
	}
	XFree(protocols);
    }
        
    return found;
    
} /* XteaWillInitiate */


/************************************************************************/
/*									*/
/* static void								*/
/* XteaCheckMessages(Widget, XtPointer, XEvent *, Boolean *)		*/
/*									*/
/*	XteaCheckMessages is a ClientMessage event handler that handles	*/
/*	incoming ClientMessage events from external agents.  This 	*/
/*	handler is added by a call to XteaSetupClient.			*/
/*									*/
/*	The ClientMessage event looks like the following:       	*/
/*									*/
/*		cme->type		== ClientMessage		*/
/*		cme->message_type	== Atom ("ICE_PROTOCOLS")	*/
/*		cme->format		== 32				*/
/*		cme->data.l[0]		== protoAtom			*/
/*		cme->data.l[1]		== Timestamp			*/
/*		cme->data.l[2]		== agentWindow			*/
/*		cme->data.l[3]		== networkIdsAtom		*/
/*									*/
/*	The protoAtom entry is the name of the protocol the agent 	*/
/*	wishes to speak.  The networkIdsAtom is the list of ICE network	*/
/*	ids the client must use to send an ICE protocol setup message 	*/
/*	to the agent.  The agentWindow is the id of the agent's window 	*/
/*	that contains the networkIds property.				*/
/*									*/
/************************************************************************/
static void 
XteaCheckMessages(Widget	w, 
		  XtPointer 	clientData, 
		  XEvent 	*event, 
		  Boolean 	*cont)
{
    XteaProtoTableEntry	*current = knownProtocols;
    Display 		*theDisplay = XtDisplay(w);
    XClientMessageEvent *cme = (XClientMessageEvent *) event;
    Atom 		protoAtom;
    Atom		networkIdsAtom;
    Window		agentWindow;
    
#ifdef XTEA_DEBUG     /* blsx */
    printf("Xtea.c: in XteaCheckMessages\n");
#endif

    /* Make sure this is a ClientMessage event before doing anything.
     */
    if (event->type != ClientMessage) 
	return;

    /* If the message type is not an ICE_PROTOCOLS message,
     * then we know not to process this event any further.
     */
    if ((cme->message_type != XInternAtom(theDisplay, ICE_PROTOCOLS, False)) ||
	(cme->format != 32)) 
        return;

    /* Get the information supplied by the ClientMessage event.
     */
    protoAtom      = cme->data.l[0];
    agentWindow    = cme->data.l[2];
    networkIdsAtom = cme->data.l[3];

    /* Look in the table of known protocols for the protocol the
     * agent wishes to speak.  If this protocol is in the list of
     * known protocols, call the protocol initialization routine
     * with the agents list of ICE network ids.
     */
    while(current) {
#ifdef XTEA_DEBUG 		/* blsx */
		printf("Xtea.c: looping through known protocols...\n");
#endif
	if (current->protoAtom == protoAtom) {
	    XTextProperty textProp;

	    if(XGetTextProperty(theDisplay,
				agentWindow,
				&textProp,
				networkIdsAtom) != 0) {
		(*(current->initializerProc))(w,
					      cme,
					      (char *)textProp.value,
					      current->clientData);
#ifdef XTEA_DEBUG 		/* blsx */
		printf("Xtea.c: Called protocol initialization routine!\n");
#endif

		XFree(textProp.value);
		return;
	    }
	    else {
		XteaSendFailure(theDisplay,cme,XtWindow(w),
				IceInitiateOpenFailed);
		return;
	    }
	}
	current = current->next;
    }

    /* If we made it this far, this client doesn't know about the
     * the protocol in the list.
     */
    XteaSendFailure(theDisplay,cme,XtWindow(w),IceInitiateUnknownProtocol);
    return;

} /* XteaCheckMessages */


/************************************************************************/
/*									*/
/* Status								*/
/* XteaUpdateICE_PROTOCOLS(Display *, Window, Atom, Boolean)		*/
/*									*/
/*	Display *dpy		- The display.				*/
/*	Window	w		- The client window.			*/
/*	Atom	prot		- The protocol.				*/
/*	Boolean add		- Add or remove the protocol.		*/
/*									*/
/*	Adds or removes the protocol to/from the ICE_PROTOCOLS property	*/
/*	of the top level window of the app.  Will return BadValue if	*/
/*	there is nothing to remove and add = False.  Otherwise, it will	*/
/*	return the status of XteaSetICE_PROTOCOLS.			*/
/*									*/
/************************************************************************/
Status
XteaUpdateICE_PROTOCOLS(Display	*dpy,
			Window	w,
			Atom	prot,
			Boolean add)
{
    Atom	*wmProtocols;
    int		i,pos,count=0;
    Status	status;
    
#ifdef XTEA_DEBUG     /* blsx */
    printf("Xtea.c: in XteaUpdateICE_PROTOCOLS\n");
#endif

/* If we're going to add a protocol, either realloc the existing
     * list or create a new one.
     */
    if (add) {
        if (XteaGetICEProtocols(dpy,w,&wmProtocols,&count) == Success)
	    XtRealloc((char *)wmProtocols,sizeof(Atom)*(count+1));
	else
	    wmProtocols=(Atom *)XtMalloc(sizeof(Atom));
        
	wmProtocols[count] = prot;
	count++;
    }
    /* Otherwise, we're going to remove a protocol.  If the the
     * protocol to remove is not in the list, return FALSE.  
     * Otherwise remove it from the list.
     */
    else if (XteaGetICEProtocols(dpy,w,&wmProtocols,&count) == Success)
    {
	i = 0; pos = -1;
	while((i < count) && (pos == -1)) {
	    if (wmProtocols[i] == prot)
		pos = i;
	    else
		i++;
	}
	if (pos == -1) {	/* Not in list */ 
	    XtFree((char *)wmProtocols);	    
	    return BadValue;
	}
	else {			/* Remove it   */
	    count--;
	    while (pos < count) {
		wmProtocols[pos] = wmProtocols[pos+1];
		pos++;
	    }
	}
    }
    else
	return BadValue; /* No ICE_PROTOCOLS == nothing to remove */

    /* Set the new values 
     */
    status = XteaSetICEProtocols(dpy,w,wmProtocols,count);
    XtFree((char *)wmProtocols);	    
    return status;

} /* XteaUpdateICE_PROTOCOLS */


/************************************************************************/
/*									*/
/* Status								*/
/* XteaAddProtocol(Widget, Atom, XTeaProtoInitializer, XtPointer)	*/
/*			       						*/
/*	Widget  w		- The widget.				*/
/*	Atom	prot		- The protocol.				*/
/*	XteaProtocolInitializer proc - The init proc.			*/
/*	XtPointer closure 	- Client Data.				*/
/*									*/
/*	Add the protocol, designated by the unique name, prot, and its 	*/
/*	initialization routine to the list of known protocols.  The 	*/
/*	client data is passed onto the init proc.  If the prototocol 	*/
/*	atom is already in the list, the proc and closure entries are	*/
/*	overridden.  Returns BadValue if the protocol or proc are 	*/
/*	missing.  Returns BadAlloc if the new entry could not be	*/
/*	created.  Otherwise, it returns the result of the call to	*/
/*	XteaUpdateICE_PROTOCOLS.					*/
/*									*/
/************************************************************************/
Status 
XteaAddProtocol(Widget 			w,
		Atom 		     	prot, 
		XteaProtoInitializer 	proc,
		XtPointer		closure)
{
    XteaProtoTableEntry *newEntry, *current, *prev;

#ifdef XTEA_DEBUG     /* blsx */
    printf("Xtea.c: in XteaAddProtocol\n");
#endif

/* The protocol to be added must be have a name and an
     * initialization procedure.
     */
    if (!prot || (proc == NULL))
	return BadValue;

    /* Create and initialize a table node for the new initializer.
     */
    if ((newEntry = (XteaProtoTableEntry *)XtNew(XteaProtoTableEntry)) == NULL)
	return BadAlloc;

    newEntry->topLevel		= w;
    newEntry->protoAtom 	= prot;
    newEntry->initializerProc 	= proc;
    newEntry->clientData	= closure;
    newEntry->next 		= NULL;

    /* Install it in the table.
     */
    if (knownProtocols == NULL) {
	knownProtocols = newEntry;
	return(XteaUpdateICE_PROTOCOLS(XtDisplay(w),
				       XtWindow(w),
				       prot,
				       True));
    }
    /* Search the list for a matching node and override it, 
     * or append at the end of no matching node is found.
     */
    else {
	current = knownProtocols;
	while (current != NULL) 
	{
	    if (current->protoAtom == prot) {
		current->initializerProc = proc;
		current->clientData = closure;
		XtFree((char *)newEntry);
		break;
	    } 
	    else {
		prev = current;
		current = current->next;
	    }
	}

	/* We're at the end of the list and haven't found a match. 
	 */
	if (current == NULL) {
	    prev->next = newEntry;
	    return(XteaUpdateICE_PROTOCOLS(XtDisplay(w),
					   XtWindow(w),
					   prot,
					   True));
	}
    }

    return Success;

} /* XteaAddProtocol */


/************************************************************************/
/*									*/
/* status								*/
/* XteaRemoveProtocol(Widget, Atom)					*/
/*									*/
/*	Widget  w	- The widget.					*/
/*	Atom prot	- The atomized name of the protocol.		*/
/*									*/
/*	Remove the protocol from the list of known protocols.  Returns	*/
/*	BadValue if the protocol is not in the list of known protocols.	*/
/*	Otherwise, it returns the status of the call to			*/
/*	XteaUpdateICE_PROTOCOLS.					*/
/*									*/
/************************************************************************/
Status
XteaRemoveProtocol(Widget w,
		   Atom prot)
{
    XteaProtoTableEntry *current = knownProtocols;
    XteaProtoTableEntry *prev = NULL;
    Status		status = BadValue;
    
#ifdef XTEA_DEBUG     /* blsx */
    printf("Xtea.c: in XteaRemoveProtocol\n");
#endif

while (current) {
	if (current->protoAtom == prot) {
	    if (prev) 
		prev->next = current->next;
	    else 
	        knownProtocols = current->next;

	    status = XteaUpdateICE_PROTOCOLS(XtDisplay(w),
					     XtWindow(w),
					     prot,
					     False);
	    XtFree((char *)current);
	    break;
	}

	prev = current;
	current = current->next;
    }
    return status;

} /* XteaRemoveProtocol */


/************************************************************************/
/*									*/
/* Status								*/
/* XteaSetupClient(Widget)						*/
/*									*/
/*	Widget w       		- The client's top level shell.		*/
/*									*/
/*	Sets up a ClientMessage event handler that listens only for	*/
/*	ClientMessages of type ICE_PROTOCOLS.	Always returns Success.	*/
/*									*/
/************************************************************************/
Status
XteaSetupClient(Widget w)
{
    /* Set up the ClientMessage handler.
     */
#ifdef XTEA_DEBUG /* blsx */
printf("Xtea.c: in XtSetupClient\n");
#endif

    XtAddEventHandler(w,		/* The widget to receive the events */
                      (EventMask) 0,	/* The EventMask		    */
		      TRUE,		/* Call on nonmaskable events	    */
	              XteaCheckMessages,/* The handler.			    */
	              NULL);		/* Client data.			    */

    return Success;
    
} /* XteaSetupClient */


/************************************************************************/
/*									*/
/* Status								*/
/* XteaSetupAgent(Display *, Window, char *)				*/
/*									*/
/*	Display *dpy		- The display.				*/
/*	Window w		- The agent's window.			*/
/*	char *ids		- The agent's ICE network IDs.		*/
/*									*/
/*	Sets up the ICE_NETWORK_IDS text property on the window passed 	*/
/*	by the agent.  Returns BadValue of the ids property is not	*/
/*	passed.  Otherwise it returns Success.				*/
/*									*/
/************************************************************************/
Status
XteaSetupAgent(Display	*dpy,
	       Window	w,
	       char	*ids)
{
    XTextProperty	textProp;
    
#ifdef XTEA_DEBUG     /* blsx */
    printf("Xtea.c: in XteaSetupAgent\n");
#endif

if (ids == NULL)
	return BadValue;
    
    textProp.value 	= (unsigned char *)ids;
    textProp.encoding 	= XA_STRING;
    textProp.format 	= 8;
    textProp.nitems 	= strlen(ids);

    XSetTextProperty(dpy,w,
		     &textProp,
		     XInternAtom(dpy,ICE_NETWORK_IDS,False));

    return Success;
    
} /* XteaSetupAgent */


/************************************************************************/
/*									*/
/* Status								*/
/* XteaInitiateConnection(Display *, Window, Window, Atom)		*/
/*									*/
/*	Display	*dpy		- The display.				*/
/*	Window cw		- Top level window of the client.	*/
/*	Window aw		- Agent window with network ID's prop.  */
/*	Atom prot		- Atomized name of the protocol.	*/
/*									*/
/*	Sets up a ClientMessage on behalf of the agent and sends it to	*/
/*	the client window.  Returns BadMatch if the client does not 	*/
/*	speak the protocol.  Otherwise, it sends the message and 	*/
/*	returns Success.						*/
/*									*/
/************************************************************************/
Status
XteaInitiateConnection(Display	*dpy,
		       Window	cw,
		       Window	aw,
		       Atom	prot)
{
    Atom		*wmProtocols;
    int			found,count;
    XClientMessageEvent	cme;
    Status		status;
    
#ifdef XTEA_DEBUG     /* blsx */
    printf("Xtea.c: in XteaInitiateConnection\n");
#endif

/* Make sure the clientWindow has the protocol atom in its 
     * ICE_PROTOCOLS property.
     */
    found  = 0;
    status = XteaGetICEProtocols(dpy,cw,&wmProtocols,&count);
    
    if (status == Success) {
	int i = 0;
	while ((i < count) && (found == 0)) {
	    if (wmProtocols[i] == prot)
		found = 1;
	    else
		i++;
	}
	XtFree((char *)wmProtocols);	    
    }
    else
	return status;
    
    if (found == 0)
	return BadMatch;
    
    cme.window 	  	= cw;
    cme.type 	   	= ClientMessage;
    cme.message_type 	= XInternAtom(dpy,ICE_PROTOCOLS,False);
    cme.format 		= 32;
    cme.data.l[0] 	= prot;
    cme.data.l[1] 	= CurrentTime;
    cme.data.l[2] 	= aw;
    cme.data.l[3] 	= XInternAtom(dpy,ICE_NETWORK_IDS,False);
    cme.data.l[4] 	= 0L;

    XSendEvent(dpy,cw,True,0,(XEvent *)&cme);

    return Success;
    
} /* XteaInitiateConnection */

/************************************************************************/
/*									*/
/* Status								*/
/* XteaSendFailure(Display *, XClientMessageEvent *, Window, int);	*/
/*									*/
/*	Display	*dpy		- The display.				*/
/*	XClientMessageEvent * 	- Original initiate client 		*/
/*				  message event.			*/
/*	Window cw		- Top level window of the client that	*/
/*				  received the ICE_PROTOCOLS message.	*/
/*	int reason 		- Reason for initiate failure.		*/
/*								        */
/*	Sets up a ClientMessage on behalf of the client and sends it to	*/
/*	the agent saying the ICE initiate attempt failed.		*/
/*									*/
/************************************************************************/
Status
XteaSendFailure(Display			*dpy,
		XClientMessageEvent     *initCme,
		Window			cw,
		unsigned int		reason)
{
    XClientMessageEvent	cme;

#ifdef XTEA_DEBUG     /* blsx */
    printf("Xtea.c: in XteaSendFailure\n");
#endif

cme.window 		= initCme->data.l[2];
    cme.type 		= ClientMessage;
    cme.message_type 	= XInternAtom(dpy, ICE_INITIATE_FAILED, False);
    cme.format 		= 32;
    cme.data.l[0] 	= initCme->data.l[0]; /* ICE Subprotocol Atom   */
    cme.data.l[1] 	= initCme->data.l[1]; /* Timestamp		*/
    cme.data.l[2] 	= cw;
    cme.data.l[3]      	= reason;
    cme.data.l[4] 	= 0L;

    XSendEvent(dpy,initCme->data.l[1],True,0,(XEvent *)&cme);

    return Success;
    
} /* XteaSendFailure */



