/*	Note:  The use of this program shall be restricted to educational,
*	non-commercial use only.  All inquiries regarding licensing this
*	software for commercial use may be addressed to:
*	
*	Georgia Tech Research Corporation
*	Technology Licensing
*	Centennial Research Building - Rm. 275
*	400 Tenth Street, N.W.
*	Atlanta, Georgia  30332-0415
*	
*	*************************************************
*	
*	       Copyright 1996
*	Georgia Tech Research Corporation
*	    All Rights Reserved
*/

/*
* $Log: dectalk.c,v $
* Revision 0.1  1997/09/02 19:08:14  rcsg
* Initial revision
*
*/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/wait.h>

#include <netinet/in.h>

#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <signal.h>
#include <sgtty.h>
#include <stdio.h>
#include <fcntl.h>

int
open_dectalk(char *hostname)
{
  int dectalk;
  int port = 1330;
  struct  sockaddr_in sin;
  struct hostent * host;

  dectalk = socket(AF_INET, SOCK_STREAM, 0);
  if (dectalk == -1) {
    perror("dectalk: socket");
    return -1;
    }

  if (hostname)
      host = gethostbyname(hostname);
  else
      host = gethostbyname("localhost");
  
  if (host == NULL) {
    perror("dectalk: gethostbyname");
    return -1;
    }

  sin.sin_family = AF_INET;
  sin.sin_port = htons(port);
  sin.sin_addr.s_addr = *(u_long*) host->h_addr;
  if (connect(dectalk, (struct sockaddr *) &sin, sizeof (sin))) {
    perror("dectalk: connect");
    return -1;
    }

  return dectalk;
}
  
int
set_speech_rate(int dectalk, int rate)
{
  char buffer[256];

  sprintf(buffer, "\033P0;0z:ra%d\033\\", rate);
  write(dectalk, buffer, strlen(buffer));
}



int
stop_speech(int dectalk)
{
  char * buffer = "\033P0;10z\033\\";
  write(dectalk, buffer, 9);
}

int
mark_index(int dectalk, int index)
{
  char buffer[256];

  sprintf(buffer, "\033P0;20;%dz\033\\", index);
  write(dectalk, "\033P0;20;1z\033\\", 11);
}



int
sync_speech(int dectalk)
{
  char * buffer = "\033P0;11z\033\\";
  write(dectalk, buffer, 9);
}


int
dial_phone(int dectalk, int phone_number)
{
  char buffer[256];

  sprintf(buffer, "\033P0;60;20;40z%s\033\\", phone_number);
  write(dectalk, buffer, strlen(buffer));
}

int 
hangup_phone(int dectalk)
{
  char * buffer = "P0;60;11z\\";
  write(dectalk, buffer, strlen(buffer));
}

int
add_to_dictionary(int dectalk, char *word, char *pronun)
{
  char buffer[256];
  char * strchr();

  *strchr(pronun, '\n') = '\0';
  sprintf(buffer, "P0;40z %s %s \\", word, pronun);
  write(dectalk, buffer, strlen(buffer));
}


#define is_member(c, s)  strchr((s), (c))
#define SPECIAL_CHARS	"!@#$%^&*()/[]{}"

int
speak_unix(int dectalk, char *word)
{
  int size = strlen(word);
  int i = 0, o = 0;
  char buffer[256];
  char last_letter;

  while ((word[i] == ' ') && (i < size)) i++;

  for (o = 0; i < size; i++) {
    if (iscntrl(word[i]))
      continue;

    if (isalnum(word[i])) {
      buffer[o++] = word[i];
      continue;
      }

    if (word[i] == ' ') {
      buffer[o++] = word[i];
      continue;
      }

    if (is_member(word[i], SPECIAL_CHARS)) {
      buffer[o++] = ' ';
      buffer[o++] = word[i];
      buffer[o++] = ' ';
      }
    }

  if (o == 0) return;
  o--;
  while (buffer[o] == ' ')
    o--;
  o++;
  buffer[o++] = '.'; /* awful hack to get it to talk immediately */
  buffer[o++] = '\n';
  buffer[o] = '\0';
  write(dectalk, buffer, o);
}
