/* $Header: /usr/mercator/newsource/dectalk/RCS/dectalkd.c,v 0.2 1998/05/21 23:46:49 sellden Exp sellden $ */
/*	Note:  The use of this program shall be restricted to educational,
*	non-commercial use only.  All inquiries regarding licensing this
*	software for commercial use may be addressed to:
*	
*	Georgia Tech Research Corporation
*	Technology Licensing
*	Centennial Research Building - Rm. 275
*	400 Tenth Street, N.W.
*	Atlanta, Georgia  30332-0415
*	
*	*************************************************
*	
*	       Copyright 1996
*	Georgia Tech Research Corporation
*	    All Rights Reserved
*/

/* 
 * dectalkd.c - Simple daemon to shovel characters to/from DECtalk
 * 
 * Author:	Christopher A. Kent
 * 		Dept. of Computer Sciences
 * 		Purdue University
 * Date:	Mon Aug 20 1984
 * Copyright (c) 1984 Christopher A. Kent
 *
 * phone hacks:
 *		Bala Krishnamurthy Mon Jul 14 11:12:29 1986
 *		bala@purdue
 *
 *  commented out SO_DONTLINGER for 4.3 NFS
 */

/*
 * $Log: dectalkd.c,v $
 * Revision 0.2  1998/05/21 23:46:49  sellden
 * Changed global var -daemon- to -isDaemon- to avoid conflicts
 *
 * Revision 0.1  1997/09/02 19:08:14  rcsg
 * Initial revision
 *
 */

static char rcs_ident[] = "$Header: /usr/mercator/newsource/dectalk/RCS/dectalkd.c,v 0.2 1998/05/21 23:46:49 sellden Exp sellden $";

#ifdef USONIX
#include <sys/time.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <netinet/in.h>

#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <termio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef __linux__
#define	TALKER	"/dev/ttyS2"
#else
#define	TALKER	"/dev/ttyb"
#endif

#ifdef USONIX
static struct	sockaddr_in sin = { AF_INET};
#else
struct	sockaddr_in sin = { AF_INET};
#endif
extern	int errno;
int isDaemon = 1;
char *ttydev = NULL;

void
reapchild(int dummy)
{
    pid_t pid;

    while (pid = waitpid((pid_t) -1, NULL, WNOHANG)) {
	if (pid < 0) {
	    if (errno == EINTR)
		continue;
	    if (errno == ECHILD)
		break;

	    exit(1);
	} else
	    return;
    }
}

int
main(int argc, char **argv)
{
    struct sockaddr_in from;
    struct servent *sp;
    int f, options = 0;
    int c;
    extern char *optarg;
    int port;
    int one = 1;
    struct linger ling;

    ttydev = strdup(TALKER);

    /*
      sp = getservbyname("dectalk", "tcp");
      if (sp == 0) {
      fprintf(stderr, "dectalkd: tcp/dectalk: unknown service\n");
      exit(1);
      }
      sin.sin_port = sp->s_port;
      */
    sin.sin_port = 1330;

    /*
     * Args:
     * -d 		enable debugging
     * -f		run in foreground (no daemon mode)
     * -t <dev>	tty to use
     * -p <port>	specify alternate port
     */
    while ((c = getopt(argc, argv, "dft:p:")) != EOF) {
	switch (c) {
	case 'd':
	    options |= SO_DEBUG;
	    break;
	case 'f':
	    isDaemon = 0;
	    break;
	case 'p':
	    port = atoi(optarg);
	    sin.sin_port = htons((u_short) port);
	    break;
	case 't':
	    free(ttydev);
	    ttydev = strdup(optarg);
	    break;
	default:
	    fprintf(stderr, "Usage: %s: -d -f -t tty -p port\n\t-d\tEnable debugging\n\t-f\tRun in foreground\n\t-t tty\tUse specified tty\n\t-p port\tUse specified port\n", argv[0]);
	    exit(1);
	}
    }
    f = socket(AF_INET, SOCK_STREAM, 0);
    if (f < 0) {
	perror("dectalkd: socket");
	exit(1);
    }

    ling.l_onoff = 0;
    
    if (options & SO_DEBUG)
	if (setsockopt(f, SOL_SOCKET, SO_DEBUG, (const char *) &one, 1) < 0)
	    perror("dectalkd: setsockopt (SO_DEBUG)");
    if (setsockopt(f, SOL_SOCKET, SO_KEEPALIVE, (const char *) &one,
		   sizeof(int)) < 0)
	perror("dectalkd: setsockopt (SO_KEEPALIVE)");
    if (setsockopt(f, SOL_SOCKET, SO_REUSEADDR, (const char *) &one,
		   sizeof(int)) < 0)
	perror("dectalkd: setsockopt (SO_REUSEADDR)");
    if (setsockopt(f, SOL_SOCKET, SO_LINGER, (const char *) &ling,
		   sizeof(struct linger)) < 0)
	perror("dectalkd: setsockopt (SO_LINGER)");
    
    if (bind(f, (struct sockaddr *) &sin, sizeof (sin)) < 0) {
	perror("dectalkd: bind");
	exit(1);
    }
    signal(SIGCHLD, reapchild);
    if (isDaemon && fork())
	exit(0);
    listen(f, 1);
    for (;;) {
	int s, len = sizeof (from);

	s = accept(f, (struct sockaddr *) &from, &len);
	if (s < 0) {
	    if (errno == EINTR)
		continue;
	    perror("dectalkd: accept");
	    continue;
	}
	doit(s, &from);
	close(s);
    }

}

int
doit(int f, struct sockaddr_in *fromp)
{
    int dt;
    struct termio ttyconfig;
    int fcc=0, dcc=0;
    int c;
    char *mesg = "This is DECtalk\n";
    int on = 1;

    dt = open(ttydev, O_RDWR|O_NONBLOCK, 0);
    if (dt < 0) {
	perror("Can't open DECtalk");
	exit(1);
    }
    if (lockf(dt, F_TLOCK, 0) < 0) {
	if (errno == EAGAIN) {
	    fprintf(stderr, "Another process has tty locked.\n");
	    exit(1);
	} else {
	    perror("lockf failed");
	    exit(2);
	}
    }
    /* 
     * Talker wants 1200 baud, raw, no echo.
     */
    if (ioctl(dt, TCGETA, &ttyconfig) == -1) {
	perror("ioctl");
	lockf(dt, F_ULOCK, 0);
	exit(1);
    }

    ttyconfig.c_cflag &= ~(CBAUD|CIBAUD);
    ttyconfig.c_cflag |= B9600;

    /* turn off many port defaults */
    ttyconfig.c_lflag &= ~(ICANON|ECHO);
    if (ioctl(dt, TCSETA, &ttyconfig) == -1) {
	perror("ioctl");
	lockf(dt, F_ULOCK, 0);
	exit(1);
    }
	
    write(f, mesg, strlen(mesg));

    for (;;) {
	fd_set ibits, obits;

	FD_ZERO(&ibits);
	FD_ZERO(&obits);
	FD_SET(f, &ibits);
	FD_SET(dt, &ibits);

	select(((f > dt) ? f : dt) + 1, &ibits, &obits, 0, 0);

	if (FD_ISSET(f,&ibits)) {
	    fcc = read(f, &c, 1);
	    if (fcc < 0 && errno == EWOULDBLOCK)
		fcc = 0;
	    else
		write(dt, &c, 1);
	}
	if (FD_ISSET(dt, &ibits)) {
	    dcc = read(dt, &c, 1);
	    if (dcc < 0 && errno == EWOULDBLOCK)
		dcc = 0;
	    else
		write(f, &c, 1);
	}
	if (fcc <= 0 || dcc < 0)
	    break;
    }

    lockf(dt, F_ULOCK, 0);
    close(dt);
    close(f);
}
