/* $Header: /usr/mercator/newsource/dectalk/RCS/speak.c,v 0.1 1997/09/02 19:08:14 rcsg Exp $ */

/*	Note:  The use of this program shall be restricted to educational,
*	non-commercial use only.  All inquiries regarding licensing this
*	software for commercial use may be addressed to:
*	
*	Georgia Tech Research Corporation
*	Technology Licensing
*	Centennial Research Building - Rm. 275
*	400 Tenth Street, N.W.
*	Atlanta, Georgia  30332-0415
*	
*	*************************************************
*	
*	       Copyright 1996
*	Georgia Tech Research Corporation
*	    All Rights Reserved
*/
/* 
 * speak.c - Connect to DECtalk server and speak!
 * 
 * Author:	Christopher A. Kent
 * 		Dept. of Computer Sciences
 * 		Purdue University
 * Date:	Mon Aug 20 1984
 * Copyright (c) 1984 Christopher A. Kent
 *
 */

/*
 * $Log: speak.c,v $
 * Revision 0.1  1997/09/02 19:08:14  rcsg
 * Initial revision
 *
 */

static char rcs_ident[] = "$Header: /usr/mercator/newsource/dectalk/RCS/speak.c,v 0.1 1997/09/02 19:08:14 rcsg Exp $";

#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <netinet/in.h>      

#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <stdio.h>

char	*host = "merlin.cs.purdue.edu";		/* oh, for resource location... */
char	*phrase = "This is Worker speaking please press a key.";
int	s;				/* socket to dectalk */
int	GiveUp();

main(argc, argv)
char **argv;
{
	struct	servent *sp;		/* service table entry */
	struct	hostent	*hp;		/* host table entry */
	struct	sockaddr_in sin;	/* socket destination address */
	char	*name;			/* my name */
	char	*number = NULL;		/* number to dial (if any) */
	char	buf[132];		/* input buffer */
	char	c;			/* tiny input buffer */
	int	i;

	name = argv[0];

	argc--, argv++;
	while (argc > 0 && **argv == '-') {
		switch(argv[0][1]) {
		case 'd':
			number = argv[1];
			argc-=2, argv+=2;
			break;

		case 'h':
			host = argv[1];
			argc-=2, argv+=2;
			break;

		case 'p':
			phrase = argv[1];
			argc-=2, argv+=2;
			break;

		default:
			printf("usage: %s [-d number] [-p phrase]", name);
			printf(" [-h host]\n");
			exit(1);
		}
	}

	if((hp = gethostbyname(host)) == (struct hostent *)NULL){
		fprintf(stderr, "%s: Unknown host %s\n", name, host);
		exit(1);
	}

	if((sp = getservbyname("dectalk", "tcp")) == (struct servent *)NULL) {
		fprintf(stderr, "%s: tcp/dectalk: unknown service\n", name);
		exit(1);
	}

	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		sprintf(buf, "%s: socket", argv[0]);
		perror(buf);
		exit(1);
	}

	sin.sin_family = hp->h_addrtype;
	bcopy(hp->h_addr, (caddr_t)&sin.sin_addr, hp->h_length);
	sin.sin_port = sp->s_port;
        if (connect(s, (caddr_t)&sin, sizeof(sin), 0) < 0) {
                sprintf(buf, "%s: Can't connect to port %d at %s",
                           name, ntohs(sin.sin_port), host);
		perror(buf);
		exit(1);
	}

	do{
		i = read(s, &c, 1);
		if (i <= 0){		/* wait out the prompt */
			fprintf(stderr, "talker already in use\n");
			exit(0);
		}
	}while (c != '\n');

	signal(SIGHUP, GiveUp);
	signal(SIGINT, GiveUp);
	signal(SIGTERM, GiveUp);

	if (number != NULL)
		dtDial(number);
	if (argc == 0) {
		while(fgets(buf, sizeof(buf) - 1, stdin) != NULL)
			write(s, buf, strlen(buf));
	} else while (argc-- > 0) {
		write(s, " ", 1);
		write(s, *argv, strlen(*argv));
		argv++;
	}

	dtSync();

	if (number != NULL)
		dtHangUp();

	exit(0);
}

dtDial(n)
char *n;
{
	char	dialbuf[132];
	char	c;
	int	tries = 8;

	sprintf(dialbuf, "[:np]\033P0;60;11;40z%s\033\\", n);
	write(s, dialbuf, strlen(dialbuf));
	signal(SIGALRM, GiveUp);
	alarm(30);
	dtSwallowReply();
	alarm(0);
	sleep(1);

	/* 
	 * Prompt the user with the specified phrase until he
	 * signifies that he's answered the phone. (Keypad was
	 * enabled in the original string.)
	 */

	if (strlen(phrase) == 0)	/* null phrase -> no prompting */
		return;

	sleep(2);
	while (--tries){
		write(s, phrase, strlen(phrase));	/* prompt */
		dtSync();
		sprintf(dialbuf, "\033P0;60;20;30;3z\033\\");
		write(s, dialbuf, strlen(dialbuf));	/* wait for key */
		dtSwallowReply();
		read(s, &c, 1);
		if (c == '\033')
			while (read(s, &c, 1) && c != '\\')
				;	/* must be a PH_TIMEOUT */
		else
		break;		/* user typed something */
	}

	fprintf(stderr, "%s's home\n", tries?"Someone":"No one");

	/* 
	 * Now, need to disable the keypad waiting garbage.
	 */

	sprintf(dialbuf, "\033P0;60;21;30;0z\033\\");
	write(s, dialbuf, strlen(dialbuf));
	dtSwallowReply();
}

dtHangUp()
{
	char	hangbuf[32];
	char	c;

	sprintf(hangbuf, "\033P0;60;11z\033\\"); /* hang up the phone */
	write(s, hangbuf, strlen(hangbuf));
	dtSwallowReply();
}

dtSync()
{
	char	buf[32];

	sprintf(buf, "\033P0;11z\033");
	write(s, buf, strlen(buf));
}

dtSwallowReply()
{
	char	c;
	int	i;

	do {
		i = read(s, &c, 1);
		if (i < 0){
			perror("SwallowReply");
			exit(1);
		}

		if (i == 0)
			GiveUp();
	} while (c != '\\');
}

GiveUp()
{
	dtHangUp();
	exit(2);
}
